
#include "rhis.h"
#include "rhtree.h"

//Cache management

PageCache::PageCache()
{
	int i;
	
	numPages = 0;
	for (i=0; i< MAX_CACHE; i++)
	{
		m_pagecache[i] = NULL;
		m_offsetpage[i] = RWNIL;
		m_lockpagecount[i] = 0;
		m_modifiedflag[i] = FALSE;
	}
	m_pfm = NULL;

}

PageCache::~PageCache()
{
	if (NULL != m_pfm)
	{
		FlushCache();
		delete m_pfm;
		m_pfm = NULL; // Not necessary though.
	}
}

int PageCache::CreateIndex(const char *szFileName)
{
	if (NULL != m_pfm)
	{
		advise("PageCache::CreateIndex(): m_pfm not null and index already exists.");
		return 1;
	}
	m_filename = szFileName;

	m_pfm = new RWFileManager(m_filename);

	if (NULL == m_pfm)
	{
		advise("PageCache::CreateIndex(): Unable to create the file manager for deletion.");
		return 1;
	}

	m_pfm->Erase(); // Erase the file..
	delete m_pfm; // Start it again...
	
	m_pfm = new RWFileManager(m_filename);

	if (NULL == m_pfm)
	{
		advise("PageCache::CreateIndex(): Unable to create the file manager.");
		return 1;
	}

	//allocate for the offset to start the file
	m_pfm->allocate(sizeof(RWoffset));
	RWoffset offset = RWNIL;
	m_pfm->SeekTo(m_pfm->start());
	m_pfm->Write(offset);
	
	return 0;
}

int PageCache::OpenIndex(const char *szFileName)
{
	if (NULL != m_pfm)
	{
		advise("PageCache::OpenIndex(): m_pfm not null and index already exists.");
		return 1;
	}
	m_filename = szFileName;

	m_pfm = new RWFileManager(m_filename);

	if (NULL == m_pfm)
	{
		advise("PageCache::OpenIndex(): Unable to create the file manager.");
		return 1;
	}

	return 0;
}

int PageCache::CloseIndex()
{
	int nRetVal = 0;

	if (NULL != m_pfm)
	{
		nRetVal = FlushCache();
		delete m_pfm;
		m_pfm = NULL;
	} else
		advise("PageCache::CloseIndex(): No active Index");

	return nRetVal;
}

int PageCache::RemoveIndex()
{
	int nRetVal = 0;

	if (NULL != m_pfm)
	{
		FlushCache();
		m_pfm->Erase();
		delete m_pfm;
		m_pfm = NULL;
	} else
		advise("PageCache::CloseIndex(): No active Index");

	return nRetVal;
}

int PageCache::FlushCache()
{
	int i;

	if (NULL == m_pfm)
	{
	 	advise("PageCache::FlushCache(): No active Index");
		return 1;
	}

	for (i=0; i<MAX_CACHE; i++)
	{
		if (NULL != m_pagecache[i])
		{
		 	if (0 == m_lockpagecount[i])
			{
				if (m_modifiedflag[i])
				{
					m_pagecache[i]->WritePage(m_pfm, m_offsetpage[i]);
					stats.IncWritePages(m_pagecache[i]->GetPageType());
				}
			} else
			{
			 	advise("PageCache::FlushCache(): Locked page.");
				return 1;
			}
			delete m_pagecache[i];
		}

		m_pagecache[i] = NULL;
		m_lockpagecount[i] = 0;
		m_modifiedflag[i] = FALSE;
		m_offsetpage[i] = RWNIL;
	}
	return 0;
}

RWoffset PageCache::NewPage(int nType)
{
	int offset = RWNIL;

	if (NULL == m_pfm)
	{
	 	advise("PageCache::NewPage(): No active Index");
		return RWNIL;
	}

	switch (nType)
	{
	 	case HEADPAGETYPE:
			{
				HeadPage headPage;
				headPage.SetPageType(nType);
				offset = m_pfm->allocate(headPage.GetSize());
				headPage.WritePage(m_pfm, offset);

				break;
			}

		case DIRPAGETYPE:
		case DATAPAGETYPE:
			{
				DataPage dataPage;
				dataPage.SetPageType(nType);
				offset = m_pfm->allocate(dataPage.GetSize());
				dataPage.WritePage(m_pfm, offset);

				break;
			}

		default:
			advise("PageCache::NewPage(): Unknown type.");
			return RWNIL;
	}
	if (RWNIL == offset)
	{
	 	advise("PageCache::NewPage(): unable to allocate the page.");
		return RWNIL;
	}

	stats.IncNewPages(nType);
	return offset;
}

StoragePage *PageCache::GetPage(RWoffset offset, int nType)
{
	int i;


	if (RWNIL == offset)
	{
		advise("PageCache::GetPage(): NIL Page offset passed");
		return NULL;
	}
	for (i=0; i < MAX_CACHE; i++)
	{
	 	if (offset == m_offsetpage[i])
		{
			StoragePage *pPage = m_pagecache[i];
			IncreasePriorityPage(i);
			LockPage(offset, nType);
			return pPage;
		}
		else if (NULL == m_pagecache[i])
			break;
	}

	StoragePage *pStorage = ReadDiskPage(offset, nType);

	if (NULL == pStorage)
	{
		advise("PageCache::GetPage(): Null page read from the file");
		return NULL;
	}

	// if there is space in the cache.
	if (i >= MAX_CACHE)
	{
	 	// Get rid of one of the pages.
	 	i = DumpPageToDisk();
		if (i <0)
		{
			advise("PageCache::GetPage(): Unable to purge the page.");
			return NULL;
		}
	}

 	m_pagecache[i] = pStorage;
	m_offsetpage[i] = offset;
	m_lockpagecount[i] = FALSE;
	m_modifiedflag[i] = FALSE;

	IncreasePriorityPage(i);
	LockPage(offset, nType);

	return pStorage;

}

int PageCache::PutPage(RWoffset offset, int nType)
{
	int i;

	i = FindPage(offset, nType);

	if (i < 0)
	{
		 	advise("PageCache::PutPage(): Unable to find the page.");
			return 1;
	}

	m_modifiedflag[i] = TRUE;

	return UnLockPage(offset, nType);
}

int PageCache::LockPage(RWoffset offset, int nType)
{
	int i;

	i = FindPage(offset, nType);

	if (i < 0)
	{
		 	advise("PageCache::LockPage(): Unable to find the page.");
			return 1;
	}

	m_lockpagecount[i] ++;

	IncreasePriorityPage(i);

	return 0;
}

int PageCache::UnLockPage(RWoffset offset, int nType)
{
	int i;

	i = FindPage(offset, nType);

	if (i < 0)
	{
	 	advise("PageCache::UnLockPage(): Unable to find the page.");
		return 1;
	}

	if (m_lockpagecount[i] >0)
		m_lockpagecount[i] --;
	else
	{
	 	advise("PageCache::UnLockPage(): The page is not locked.");
		return 1;
	}

	return 0;
}


int PageCache::FindPage(RWoffset offset, int nType)
{
	int i;

	ASSERT(RWNIL != offset);

	for (i=0; i < MAX_CACHE; i++)
	{
	 	if (offset == m_offsetpage[i])
		{
			return i;
		}
		else if (NULL == m_pagecache[i])
		{
		 	advise("PageCache::LockPage(): Unable to find the page");
			return -1;
		}
	}

	return -1;
}

int PageCache::DumpPageToDisk()
{
	int i;

	//Find a missed empty page....
	for (i = 0 ; i < MAX_CACHE; i++)
		if (NULL == m_pagecache[i])
			return i;

	// Start from the back since LRU cache is being implemented.
	for (i=MAX_CACHE -1 ; i >=0; i--)
	{
		 	// This should not happen..
		 	if (NULL == m_pagecache[i])
				return i;

			//Find an unlocked page
		 	if (0 == m_lockpagecount[i])
			{
				if (m_modifiedflag[i])
				{
					m_pagecache[i]->WritePage(m_pfm, m_offsetpage[i]);
					stats.IncWritePages(m_pagecache[i]->GetPageType());
				}
				delete m_pagecache[i];

				m_pagecache[i] = NULL;
				m_lockpagecount[i] = 0;
				m_modifiedflag[i] = FALSE;
				m_offsetpage[i] = RWNIL;
				break;
			}		
	}

	return i;
}


int PageCache::IncreasePriorityPage(int nNum)
{
	StoragePage *pPage;
	int i, LockCount, ModifiedFlag;
	RWoffset offset;

	
	if ((nNum >= MAX_CACHE) || (nNum < 0))
	{
		advise("PageCache::IncreasePriorityPage(): The index is out of range.");
		return 1;
	}
	pPage = m_pagecache[nNum];
	LockCount = m_lockpagecount[nNum];
	ModifiedFlag = m_modifiedflag[nNum];
	offset = m_offsetpage[nNum];

	for (i= nNum; i >0; i--)
	{
		m_pagecache[i] = m_pagecache[i-1];
		m_lockpagecount[i] = m_lockpagecount[i-1];
		m_modifiedflag[i] = m_modifiedflag[i-1];
		m_offsetpage[i] = m_offsetpage[i-1];
	}

	m_pagecache[0] = pPage;
	m_lockpagecount[0] = LockCount;
	m_modifiedflag[0] = ModifiedFlag;
	m_offsetpage[0] = offset;

	return 0;
	
}

//Read from the disk
StoragePage* PageCache::ReadDiskPage(RWoffset nOffset, int nType)
{
	StoragePage* pPage = NULL;
	
	if (NULL == m_pfm)
	{
		advise("PageCache::ReadDiskPage(): No active index.");
		return NULL;
	}

	// Special case: For this page the offset could be wrong so read it from the
	// beginning of the file.
	if ((HEADPAGETYPE == nType) && (RWNIL == nOffset))
	{
		m_pfm->SeekTo(m_pfm->start());
		m_pfm->Read(nOffset);
	}

	if (RWNIL == nOffset)
	{
		advise("PageCache::ReadDiskPage(): Incorrect offset.");
		return NULL;
	}

	switch(nType)
	{
		case HEADPAGETYPE:
			pPage = new HeadPage;
			break;
		case DIRPAGETYPE:
		case DATAPAGETYPE:
			pPage = new DataPage;
			break;

		default:
	 		advise("PageCache::ReadDiskPage(): Incorrect pagetype.");
			return NULL;
	}

	if (NULL == pPage)
	{
 		advise("PageCache::ReadDiskPage(): Unable to allocate the page space.");
		return NULL;
	}

	pPage->SetPageType(nType);
	pPage->SetOffset(nOffset);
	m_pfm->SeekTo(nOffset);
	pPage->ReadPage(m_pfm, nOffset);

	stats.IncReadPages(pPage->GetPageType());

	return pPage;

}

RWoffset PageCache::GetFirstPageOffset()
{
	RWoffset offset;

	m_pfm->SeekTo(m_pfm->start());
	m_pfm->Read(offset);

	return offset;
}

int PageCache::PutFirstPageOffset(RWoffset offset)
{
	m_pfm->SeekTo(m_pfm->start());
	m_pfm->Write(offset);

	return 0;
}

RecordEntry PageCache::GetPageEntry(RWoffset offset, int nType, int index)
{
	RecordEntry record;
	StoragePage* pPage = GetPage(offset, nType);

	if (NULL == pPage)
	{
		advise("PageCache::GetPageEntry(): Unable to get the page.");
		return record;
	}

	record = pPage->GetEntry(index);
	UnLockPage(offset, nType);

	return record;
}

int PageCache::NumOfLockedPages()
{
	int i, nCount =0;

	if (NULL == m_pfm)
	{
	 	advise("PageCache::CountLockedPages(): No active Index");
		return 1;
	}

	for (i=0; i<MAX_CACHE; i++)
		if (NULL != m_pagecache[i])
		 	if (m_lockpagecount[i])
				nCount++;

	return nCount;
}
