
// datapage.cpp : this keeps track of all the data pages.

#include "rhis.h"
#include "rhtree.h"
#include <stdlib.h>

// Read the different type of data pages.

int StoragePage::GetSize()
{
	return sizeof(m_type);
}

int StoragePage::ClearPage()
{
	return 0;
}

int StoragePage::PrintPage(ostream &os) const
{
	os << "Storage Page: " << m_type << " " << m_offset << endl;
	return 0;
}

// Functions for the head page which is the first page in the file.

int HeadPage::WritePage(RWFileManager *pFM, RWoffset offset)
{
	int i;

	pFM->SeekTo(offset);

	pFM->Write(m_type);
	pFM->Write(m_numpages);
	pFM->Write(m_depth);
	pFM->Write(m_numentries);
	
	for (i=0; i< m_numentries; i++)
		m_rootpage[i].WriteRecord(pFM);

	return 0;
}

int HeadPage::ReadPage(RWFileManager *pFM, RWoffset offset)
{
	int i;
	
	pFM->SeekTo(offset);

	pFM->Read(m_type);
	pFM->Read(m_numpages);
	pFM->Read(m_depth);
	pFM->Read(m_numentries);
	
	for (i=0; i< m_numentries; i++)
		m_rootpage[i].ReadRecord(pFM);

	return 0;
}

int HeadPage::GetSize()
{
	return sizeof(m_type) + sizeof(m_numpages) +
		sizeof(m_depth) + sizeof(m_numentries) + MAX_FIRSTPAGE * m_rootpage[0].GetSize();
}

int HeadPage::FindRootIndex(long lTime)
{
	int index, low, high;
	int found = FALSE;

	low = 0;
	high = m_numentries -1;


	if (high <0)
	{
		advise("HeadPage::FindRootIndex(): No entries in the page.");
		return -1;
	}

	//Find the latest entry
	if (0L == lTime)
		return high;

	if (m_rootpage[low].IsAlive(lTime))
		return low;

	if (m_rootpage[high].IsAlive(lTime))
		return high;
	
	while (low < high)
	{
		index = (low + high) /2;
		if (m_rootpage[index].IsAlive(lTime))
			return index;
		
		if (m_rootpage[index].m_ins > lTime)
			high = index;
		else
			low = index;
	}
	
	advise("HeadPage::FindRootIndex(): Unable to find the entry.");

	return -1;
}

RecordEntry & HeadPage::GetEntry(int index)
{
 	return m_rootpage[index];
}

int HeadPage::AddEntry(const RecordEntry &record)
{
	if (m_numentries >= MAX_FIRSTPAGE)	
	{
		advise("HeadPage::AddEntry() Full page.");
		return 1;
	}
	m_rootpage[m_numentries] = record;
	m_numentries++;
	
	return 0;
}

int HeadPage::ModifyEntry(int index, const RecordEntry &record)
{
	if (index >= m_numentries)	
	{
		advise("HeadPage::ModifyEntry() Out of range.");
		return 1;
	}
	m_rootpage[index] = record;
	
	return 0;
}

int HeadPage::ClearPage()
{
	m_depth = 0;
	m_numpages = 0;
	m_numentries = 0;
	return 0;
}

int HeadPage::PrintPage(ostream &os) const
{
	os << "Head Page: " << m_type << " " << m_offset << endl << m_numpages;
	os << " " << m_depth << " " << m_numentries << endl;
	int i;
	 
	for (i=0; i < m_numentries; i++)
	{
	 	os << i << "th entry: ";
		m_rootpage[i].PrintEntry(os);
	}
	return 0;
}

// Functions for the data pages

int DataPage::WritePage(RWFileManager *pFM, RWoffset offset)
{
	int i;

	pFM->SeekTo(offset);

	pFM->Write(m_type);
	pFM->Write(m_numentries);
	pFM->Write(m_numaliveentries);
	
	for (i=0; i< m_numentries; i++)
		m_entry[i].WriteRecord(pFM);

	return 0;
}

int DataPage::ReadPage(RWFileManager *pFM, RWoffset offset)
{
	int i;
	
	pFM->SeekTo(offset);

	pFM->Read(m_type);
	pFM->Read(m_numentries);
	pFM->Read(m_numaliveentries);
	
	for (i=0; i< m_numentries; i++)
		m_entry[i].ReadRecord(pFM);

	return 0;
}

int DataPage::GetSize()
{
	return sizeof(m_type) + sizeof(m_numentries) +
		sizeof(m_numaliveentries) + (MAX_ENTRYCAPACITY) * m_entry[0].GetSize();
}


int DataPage::VersionSplit(DataPage *pPage, long lTime)
{
	int i;
	if (NULL == pPage)
	{
		advise("DataPage:VersionSplit(): NULL page passed");
		return 1;
	}

	if (pPage->m_type != m_type)
	{
		advise("DataPage:VersionSplit(): Incompatible types.");
		return 1;
	}

	pPage->m_numentries = 0;
	pPage->m_numaliveentries = 0;
	for (i=0; i < m_numentries; i++)
	{
		if (m_entry[i].IsAlive(lTime))
		{
			pPage->AddEntry(m_entry[i]);
			pPage->m_numaliveentries++;
		}
	}

	return 0;
}

int DataPage::AddEntry(const RecordEntry &record)
{
	if (m_numentries >= MAX_ENTRYCAPACITY)	
	{
		advise("DataPage::AddEntry() Full page.");
		return 1;
	}
	m_entry[m_numentries] = record;
	m_numentries++;
	
	return 0;
}

int DataPage::ModifyEntry(int index, const RecordEntry &record)
{
	if (index >= m_numentries)	
	{
		advise("DataPage::ModifyEntry() Out of range.");
		return 1;
	}
	m_entry[index] = record;
	
	return 0;
}

// Updates the number of alive entries at a given time.
int DataPage::UpdateNumAliveEntries(long lTime)
{
	m_numaliveentries = 0;
	for (int i=0; i< m_numentries; i++)
	{
		if (m_entry[i].IsAlive(lTime))
			m_numaliveentries++;
	}
	return 0;
}

// Returns the number of alive entries at a given time.
int DataPage::ComputedNumberOfAliveEntries(long lTime)
{
	int nAliveEntries = 0;
	for (int i=0; i< m_numentries; i++)
	{
		if (m_entry[i].IsAlive(lTime))
			nAliveEntries++;
	}
	return nAliveEntries;
}

int DataPage::ClosestAliveEntry(const RecordEntry &record, long lTime)
{
	int i=0, index=-1;
	double MinAreaInc=0.0;
	double MinArea=0.0;

	for (i=0; i< m_numentries; i++)
	{
		if (m_entry[i].IsAlive(lTime))
		{
			double AreaInc = m_entry[i].AreaIncrease(record);
			double Area = m_entry[i].Area();

			if ((-1 == index) || (AreaInc < MinAreaInc) ||
				((AreaInc == MinAreaInc) && (Area < MinArea)))
			{
				index = i;
				MinAreaInc = AreaInc;
				MinArea = Area;
			}		
		}
	}

	if (-1 == index)
		advise("DataPage::ChooseLeafNode(): Unable to find the minimum entry");

	return index;

}

RecordEntry &DataPage::GetEntry(int index)
{
 	return m_entry[index];
}

RecordEntry DataPage::ComputeCoverEntry(long lTime)
{
	RecordEntry coverRecord;
	int i;

	for (i=0; i < m_numentries; i++)
	{
	 	if (m_entry[i].IsAlive(lTime))
		{
		 	coverRecord.CoverEntry(m_entry[i]);
		}
	}

	return coverRecord;

}

int DataPage::ClearPage()
{
	m_numentries = 0;
	m_numaliveentries = 0;
	return 0;
}

int DataPage::AddAliveEntries(DataPage *pPage, long lTime)
{
	int i;

	for (i=0 ; i < pPage->m_numentries; i++)
	{
	 	if (pPage->GetEntry(i).IsAlive(lTime))
		{
			AddEntry(pPage->GetEntry(i));
			m_numaliveentries++;
		}
	}

	return 0;
}

int DataPage::AddAliveEntriesChangeInsertionTime(DataPage *pPage, long lTime)
{
	int i;

	for (i=0 ; i < pPage->m_numentries; i++)
	{
		RecordEntry record = pPage->GetEntry(i);
	 	if (record.IsAlive(lTime))
		{
			// Set the insertion time to the current time lTime
			record.m_ins = lTime;
			AddEntry(record);
			m_numaliveentries++;
		}
	}

	return 0;
}

int DataPage::FindSibling(const RecordEntry record, int SkipIndex, long lTime)
{
	double fAreaIncrease = 0.0, fAreaOverlap = 0.0;
	double fArea =0.0, fMargin = 0.0;
	int i, retIndex = -1;
	double fAreaParameterMeasure;

	if (RHTree::RandomSiblingAlgorithm == RHTree::m_SiblingAlgorithmType)
	{
		RWGOrderedVector(int) intArray;
		intArray.clear();

		// Insert all the candidates in the integer array one by one and then
		// pick one up randomly
		for (i = 0; i < m_numentries; i++)
		{
			if (SkipIndex == i)
				continue;

	 		if (m_entry[i].IsAlive(lTime))
				intArray.insert(i);
		}
		ASSERT(0 != intArray.entries());
		i = rand() % intArray.entries();
		retIndex = intArray[i];

	} else
	{

		for (i = 0; i < m_numentries; i++)
		{
			if (SkipIndex == i)
				continue;

	 		if (m_entry[i].IsAlive(lTime))
			{
		 		if (-1 == retIndex)
				{
			 		retIndex = i;
					RecordEntry AdjustedEntry = m_entry[i].ReturnAdjustedEntry(record);
					fAreaIncrease = m_entry[i].AreaIncrease(record);
					fAreaOverlap = m_entry[i].OverlapArea(record);
					fArea = m_entry[i].Area();
					fMargin = AdjustedEntry.Margin();
					fAreaParameterMeasure = RHTree::m_fAreaPerimeterTradeoff*fMargin
						+ AdjustedEntry.Area();
				} else
				{
					RecordEntry AdjustedEntry1 = m_entry[i].ReturnAdjustedEntry(record);
					double fAreaIncrease1 = m_entry[i].AreaIncrease(record);
					double fAreaOverlap1 = m_entry[i].OverlapArea(record);
					double fArea1 = m_entry[i].Area();
					double fMargin1 = AdjustedEntry1.Margin();
					double fAreaParameterMeasure1 = RHTree::m_fAreaPerimeterTradeoff*fMargin1
						+ AdjustedEntry1.Area();
					switch(RHTree::m_SiblingAlgorithmType)
					{
					case RHTree::AreaMinimizeAlgorithm:
						{
							if ((fAreaIncrease1 < fAreaIncrease)||
								((fAreaIncrease1 == fAreaIncrease) && (fArea1 < fArea)))
							{
				 				retIndex = i;
							}
						}
						break;
					case RHTree::OverlapMaximizeAlgorithm:
						{
							if ((fAreaOverlap1 > fAreaOverlap)||
								((fAreaOverlap1 == fAreaOverlap) && (fArea1 < fArea)))
							{
				 				retIndex = i;
							}
						}
						break;
					case RHTree::MarginMinimizeAlgorithm:
						{
							if ((fMargin1 < fMargin)||
								((fMargin1 == fMargin) && (fArea1 < fArea)))
							{
				 				retIndex = i;
							}
						}
						break;
					case RHTree::AreaParameterAlgorithm:
						{
							if ((fAreaParameterMeasure1 < fAreaParameterMeasure)||
								((fAreaParameterMeasure1 == fAreaParameterMeasure)
								&& (fArea1 < fArea)))
							{
				 				retIndex = i;
							}
						}
						break;
					default:
						advise("DataPage::FindSibling(): Invalid Sibling finding algorithm");
						return -1;
						break;
					}
					if (i == retIndex)
					{
						fAreaIncrease = fAreaIncrease1;
						fAreaOverlap = fAreaOverlap1;
						fArea = fArea1;
						fMargin = fMargin1;
						fAreaParameterMeasure = fAreaParameterMeasure1;
					}
				}
			}
		}
	}

	if (retIndex < 0)
	{
	 	advise("DataPage::FindSibling(): Unable to find the sibling.");
		return -1;
	}

	return retIndex;
}

int DataPage::PrintPage(ostream &os) const
{
	os << "Data Page: " << m_type << " " << m_offset << endl;
	os << m_numentries << " " << m_numaliveentries << endl;
	int i;
	 
	for (i=0; i < m_numentries; i++)
	{
	 	os << i << "th entry: ";
		m_entry[i].PrintEntry(os);
	}
	return 0;
}
