
// The main file for managing the rtree history.

#include <iostream.h>
#include <stdlib.h>
#include <time.h>
#include <fstream.h>
#include <rw/cstring.h>
#include <rw/rwfile.h>
#include "rhis.h"
#include "rhtree.h"

#define INPUT_FILE	"t.txt"
#define	RH_DATABASE	"t.rht"

//Declaration of the stats variable
CStats stats;
CGeneralStats genStats;

int
main(int argc, char* argv[])
{
	int i;
	CStats insStats, delStats, queryStats;
	CQueryProfileStats queryProfileStats;
	RWCString sDataFile = INPUT_FILE;
	RWCString sDBName = RH_DATABASE;

	// Initialize the random seed
	srand( (unsigned)time( NULL ) );

	for (i=1; i < argc; i++)
	{
		if (!strcmp(argv[i], "-i"))
			sDataFile = argv[++i];
		else if (!strcmp(argv[i], "-d"))
			sDBName = argv[++i];
		else if (!strcmp(argv[i], "-h"))
		{
		 	cout << "Usage: " << argv[0] << " [-i input_data_file]"
		 		<< " [-d database name] [-h] [-debug] [-querydebug] [-deletedebug]"
				<< " [-queryprofile queryblocksize] [-nocache/-cache]"
				<< " [-splitalgotype number] [-siblingalgotype number]"
				<< " [-areaperimetertradeoff number]"
				<< " [-noverma] [-notrasactionintervalquery]" << endl;
			return 0;
		} else if (!strcmp(argv[i], "-querydebug"))
		{
			RHTree::bQueryDebug = TRUE;
		} else if (!strcmp(argv[i], "-deletedebug"))
		{
			RHTree::bDeleteDebug = TRUE;
		} else if (!strcmp(argv[i], "-debug"))
		{
			RHTree::bQueryDebug = TRUE;
			RHTree::bDeleteDebug = TRUE;
		} else if (!strcmp(argv[i], "-nocache"))
		{
			RHTree::bCacheEnabled = FALSE;
		} else if (!strcmp(argv[i], "-cache"))
		{
			RHTree::bCacheEnabled = TRUE;
		} else if (!strcmp(argv[i], "-queryprofile"))
		{
			RHTree::m_bQueryProfileEnabled = TRUE;
			RHTree::m_nQueryProfileBlockSize = atoi(argv[++i]);
		} else if (!strcmp(argv[i], "-splitalgotype"))
		{
			RHTree::m_SplitAlgorithmType = atoi(argv[++i]);
		} else if (!strcmp(argv[i], "-siblingalgotype"))
		{
			RHTree::m_SiblingAlgorithmType = atoi(argv[++i]);
		} else if (!strcmp(argv[i], "-areaperimetertradeoff"))
		{
			RHTree::m_fAreaPerimeterTradeoff = atof(argv[++i]);
		} else if (!strcmp(argv[i], "-noverma"))
		{
			RHTree::m_bVermaUpdatingEnabled = FALSE;
		} else if (!strcmp(argv[i], "-notrasactionintervalquery"))
		{
			RHTree::m_bTransactionIntervalQueryEnabled = FALSE;
		} else
		{
		 	advise("Invalid argument");
			return 1;
		}
	}

	cout << "Starting execution." << endl << "Input file = " << sDataFile << " Database = " <<
		sDBName << endl << " Number of Dimensions = " << NUM_DIM << endl;

	// Check out the RHTree choices
	RHTree::PrintRHTreeChoices(cout);

	RHTree rhtree;
	rhtree.CreateTree(sDBName);

	// input file
	fstream iFile(sDataFile, ios::in);

	if ((iFile.fail()) || (iFile.eof()))
	{
	 	advise("main(): Unable to open the file.");
		return 1;
	}
	RecordEntry record;

	double fUniverse, fTrials, fTmp;
	int nOpType, nRetVal;
	long lCurTime, lOid;

	// This are not used though but are there in the file.
	// iFile >> fUniverse >> fTrials >> fTmp >> fTmp >> fTmp;

	iFile >> lCurTime;

	while (-1 != lCurTime)
	{
		for (i=0; i < NUM_DIM; i++)
		{
			iFile >> record.m_low[i];
			iFile >> record.m_high[i];
		}

		// These are used for the record.
		iFile >> nOpType;
		iFile >> record.m_ins;
		iFile >> record.m_del;
		iFile >> lOid;

		// cout << "Operation: " << nOpType << " Cur Time: " << lCurTime << " ";
		record.PrintEntry(cout);

		switch(nOpType)
		{
			case OP_INSERT:
				stats = insStats;
				record.m_del = -1L;
				nRetVal = rhtree.InsertRecord(record);
				genStats.IncNumInserts();
				if (!rhtree.bCacheEnabled)
					rhtree.FlushCache();
				insStats = stats;
				break;

			case OP_DELETE:
				stats = delStats;
				record.m_ins = lCurTime;
				record.m_del = lCurTime;
				nRetVal = rhtree.DeleteRecord(record);
				genStats.IncNumDeletes();
				if (!rhtree.bCacheEnabled)
					rhtree.FlushCache();
				delStats = stats;
				break;

			case OP_QUERY:
				{
					int bDone = FALSE;
					stats = queryStats;
					if (RHTree::m_bQueryProfileEnabled)
						queryProfileStats.RememberStats(genStats, stats);

					// We want to find all the records which are
					// alive at m_del actually so increment it by one
					// otherwise the records which are inserted at
					// m_del are going to be missed.
					record.m_del++;
					nRetVal = rhtree.QueryRecords(record,
						&(RHTree::QueryDirectoryRecordFunction),
						&(RHTree::QueryDataRecordFunction), TRUE, bDone);
					genStats.IncNumQueries();
					if (!rhtree.bCacheEnabled)
						rhtree.FlushCache();

					if (RHTree::m_bQueryProfileEnabled)
						queryProfileStats.UpdateProfileStats(genStats, stats);
					queryStats = stats;
					break;
				}
		 	default:
				advise("main(): Invalid operation type.");
				return 1;
		}

		if (0 != nRetVal)
		{
		 	advise("main(): Invalid return type from the operation.");
			return 1;
		}


		iFile >> nOpType;
		cout << "Operation  " << nOpType << "  " ;

		if (iFile.eof())
		{
			advise("main(): unexpected EOF for the input file.");
			return 1;
		}
	}
	rhtree.CloseTree();

	cout << endl;
	insStats.PrintStats(cout, "Insert ", genStats.NumberOfInserts());
	cout << endl;
	delStats.PrintStats(cout, "Delete ", genStats.NumberOfDeletes());
	cout << endl;
	insStats = insStats + delStats;
	insStats.PrintStats(cout, "Update ", genStats.NumberOfInserts() + genStats.NumberOfDeletes());
	cout << endl;
	queryStats.PrintStats(cout, "Query ", genStats.NumberOfQueries());
	cout << endl;
	genStats.PrintGeneralStats(cout);
	cout << endl;
	if (RHTree::m_bQueryProfileEnabled)
		queryProfileStats.PrintStats(cout);
	cout << endl;

	cout << "Quiting The program." << endl;

	return 0;
}
