//This file has the definition of the Rtree...with history !!

#include "rhis.h"
#include "rhtree.h"

// Static Variables to store the user options.
int RHTree::bQueryDebug = FALSE;
int RHTree::bDeleteDebug = FALSE;
int RHTree::bCacheEnabled = TRUE;
int RHTree::m_SplitAlgorithmType = RHTree::RStarTreeAlgorithm;
int RHTree::m_SiblingAlgorithmType = RHTree::MarginMinimizeAlgorithm;
double RHTree::m_fAreaPerimeterTradeoff  = 0.0;
int RHTree::m_bQueryProfileEnabled = FALSE;
int RHTree::m_nQueryProfileBlockSize = 0;
int RHTree::m_bVermaUpdatingEnabled = TRUE;
int RHTree::m_bTransactionIntervalQueryEnabled = TRUE;

RHTree::RHTree():m_offset(MAX_PATH),m_index(MAX_PATH)
{
	m_filename = "";
	m_cache = NULL;
	m_offset.clear();
	m_index.clear();
	currentTime = 1L;
	m_ScannedPageTable.clear();
}

RHTree::~RHTree()
{
	//close the file manager.
	if (NULL != m_cache)
		delete m_cache;
}

void RHTree::PrintRHTreeChoices(ostream &os)
{
	os << "Number of Dimensions: " << NUM_DIM << endl;
	os << "Query debugging " << (bQueryDebug ?  "enabled" : "disabled") << endl;
	os << "Deletion debugging " << (bDeleteDebug ?  "enabled" : "disabled") << endl;
	os << "Cache " << (bCacheEnabled ?  "enabled" : "disabled") << endl;
	os << "Verma updating method " << (m_bVermaUpdatingEnabled ?  "enabled" : "disabled") << endl;
	os << "Transaction Interval Query Enabled " <<
		(m_bTransactionIntervalQueryEnabled ?  "enabled" : "disabled") << endl;

	switch(m_SplitAlgorithmType)
	{
		case QuadraticAlgorithm:
			os << "Quadratic";
			break;
		case LinearAlgorithm:
			os << "Linear";
			break;
		case RStarTreeAlgorithm:
			os << "Rstar Tree";
			break;
		default:
			os << "Unknown";
			advise("Unknown splitting algorithm");
			break;
	}
	os << endl;
	os << "Sibling Finding Algorithm Type: ";
	switch(m_SiblingAlgorithmType)
	{
		case AreaMinimizeAlgorithm:
			os << "Minimum Area Increase";
			break;
		case OverlapMaximizeAlgorithm:
			os << "Maximize Overlap";
			break;
		case MarginMinimizeAlgorithm:
			os << "Minimize Margin";
			break;
		case AreaParameterAlgorithm:
			os << "Area/Parameter Tradeoff with value " << m_fAreaPerimeterTradeoff;
			break;
		case RandomSiblingAlgorithm:
			os << "Random Sibling";
			break;
		default:
			os << "Unknown";
			advise("Unknown sibling finding algorithm");
			break;
	}
	os << endl;

	if (m_bQueryProfileEnabled)
		os << "Query profiling enabled with block size "
			<< m_nQueryProfileBlockSize;
	else
		os << "Query profiling disabled";

	os << endl;

	return;
}

int RHTree::CreateTree(const char* szFileName)
{
	int nRetVal = 0;

	if (NULL != m_cache)
	{
	 	advise("RHTree::CreateTree(): already an active index.");
		return 1;
	}	
	m_cache = new PageCache;

	if (NULL == m_cache)
	{
	 	advise("RHTree::CreateTree(): unable to allocate the cache variable.");
		return 1;
	}

	if (0 != (nRetVal = m_cache->CreateIndex(szFileName)))
	{
		advise("RHTree::CreateTree(): Unable to create the index");
	 	return nRetVal;
	}

	RWoffset HeadPageOffset = m_cache->NewPage(HEADPAGETYPE);

	if (RWNIL == HeadPageOffset)
	{
		advise("RHTree::CreateTree(): Unable to create the headpage");
	 	return 1;
	}

	m_cache->PutFirstPageOffset(HeadPageOffset);

	HeadPage *pHead = (HeadPage *) m_cache->GetPage(HeadPageOffset, HEADPAGETYPE);
	if (NULL == pHead)
	{
	 	advise("RHTree::CreateTree(): unable to get the head page.");
		return 1;
	}

	RWoffset RootPageOffset = m_cache->NewPage(DATAPAGETYPE);
	if (RWNIL == RootPageOffset)
	{
		advise("RHTree::CreateTree(): Unable to create the rootpage");
	 	return 1;
	}

	RecordEntry record;

	record.m_info = RootPageOffset;
	record.m_ins = 0L;
	record.m_del = -1L;

	pHead->AddEntry(record);

	m_cache->PutPage(HeadPageOffset, HEADPAGETYPE);

	return 0;
}

int RHTree::CloseTree()
{
	int nRetVal = 0;

	if (NULL != m_cache)
		nRetVal = m_cache->CloseIndex();

	delete m_cache;
	m_cache = NULL;

	return nRetVal;
}

int RHTree::RemoveTree()
{
	if (NULL != m_cache)
	{
		m_cache->RemoveIndex();
	}

	return 0;
}

int RHTree::OpenTree(const char *szFileName)
{
	int nRetVal = 0;

	if (NULL != m_cache)
	{
	 	advise("RHTree::OpenTree(): already an active index.");
		return 1;
	}	
	m_cache = new PageCache;

	if (0 != (nRetVal = m_cache->OpenIndex(szFileName)))
	{
		advise("RHTree::OpenTree(): Unable to open the index");
	 	return nRetVal;
	} 

	return 0;
}

// Insert records
int RHTree::InsertRecord(RecordEntry &record)
{
	RWoffset leafOffset = ChooseLeafNode(record, record.m_ins);
	
	if (RWNIL == leafOffset)
	{
		advise("RHTree::InsertRecord(): Null Offset.");
		return 1;
	}

	if (0!= UpdateNode(leafOffset, DATAPAGETYPE, 0, NULL, NULL, 1, &record, 1, record.m_ins))
	{
		advise("RHTree::InsertRecord(): Unable to update the node.");
		return 1;
	}

	m_depth = m_offset.entries();
	return ClearPath();
}

// Delete records
// Make sure that for the input record the fields other than the transaction
// time field are set correctly and the transaction time (both m_ins and m_del)
// are set at currenttime
int RHTree::DeleteRecord(RecordEntry &record)
{
	int index, bDone = FALSE;
	
	if ((0 != QueryRecords(record, &(RHTree::DeleteQueryDirectoryRecordFunction),
		&(RHTree::DeleteQueryDataRecordFunction), FALSE, bDone)) || !bDone)
	{
		advise("RHTree::DeleteRecord(): Unable to find the record to be deleted.");
		record.PrintEntry(cerr);
		return 1;
	}

	index = m_index[m_depth -1];
	RecordEntry QueriedRecord = m_cache->GetPageEntry(m_offset[m_depth-1], DATAPAGETYPE, index);
	QueriedRecord.m_del = record.m_ins;

	if (0!= UpdateNode(m_offset[m_depth-1], DATAPAGETYPE, 1, &index, &QueriedRecord, 0, NULL,
			-1, record.m_del))
	{
		advise("RHTree::DeleteRecord(): Unable to update the node.");
		return 1;
	}
	// restore the m_depth count.
	m_depth = m_offset.entries();
	return ClearPath();
}

//The record has all the information regarding the overlaps etc.
int RHTree::QueryRecords(RecordEntry &record, QUERYFNPTR pDirFunction,
		QUERYFNPTR pDataFunction, int bCleanup, int &bDone)
{
	int index, nStartIndex;
	HeadPage *pHead;
	RWoffset offset;

	m_offset.clear();
	m_index.clear();
	m_depth = 0;
	if (m_bTransactionIntervalQueryEnabled)
		m_ScannedPageTable.clearAndDestroy();

	offset = m_cache->GetFirstPageOffset();

	pHead = (HeadPage *) m_cache->GetPage(offset, HEADPAGETYPE);
	

	if (NULL == pHead)
	{
	 	advise("RHTree::QueryRecords(): NULL head page.");
		return RWNIL;
	}


	// Should be modified to support a range of time.
	nStartIndex = pHead->FindRootIndex(record.m_ins);

	// Unlock it here since lock it again anyway.
	m_cache->UnLockPage(offset, HEADPAGETYPE);

	if (nStartIndex < 0)
	{
	 	advise("RHTree::QueryRecords(): Negative Index.");
		return 1;
	}

	ASSERT(0 == m_cache->NumOfLockedPages());
	// We need to look at all the entries in the time range
	for (index = nStartIndex; index < pHead->m_numentries; index++)
	{
		if (0 != m_depth)
		{
	 		advise("RHTree::QueryRecords(): There is something in the path array.");
			return 1;
		}

		offset = m_cache->GetFirstPageOffset();
		pHead = (HeadPage *) m_cache->GetPage(offset, HEADPAGETYPE);

		if (NULL == pHead)
		{
	 		advise("RHTree::QueryRecords(): NULL head page.");
			return RWNIL;
		}

		m_offset.insert(offset);		// Of the head page...
		m_index.insert(index);			// index in the head page !!
		m_depth++;

		if (!pHead->GetEntry(index).OverlapsAndIsAlive(record))
		  break;

		offset = pHead->GetEntry(index).m_info;
		if (RWNIL == offset)
		{
	 		advise("RHTree::QueryRecords(): NULL offset.");
			return 1;
		}

		if (0 != ScanNode(offset, record, pDirFunction, pDataFunction, bDone))
		{
	 		advise("RHTree::QueryRecords(): Unable to do the scanning of the node.");
			return 1;
		}


		if (bDone)
			break;

		// cleanup for the next run.
		if (bCleanup)
		{
			ClearPath();
			ASSERT(0 == m_cache->NumOfLockedPages());
		}
	}

	if (bCleanup)
	{
		ClearPath();
		ASSERT(0 == m_cache->NumOfLockedPages());
	}

	if (m_bTransactionIntervalQueryEnabled)
		m_ScannedPageTable.clearAndDestroy();
	return 0;
}

// Used while querying.
int RHTree::ScanNode(RWoffset offset, RecordEntry &record,
	QUERYFNPTR pDirFunction, QUERYFNPTR pDataFunction, int &bDone)
{
	int i, nRetVal, nCurDepth;
	RWCollectableInt KeyCollectable(offset);

	if (m_bTransactionIntervalQueryEnabled)
	{
		// See if already scanned in the current run then don't need to do it again
		if (m_ScannedPageTable.contains(&KeyCollectable))
			return 0;

		// If not add it now.
		m_ScannedPageTable.insert(new RWCollectableInt(offset));
	}
	DataPage *pPage =  (DataPage *) m_cache->GetPage(offset, DIRPAGETYPE);

	if (NULL == pPage)
	{
	 	advise("RHTree::ScanNode(): Unable to retrieve the page");
		return 1;
	}
	
	m_offset.insert(offset);
	m_index.insert(-1);				// Insert some invalid index.
	nCurDepth = m_depth;			// remember the current depth to insert the index
	m_depth++;
	int nPageType = pPage->GetPageType();
	

	if ((DIRPAGETYPE != nPageType) && (DATAPAGETYPE != nPageType))
	{
	 	advise("RHTree::ScanNode(): invalid page type.");
		return 1;
	}

	for (i=0; i < pPage->m_numentries; i++)
	{
		int bOverlap = FALSE;
		if (DIRPAGETYPE == nPageType)
			nRetVal = (*pDirFunction)(record, pPage->GetEntry(i), bOverlap, bDone);
		else
			nRetVal = (*pDataFunction)(record, pPage->GetEntry(i), bOverlap, bDone);

		if (0 != nRetVal)
		{
		 	advise("RHTree::ScanNode(): Bad return code.");
			return 1;
		}

		if (bDone)
		{
			m_index[nCurDepth] = i;
			return 0;
		}

		// Do a recursive scan of the node
		if (bOverlap && (DIRPAGETYPE == nPageType))
		{
		 	 if (0 != ScanNode(pPage->GetEntry(i).m_info, record, pDirFunction, pDataFunction, bDone))
			 	return 1;

			if (bDone)
			{
			 	m_index[nCurDepth] = i;
				return 0;
			}
		}
	}

	m_cache->UnLockPage(offset, nPageType);

	m_offset.removeAt(m_depth-1);
	m_index.removeAt(m_depth-1);
	m_depth--;

	return 0;
}

//Clears the pages in the cache by unlocking them.
int RHTree::ClearPath()
{
	int i;

	if ((m_depth != (int) (m_offset.entries())) ||
		(m_offset.entries() != m_index.entries()))
	{
	 	advise("RHTree::ClearPath(): Unequal depth and path lengths.");
		return 1;
	}
	for (i=0; i < m_depth; i++)
	{
		m_cache->UnLockPage(m_offset[i], (0==i) ? HEADPAGETYPE : DIRPAGETYPE);
	}

	m_offset.clear();
	m_index.clear();
	m_depth=0;
	return 0;
}

// This function does the insert of the maximum of two entries and reroutes/deletes two of the entries
// Also need to specify the index for the entries to be deleted or rerouted.
// The netchange variable specifies the number of entries lost or gained. This is useful
// to calculate the overflow or underflow.
int RHTree::UpdateNode(RWoffset offset, int nType, int nNumModifiedRecords, int indexArray[],
	RecordEntry recordArray[], int nNumNewRecords, RecordEntry newRecordArray[], int nNetChange,
	long lTime)
{

	int i;
	
	switch(nType)
	{
		case HEADPAGETYPE:
		case DIRPAGETYPE:
		case DATAPAGETYPE:
			break;
		default:
			advise("Unknown page type");
			return 1;
	}
		
	// This is just to handle the top page..
	if (HEADPAGETYPE == nType)
	{
		HeadPage *pHead = (HeadPage *) m_cache->GetPage(offset, nType);
		for (i=0; i < nNumModifiedRecords; i++)
		{
		 	if (0 != pHead->ModifyEntry(indexArray[i], recordArray[i]))
			{
			 	advise("RHTree::UpdateNode(): Unable to modify the entry in the head page.");
				return 1;
			}
		}

		// if the number of new records is two it means that the root page has split.
		if (nNumNewRecords < 2)
		{
			for (i=0; i < nNumNewRecords; i++)
			{
			 	if (0 != pHead->AddEntry(newRecordArray[i]))
				{
				 	advise("RHTree::UpdateNode(): Unable to add the entry in the head page.");
					return 1;
				}
			}
		} else // This case will create a new root page since the previous one split.
		{
			RWoffset off_newpage = m_cache->NewPage(DIRPAGETYPE);
			if (RWNIL == off_newpage)
			{
				advise("RHTree::UpdateNode(): Null New Page Offset root.");
				return -1;
			}

			// Get the new page.
			DataPage *pNewPage = (DataPage *) m_cache->GetPage(off_newpage, DIRPAGETYPE);
			RecordEntry cover;

			if (NULL == pNewPage)
			{
				advise("RHTree::UpdateNode(): Null Page root.");
				return 1;
			}
			for (i=0; i < nNumNewRecords; i++)
			{
			 	if (0 != pNewPage->AddEntry(newRecordArray[i]))
				{
				 	advise("RHTree::UpdateNode(): Unable to add the entry in the new root page.");
					return 1;
				}
				pNewPage->m_numaliveentries++;
				cover.CoverEntry(newRecordArray[i]);
			}
			
			ASSERT1(pNewPage->m_numaliveentries == pNewPage->ComputedNumberOfAliveEntries(lTime));
			m_cache->PutPage(off_newpage, DIRPAGETYPE);

			cover.m_ins = lTime;
			cover.m_del = -1L;
			cover.m_info = off_newpage;

			pHead->AddEntry(cover);
		}
		return m_cache->PutPage(offset, HEADPAGETYPE);
	}
	
	int bRootPage = IsRootPage(offset);
	// Now only type to be handled is DIRPAGETYPE or DATAPAGETYPE

	RecordEntry origRecord, modRecord;		// Has the record for the page in question.

	if (bRootPage)
		origRecord = m_cache->GetPageEntry(m_offset[m_depth-2], HEADPAGETYPE, m_index[m_depth-2]);
	else
		origRecord = m_cache->GetPageEntry(m_offset[m_depth-2], DIRPAGETYPE, m_index[m_depth-2]);

	modRecord = origRecord;

	DataPage *pPage = (DataPage *) m_cache->GetPage(offset, nType);
	ASSERT1(pPage->m_numaliveentries == pPage->ComputedNumberOfAliveEntries(lTime));

	for (i=0; i < nNumModifiedRecords; i++)
	{
	 	if (0 != pPage->ModifyEntry(indexArray[i], recordArray[i]))
		{
		 	advise("RHTree::UpdateNode(): Unable to modify the entry in the head page.");
			return 1;
		}
		if (recordArray[i].IsAlive(lTime))
			modRecord.CoverEntry(recordArray[i]);
	}

	for (i=0; i < nNumNewRecords; i++)
	{
	 	if (0 != pPage->AddEntry(newRecordArray[i]))
		{
		 	advise("RHTree::UpdateNode(): Unable to add the entry in the head page.");
			return 1;
		}
		modRecord.CoverEntry(newRecordArray[i]);
	}
	
	// Adjust the number of alive entries.
	pPage->m_numaliveentries += nNetChange;
	m_depth--;	// Change the depth.


	// See if the root page needs to be collapsed
	if (bRootPage && (DIRPAGETYPE == nType ) && (1 == pPage->m_numaliveentries))
	{
		ASSERT1(pPage->m_numaliveentries == pPage->ComputedNumberOfAliveEntries(lTime));
	 	// Back up the count..Because the last entries must be the new ones....
	 	pPage->m_numentries -= nNumNewRecords;

	 	m_cache->PutPage(offset, nType);
		modRecord.m_del = lTime;

		// This must have happened if the the net change is to reduce the number of entries.
		if ((nNetChange != -1) || (1 != nNumNewRecords ))
		{
		 	advise("RHTree::UpdateNode(): Incorrect net change count or new records.");
			return 1;
		}

		return UpdateNode(m_offset[m_depth -1], bRootPage ? HEADPAGETYPE : DIRPAGETYPE,
			1, &(m_index[m_depth-1]), &modRecord, nNumNewRecords, newRecordArray, 0, lTime);
	}

	// See if the page is ok after all the changes..
	if (((pPage->m_numaliveentries >= MIN_WVENTRY) || bRootPage) && (pPage->m_numentries <= MAX_ENTRY))
	{
		ASSERT1(pPage->m_numaliveentries == pPage->ComputedNumberOfAliveEntries(lTime));

		m_cache->PutPage(offset, nType);
		// If no change in the record.
		if (origRecord == modRecord)
		 	return 0;

		return UpdateNode(m_offset[m_depth -1], bRootPage ? HEADPAGETYPE : DIRPAGETYPE,
			1, &(m_index[m_depth-1]), &modRecord, 0, NULL, 0, lTime);
	}

	// Delete the current record
	modRecord.m_del = lTime;
	
	// Block overflow or weak version underflow.
	if ((pPage->m_numentries > MAX_ENTRY) || (pPage->m_numaliveentries < MIN_WVENTRY))
	{
		// Some of the local variable for managing the new updates.
		int l_indexArray[2]; 
		RecordEntry l_recordArray[2], l_newRecordArray[2];
		int bNodeHandled = FALSE;

		// Upon underflow in the node Varman's special case of sibling < MID_ENTRY and total entries
		// in the sibling less than MAX_WVENTRY +1
		if ((m_bVermaUpdatingEnabled) && (!bRootPage) && (pPage->m_numaliveentries < MIN_WVENTRY))
		{
			// If there is strong version underflow and it is not a rootpage
			DataPage *pParent = (DataPage *) m_cache->GetPage(m_offset[m_depth-1], DIRPAGETYPE);
			if (NULL == pParent)
			{
				advise("RHTree::UpdateNode(): Unable to get the parent (WV underflow).");
				return 1;
			}

			RecordEntry coverNode = pPage->ComputeCoverEntry(lTime);

			// Find the sibling and also give an index to skip while searching.
			int nSibIndex = pParent->FindSibling(coverNode, m_index[m_depth-1], lTime);
			if (nSibIndex < 0)
			{
				advise("RHTree::UpdateNode(): Unable to get the sibling (WV underflow).");
				return 1;
			}

			// Get the sibling entry.
			RecordEntry sibEntry = pParent->GetEntry(nSibIndex);
			m_cache->UnLockPage(m_offset[m_depth-1], DIRPAGETYPE);

			DataPage* pSibPage = (DataPage *) m_cache->GetPage(sibEntry.m_info, nType);
			if (NULL == pSibPage)
			{
				advise("RHTree::UpdateNode(): Unable to get the sibling (WV underflow).");
				return 1;
			}

			ASSERT(pSibPage->m_numaliveentries >= MIN_WVENTRY);

			if ((pSibPage->m_numaliveentries < MID_ENTRY) && (pSibPage->m_numentries <= MAX_SVENTRY))
			{
				bNodeHandled = TRUE;
				
				// Now add the alive entries from old node to the sibling node and we are done.
				// Note that the insertion time of the entries should be changed here
				// so that the entries are not obtained in duplicate during searches
				// since the old node and sibling node are alive for some time
				// together and the entries might cause confusion. 
				// This is as per updating method from Verma'a paper
				if (0 != pSibPage->AddAliveEntriesChangeInsertionTime(pPage, lTime))
				{
					advise("RHTree::UpdateNode(): Unable to copy the entries (WV underflow).");
					return 1;
				}
				
				sibEntry.m_del = lTime;
				// Set the records to be modified
				// Note the old sibling entry is deleted but a new one is put.
				l_recordArray[0] = modRecord;
				l_recordArray[1] = sibEntry;
				l_indexArray[0] = m_index[m_depth -1];
				l_indexArray[1] = nSibIndex;

				// Make the new entries.
				l_newRecordArray[0] = pSibPage->ComputeCoverEntry(lTime);
				l_newRecordArray[0].m_info = pSibPage->GetOffset();
				l_newRecordArray[0].m_ins = lTime;
				l_newRecordArray[0].m_del = -1;

				ASSERT(pSibPage->m_numaliveentries >= MIN_WVENTRY);
				ASSERT1(pSibPage->m_numaliveentries == pSibPage->ComputedNumberOfAliveEntries(lTime));

				// Put the sibling page after modifying it since unlocked again later on.
				m_cache->PutPage(sibEntry.m_info, nType);
				// Backup the entries since the new ones are not valid...
				pPage->m_numentries -= nNumNewRecords;
				m_cache->PutPage(offset, nType);

				if (0 != UpdateNode(m_offset[m_depth -1], DIRPAGETYPE, 2, l_indexArray,
					l_recordArray, 1, l_newRecordArray, -1, lTime))
				{
					advise("RHTree::UpdateNode(): Unable to update the node (Varman's WV underflow).");
					return 1;
				}
				return 0;
			}

			m_cache->UnLockPage(sibEntry.m_info, nType);

		}

		// Prepare for a version split.
		RWoffset off_newpage1 = m_cache->NewPage(pPage->GetPageType());
		if (RWNIL == off_newpage1)
		{
			advise("RHTree::UpdateNode(): Null New page Offset.");
			return -1;
		}

		// Get the new page.
		DataPage *pNewPage1 = (DataPage *) m_cache->GetPage(off_newpage1, pPage->GetPageType());
		if (NULL == pNewPage1)
		{
			advise("RHTree::UpdateNode(): Null Page.");
			return 1;
		}

		if (0 != pPage->VersionSplit(pNewPage1, lTime))
		{
			advise("RHTree::UpdateNode(): Unable to split the page");
			return 1;
		}

		if (pNewPage1->m_numentries > MAX_SVENTRY)
		{
		// Condition for strong version overflow
			RWoffset off_newpage2 = m_cache->NewPage(pPage->GetPageType());
			if (RWNIL == off_newpage2)
			{
				advise("RHTree::UpdateNode(): Null New page 2 Offset.");
				return -1;
			}

			// Get the new page.
			DataPage *pNewPage2 = (DataPage *) m_cache->GetPage(off_newpage2, pPage->GetPageType());
			if (NULL == pNewPage2)
			{
				advise("RHTree::UpdateNode(): Null Page 2.");
				return 1;
			}

			if (0 != SplitPage(pNewPage1, NULL, pNewPage1, l_newRecordArray[0], pNewPage2,
				l_newRecordArray[1], lTime))
			{
				advise("RHTree::UpdateNode(): Unable to split the page (SV overflow).");
				return 1;
			}

			if (0 != UpdateNode(m_offset[m_depth -1], bRootPage ? HEADPAGETYPE : DIRPAGETYPE,
				1, &(m_index[m_depth-1]), &modRecord, 2, l_newRecordArray, 1, lTime))
			{
				advise("RHTree::UpdateNode(): Unable to update the node (SV overflow).");
				return 1;
			}

			ASSERT(pNewPage2->m_numaliveentries >= MIN_SVENTRY);
			ASSERT1(pNewPage2->m_numaliveentries == pNewPage2->ComputedNumberOfAliveEntries(lTime));

			m_cache->PutPage(off_newpage2, nType);

		} else if ((pNewPage1->m_numentries < MIN_SVENTRY) && (!bRootPage))
		{
			// If there is strong version underflow and it is not a rootpage
			DataPage *pParent = (DataPage *) m_cache->GetPage(m_offset[m_depth-1], DIRPAGETYPE);
			if (NULL == pParent)
			{
				advise("RHTree::UpdateNode(): Unable to get the parent (SV underflow).");
				return 1;
			}

 			RecordEntry newRecord = pNewPage1->ComputeCoverEntry(lTime);

			// Find the sibling and also give an index to skip while searching.
			int nSibIndex = pParent->FindSibling(newRecord, m_index[m_depth-1], lTime);
			if (nSibIndex < 0)
			{
				advise("RHTree::UpdateNode(): Unable to get the sibling (SV underflow).");
				return 1;
			}

			RecordEntry sibEntry = pParent->GetEntry(nSibIndex);
			m_cache->UnLockPage(m_offset[m_depth-1], DIRPAGETYPE);

			DataPage* pSibPage = (DataPage *) m_cache->GetPage(sibEntry.m_info, nType);
			if (NULL == pSibPage)
			{
				advise("RHTree::UpdateNode(): Unable to get the sibling (SV underflow).");
				return 1;
			}
			sibEntry.m_del = lTime;

			// Special case from Verma and Varman to handle the sibling > MID_ENTRY case
			if (m_bVermaUpdatingEnabled && (pSibPage->m_numaliveentries >= MID_ENTRY))
			{
				if (0 != CopyEntries(pSibPage, pNewPage1,
					MIN_SVENTRY - pNewPage1->m_numaliveentries, lTime))
				{
					advise("RHTree::UpdateNode(): Unable to copy entries from the sibling.");
					return 1;
				}
				// Lock and Put the sibling page since unlocked again in the end.
				m_cache->LockPage(sibEntry.m_info, nType);
				m_cache->PutPage(sibEntry.m_info, nType);
				
				// Set the records to be modified
				// Note the old sibling entry is deleted but a new one is put.
				l_recordArray[0] = modRecord;
				l_recordArray[1] = sibEntry;
				l_indexArray[0] = m_index[m_depth -1];
				l_indexArray[1] = nSibIndex;

				// Make the new entries.
				l_newRecordArray[0] = pNewPage1->ComputeCoverEntry(lTime);
				l_newRecordArray[0].m_info = pNewPage1->GetOffset();;
				l_newRecordArray[1] = pSibPage->ComputeCoverEntry(lTime);
				l_newRecordArray[1].m_info = pSibPage->GetOffset();
				l_newRecordArray[0].m_ins = l_newRecordArray[1].m_ins = lTime;
				l_newRecordArray[0].m_del = l_newRecordArray[1].m_del = -1;

				ASSERT(pSibPage->m_numaliveentries >= MIN_WVENTRY);
				ASSERT(pNewPage1->m_numaliveentries >= MIN_WVENTRY);

				ASSERT1(pSibPage->m_numaliveentries == pSibPage->ComputedNumberOfAliveEntries(lTime));

				if (0 != UpdateNode(m_offset[m_depth -1], DIRPAGETYPE, 2, l_indexArray,
					l_recordArray, 2, l_newRecordArray, 0, lTime))
				{
					advise("RHTree::UpdateNode(): Unable to update the node (Varman's SV underflow).");
					return 1;
				}
			}
			// See if on merging the number of entries is too high.
			else if (pNewPage1->m_numaliveentries + pSibPage->m_numaliveentries > MAX_SVENTRY)
			{
				RWoffset off_newpage2 = m_cache->NewPage(nType);
				if (RWNIL == off_newpage2)
				{
					advise("RHTree::UpdateNode(): Null New page 2 Offset (SV underflow).");
					return 1;
				}

				// Get the new page.
				DataPage *pNewPage2 = (DataPage *) m_cache->GetPage(off_newpage2, nType);
				if (NULL == pNewPage2)
				{
					advise("RHTree::UpdateNode(): Null Page 2 (SV underflow).");
					return 1;
				}

				if (0 !=SplitPage(pNewPage1, pSibPage, pNewPage1, l_newRecordArray[0], pNewPage2,
					l_newRecordArray[1], lTime))
				{
					advise("RHTree::UpdateNode(): Unable to split the page (SV underflow).");
					return 1;
				}

				l_recordArray[0] = modRecord;
				l_recordArray[1] = sibEntry;
				l_indexArray[0] = m_index[m_depth -1];
				l_indexArray[1] = nSibIndex;

				ASSERT(pNewPage1->m_numaliveentries >= MIN_SVENTRY);
				ASSERT(pNewPage2->m_numaliveentries >= MIN_SVENTRY);
				ASSERT1(pNewPage2->m_numaliveentries == pNewPage2->ComputedNumberOfAliveEntries(lTime));

				if (0 != UpdateNode(m_offset[m_depth -1], DIRPAGETYPE, 2, l_indexArray,
					l_recordArray, 2, l_newRecordArray, 0, lTime))
				{
					advise("RHTree::UpdateNode(): Unable to update the node (SV underflow).");
					return 1;
				}

				m_cache->PutPage(off_newpage2, nType);
			} else
			{
				pNewPage1->AddAliveEntries(pSibPage, lTime);
				l_newRecordArray[0] = pNewPage1->ComputeCoverEntry(lTime);
				l_newRecordArray[0].m_ins = lTime;
				l_newRecordArray[0].m_del = -1;
				l_newRecordArray[0].m_info = off_newpage1;

				l_recordArray[0] = modRecord;
				l_recordArray[1] = sibEntry;
				l_indexArray[0] = m_index[m_depth -1];
				l_indexArray[1] = nSibIndex;
				ASSERT(pNewPage1->m_numaliveentries >= MIN_SVENTRY);

				if (0 != UpdateNode(m_offset[m_depth -1], DIRPAGETYPE,
					2, l_indexArray, l_recordArray, 1, l_newRecordArray, -1, lTime))
				{
					advise("RHTree::UpdateNode(): Unable to update the node (ok).");
					return 1;
				}

			}
 			m_cache->UnLockPage(sibEntry.m_info, nType);

		} else
		{
			// If everything is ok.
			l_newRecordArray[0] = pNewPage1->ComputeCoverEntry(lTime);
			l_newRecordArray[0].m_ins = lTime;
			l_newRecordArray[0].m_del = -1;
			l_newRecordArray[0].m_info = off_newpage1;
			if (0 != UpdateNode(m_offset[m_depth -1], bRootPage ? HEADPAGETYPE : DIRPAGETYPE,
				1, &(m_index[m_depth-1]), &modRecord, 1, l_newRecordArray, 0, lTime))
			{
				advise("RHTree::UpdateNode(): Unable to update the node (ok).");
				return 1;
			}
		}

		// Backup the entries since the new ones are not valid...
		pPage->m_numentries -= nNumNewRecords;

		ASSERT(pNewPage1);
		ASSERT(bRootPage || (pNewPage1->m_numaliveentries >= MIN_SVENTRY));
		ASSERT1(pNewPage1->m_numaliveentries == pNewPage1->ComputedNumberOfAliveEntries(lTime));

		m_cache->PutPage(off_newpage1, nType);
		m_cache->PutPage(offset, nType);
	} else
	{
	 	advise("RHTree::UpdateNode(): Error in checking the page entries.");
		return 1;
	}
	return 0;
}

// To get some number of entries out of a page into another page.
// This is for the case when the sibling has more than MID_ENTRY number of
// entries and hence can pick up stuff from there.
// This is the special case coming out of Varman and Verma paper.
int RHTree::CopyEntries(DataPage *pSourcePage, DataPage *pTargetPage,
	int nNumEntriesToCopy, long lTime)
{
	int i;
	RWGSortedVector(SortStruct) sortedVec(CompareSortStruct);
	RecordEntry coverRecord = pSourcePage->ComputeCoverEntry(lTime);

	for (i = 0; i < pSourcePage->m_numentries; i++)
	{
		RecordEntry record = pSourcePage->GetEntry(i);

		// Copy the alive records
		if (!record.IsAlive(lTime))
			continue;

		// Add the record first
		SortStruct s;

		// Idea is to find the list area increase
		s.fValue = coverRecord.AreaIncrease(record);
		s.nIndex = i;
		
		// Sort according to the area and insert
		sortedVec.insert(s);

	}

	if ((int)sortedVec.entries() < nNumEntriesToCopy)
	{
		advise("RHTree::CopyEntries(): Not enough entries to copy.");
		return 1;
	}

	// Pick up the closest entries to the target page.
	for (i=0 ; i < nNumEntriesToCopy; i++)
	{
		SortStruct s = sortedVec(i);
		RecordEntry record = pSourcePage->GetEntry(s.nIndex);

		if (0 != pTargetPage->AddEntry(record))
		{
			advise("RHTree::CopyEntries(): Error in adding the entry.");
			return 1;
		}
		
		// Delete it from the source page.
		record.m_del = lTime;
		if (0 != pSourcePage->ModifyEntry(s.nIndex, record))
		{
			advise("RHTree::CopyEntries(): Error in modifying the entry.");
			return 1;
		}
		
		// Time to adjust the number of alive entries.
		pSourcePage->m_numaliveentries--;
		pTargetPage->m_numaliveentries++;
	}

	return 0;
}

// This splits the page of a certain type when it has too many entries..
// It assumes that the current path in the m_offset includes this tree
// This fact is used for updating the entries before it.
// 
// Note: The pNewPage1 or pNewPage2 could be one of the input pages pPage1 or pPage2 too..
// Cover returns the new records for the created pages..
int RHTree::SplitPage(DataPage *pPage1, DataPage *pPage2,
	DataPage *pNewPage1, RecordEntry &cover1, DataPage *pNewPage2, RecordEntry &cover2,
	long lTime)
{
	if (RHTree::RStarTreeAlgorithm == m_SplitAlgorithmType)
		return RStarTreeSplitPage(pPage1, pPage2, pNewPage1, cover1, pNewPage2, cover2, lTime);
	else
		return RTreeSplitPage(pPage1, pPage2, pNewPage1, cover1, pNewPage2, cover2, lTime);
}

int RHTree::RStarTreeSplitPage(DataPage *pPage1, DataPage *pPage2,
	DataPage *pNewPage1, RecordEntry &cover1, DataPage *pNewPage2, RecordEntry &cover2,
	long lTime)
{
	RWGOrderedVector(RecordEntry) listEntry(2*MAX_ENTRYCAPACITY);
	int i;

	if ((NULL == pPage1) || (NULL == pNewPage1) || (NULL == pNewPage2))
	{
	 	advise("RHTree::RStarTreeSplitPage(): Null new data pages.");
		return 1;
	}

	for (i=0; i < pPage1->m_numentries; i++)
	{
		RecordEntry record = pPage1->GetEntry(i);
		if (record.IsAlive(lTime))
			listEntry.insert(record);
	}

	if (NULL != pPage2)
	{
		for (i=0; i < pPage2->m_numentries; i++)
		{
			RecordEntry record = pPage2->GetEntry(i);
			if (record.IsAlive(lTime))
				listEntry.insert(record);
		}
	}

	// If too many or too few entries.
	if ((listEntry.entries() <= MAX_SVENTRY) || (listEntry.entries() > 2*MAX_SVENTRY))
	{
	 	advise("RHTree::RStarTreeSplitPage(): Too few or too many entries to split.");
		return 1;
	}

	int nAxis;
	
	// Determine the axis to split by first.
	if (0 != ChooseAxisToSplit(listEntry, nAxis))
	{
	 	advise("RHTree::RStarTreeSplitPage(): Unable to choose the axis.");
		return 1;
	}
	ASSERT ((nAxis >=0) && (nAxis < NUM_DIM));

	// Now time to do the actual split.
	if (0 != SortEntryByAxis(listEntry, nAxis))
	{
	 	advise("RHTree::RStarTreeSplitPage(): Unable to sort entry by axis.");
		return 1;
	}

	size_t nNumFirstSplit;
	if (0 != ChooseSplitNumForGivenAxis(listEntry, nAxis, nNumFirstSplit))
	{
	 	advise("RHTree::RStarTreeSplitPage(): Unable to sort entry by axis.");
		return 1;
	}
	
	// Note that the number in the first split should be atleast MIN_SVENTRY
	// && less than so that the second split also has atleast that much.
	// This is to satisfy the basic properties for BRT.
	ASSERT((nNumFirstSplit >= MIN_SVENTRY) && 
		(nNumFirstSplit <= listEntry.entries() - MIN_SVENTRY));

	// Now based on the split index...add the initial entries to page1 and 
	// rest to page2

	cover1.ClearEntry();
	cover2.ClearEntry();
	pNewPage1->ClearPage();
	pNewPage2->ClearPage();
	cover1.m_ins = lTime;
	cover1.m_del = -1;
	cover2.m_ins = lTime;
	cover2.m_del = -1;
	cover1.m_info = pNewPage1->GetOffset();
	cover2.m_info = pNewPage2->GetOffset();

	size_t j;
	for (j=0; j < nNumFirstSplit; j++)
	{
		pNewPage1->AddEntry(listEntry[j]);
		pNewPage1->m_numaliveentries++;
		cover1.CoverEntry(listEntry[j]);
	}

	for (; j < listEntry.entries(); j++)
	{
		pNewPage2->AddEntry(listEntry[j]);
		pNewPage2->m_numaliveentries++;
		cover2.CoverEntry(listEntry[j]);
	}

	listEntry.clear();
	return 0;
}

// The given entries are already sorted for the given axis and now just
// get the number of entries needed to be put in the first and second splits
int RHTree::ChooseSplitNumForGivenAxis(RWGOrderedVector(RecordEntry) &listEntry,
		int nAxis, size_t &nNumFirstSplit)
{
	ASSERT((nAxis >= 0) && (nAxis < NUM_DIM));

	size_t j;
	double fOverlapArea, fMinOverlapArea, fTotalArea, fMinTotalArea;

	// Check if the number of entries are more than a certain minimum.
	ASSERT(listEntry.entries() >= 2* MIN_SVENTRY);

	// Find the distribution with the least overlap area value between
	// the bounding boxes. Resolve ties by lesser total area.
	for (j= MIN_SVENTRY ; j <= (listEntry.entries() - MIN_SVENTRY); j++)
	{
		// Compute the margin value of the bounding boxes for the split at
		// with the first split containing j entries.
		fOverlapArea =
			ComputeGoodnessValue(listEntry, j, RHTree::OverlapValueType, fTotalArea);
		
		if ((MIN_SVENTRY == j) || (fOverlapArea < fMinOverlapArea)
			|| ((fOverlapArea == fMinOverlapArea) && (fTotalArea < fMinTotalArea)))
		{
			// Remember the minimum overlap area and the distribution where it occurred.
			fMinOverlapArea = fOverlapArea;
			fMinTotalArea = fTotalArea;
			nNumFirstSplit = j;
		}
	}

	return 0;
}

int RHTree::ChooseAxisToSplit(RWGOrderedVector(RecordEntry) &listEntry,
	int &nAxis)
{
	int i;
	size_t j;
	double fMargin, fMinMargin = 0.0, fTotalArea;

	for (i=0; i < NUM_DIM; i++)
	{
		if (0 != SortEntryByAxis(listEntry, i))
		{
	 		advise("RHTree::ChooseAxisToSplit(): Unable to sort entry by axis.");
			return 1;
		}

		// Find the distribution with the least margin value.
		// Check if the number of entries are more than a certain minimum.
		ASSERT(listEntry.entries() >= 2* MIN_SVENTRY);
		for (j= MIN_SVENTRY ; j <= (listEntry.entries() - MIN_SVENTRY); j++)
		{
			// Compute the margin value of the bounding boxes for the split at
			// with the first split containing j entries.
			fMargin = 
				ComputeGoodnessValue(listEntry, j, RHTree::MarginValueType, fTotalArea);
			if (((0 == i) && (MIN_SVENTRY == j)) || (fMargin < fMinMargin))
			{
				// Remember the minimum margin and the axis where it occurred.
				fMinMargin = fMargin;
				nAxis = i;
			}
		}
	}

	return 0;
}

// Compute the required value for the given distribution
// The incoming entries (listEntry) are sorted in certain order and nNumFirstSplit is the
// number of entries in the first part of the distribution.
// Also return the area since that is useful for resolving ties
double RHTree::ComputeGoodnessValue(RWGOrderedVector(RecordEntry) &listEntry,
		size_t nNumFirstSplit, int nValueType, double &fTotalArea)
{
	RecordEntry cover1, cover2;
	size_t i;

	cover1.ClearEntry();
	cover2.ClearEntry();

	// Basic check for the splitting place
	ASSERT((nNumFirstSplit >0) && (nNumFirstSplit < listEntry.entries()));

	// cover the first part
	for (i=0; i < nNumFirstSplit; i++)
		cover1.CoverEntry(listEntry[i]);

	// cover the second part.
	for (;i<listEntry.entries(); i++)
		cover2.CoverEntry(listEntry[i]);

	fTotalArea = cover1.Area() + cover2.Area();
	switch (nValueType)
	{
		case MarginValueType:
			return (cover1.Margin() + cover2.Margin());
			break;
		case OverlapValueType:
			return cover1.OverlapArea(cover2);
			break;
		case AreaValueType:
			return cover1.Area() + cover2.Area();
		default:
	 		advise("RHTree::ComputeGoodnessValue(): Invalid value type.");
			break;
	}
	
	return 0.0;
}

int RHTree::SortEntryByAxis(RWGOrderedVector(RecordEntry) &listEntry, int nAxis)
{
	size_t i, j;
	RecordEntry record;

	// Sort the entries. Use simple insertion? sort. Change it to bubble sort or
	// heap sort or quick sort
	for (i=0; i<listEntry.entries(); i++)
		for (j=i+1; j<listEntry.entries(); j++)
		{
			if ((listEntry[i].m_low[nAxis] > listEntry[j].m_low[nAxis])
				|| ((listEntry[i].m_low[nAxis] == listEntry[j].m_low[nAxis])
				&& (listEntry[i].m_high[nAxis] > listEntry[j].m_high[nAxis])))
			{
				//Swap the entries.
				record = listEntry[i];
				listEntry[i] = listEntry[j];
				listEntry[j] = record;
			}
		}

	return 0;
}

//
// Split page by R tree algorithms--either linear or quadratic
//
int RHTree::RTreeSplitPage(DataPage *pPage1, DataPage *pPage2,
	DataPage *pNewPage1, RecordEntry &cover1, DataPage *pNewPage2, RecordEntry &cover2,
	long lTime)
{
	RWGOrderedVector(RecordEntry) listEntry(2*MAX_ENTRYCAPACITY);
	int i;

	if ((NULL == pPage1) || (NULL == pNewPage1) || (NULL == pNewPage2))
	{
	 	advise("RHTree::RTreeSplitPage(): Null new data pages.");
		return 1;
	}

	for (i=0; i < pPage1->m_numentries; i++)
	{
		RecordEntry record = pPage1->GetEntry(i);
		if (record.IsAlive(lTime))
			listEntry.insert(record);
	}

	if (NULL != pPage2)
	{
		for (i=0; i < pPage2->m_numentries; i++)
		{
			RecordEntry record = pPage2->GetEntry(i);
			if (record.IsAlive(lTime))
				listEntry.insert(record);
		}
	}

	// If too many or too few entries.
	if ((listEntry.entries() <= MAX_SVENTRY) || (listEntry.entries() > 2*MAX_SVENTRY))
	{
	 	advise("RHTree::RTreeSplitPage(): Too few or too many entries to split");
		return 1;
	}

	int index1, index2;

	cover1.ClearEntry();
	cover2.ClearEntry();
	pNewPage1->ClearPage();
	pNewPage2->ClearPage();
	cover1.m_ins = lTime;
	cover1.m_del = -1;
	cover2.m_ins = lTime;
	cover2.m_del = -1;
	cover1.m_info = pNewPage1->GetOffset();
	cover2.m_info = pNewPage2->GetOffset();

	if (RHTree::LinearAlgorithm == m_SplitAlgorithmType)
	{
		if ((0 != LinearPickSeeds(listEntry, &index1, &index2)) || index1 < 0 || index2 < 0)
		{
	 		advise("RHTree::RTreeSplitPage(): Unable to get the indexes for the seeds.");
			return 1;
		}
	} else
	{
		ASSERT(RHTree::QuadraticAlgorithm == m_SplitAlgorithmType);
		if ((0 != QuadraticPickSeeds(listEntry, &index1, &index2)) || index1 < 0 || index2 < 0)
		{
	 		advise("RHTree::RTreeSplitPage(): Unable to get the indexes for the seeds.");
			return 1;
		}
	}

	cover1.CoverEntry(listEntry[index1]);
	cover2.CoverEntry(listEntry[index2]);

	pNewPage1->AddEntry(listEntry[index1]);
	pNewPage2->AddEntry(listEntry[index2]);
	
	pNewPage1->m_numaliveentries++;
	pNewPage2->m_numaliveentries++;

	// Make sure that the correct items are removed in the right order.
	// The higher index item should be removed first.
	if (index1 > index2)
	{
		listEntry.removeAt(index1);
		listEntry.removeAt(index2);
	} else
	{
		listEntry.removeAt(index2);
		listEntry.removeAt(index1);
	}

	while (listEntry.entries() > 0)
	{
		// Check if the second page already has too many entries.
		if ((listEntry.entries() + pNewPage1->m_numentries <= MIN_SVENTRY) ||
			(pNewPage2->m_numentries >= MAX_SVENTRY))
		{
			AddAllEntries(pNewPage1, cover1, listEntry);
			break;
		} else if ((listEntry.entries() + pNewPage2->m_numentries <= MIN_SVENTRY)||
			(pNewPage1->m_numentries >= MAX_SVENTRY))
		{
			AddAllEntries(pNewPage2, cover2, listEntry);
			break;
		} else
		{
			// Pick up the last entry. This should be more efficient than getting the first one :-)
			// The algorithm is different in the quadratic pick next though
			int nNext;
			
			if (RHTree::LinearAlgorithm == m_SplitAlgorithmType)
			{
				nNext = LinearPickNext(listEntry, cover1, cover2);
			} else
			{
				ASSERT(RHTree::QuadraticAlgorithm == m_SplitAlgorithmType);
				nNext = QuadraticPickNext(listEntry, cover1, cover2);
			}

			if (nNext < 0)
			{
			 	advise("RHTree::RTreeSplitPage(): Unable to pick the next entry.");
				return 1;
			}

			RecordEntry record = listEntry[nNext];
			double areaIncrease1 = cover1.AreaIncrease(record);
			double areaIncrease2 = cover2.AreaIncrease(record);

			listEntry.removeAt(nNext);
			
			if ((areaIncrease1 < areaIncrease2) ||
				((areaIncrease1 == areaIncrease2) && (cover1.Area() < cover2.Area())))
			{
				pNewPage1->AddEntry(record);
				pNewPage1->m_numaliveentries++;
				cover1.CoverEntry(record);
			} else
			{
				pNewPage2->AddEntry(record);
				pNewPage2->m_numaliveentries++;
				cover2.CoverEntry(record);
			}
		}
	}
	return 0;
}

// Pick the first two seeds from the vector.
int RHTree::LinearPickSeeds(RWGOrderedVector(RecordEntry) &list, int *pIndex1, int *pIndex2)
{
	double fSeparation[NUM_DIM], minLow[NUM_DIM], maxHigh[NUM_DIM];
	int i, j, k, nNumEntries;
	double maxSeparation;

	ASSERT(pIndex1);
	ASSERT(pIndex2);
	*pIndex1 = -1;
	*pIndex2 = -1;

	for (i=0; i < NUM_DIM; i++)
	{
		fSeparation[i] = 0.0;
		maxHigh[i] = 0.0;
		minLow[i] = 0.0;
	}

	nNumEntries = list.entries();
	if (nNumEntries <2)
	{
	 	advise("RHTree::LinearPickSeeds(): Too few entries.");
		return 1;
	}

	for (j=0; j < nNumEntries; j++)
	for (i=0; i < NUM_DIM; i++)
	{
	 	if (0 == j) 
		{
		 	minLow[i] = list[j].m_low[i];
			maxHigh[i] = list[j].m_high[i];
		} else
		{
		 	if (minLow[i] > list[j].m_low[i])
				minLow[i] = list[j].m_low[i];
		 	if (maxHigh[i] < list[j].m_high[i])
				maxHigh[i] = list[j].m_high[i];
		}
	}
	/*if (fSeparation[i] < list[j].m_high[i] - list[j].m_low[i])
		fSeparation[i] = list[j].m_high[i] - list[j].m_low[i];*/
	
	for (i=0; i < NUM_DIM; i++)
	{
		fSeparation[i] = maxHigh[i] - minLow[i];
		if (0.0 >= fSeparation[i])
		{
		 	advise("RHTree::LinearPickSeeds(): Illegal separation.");
			return 1;
		}
	}

	maxSeparation = 0.0;
	for (j=0; j < nNumEntries; j++)
	for (k=j+1; k < nNumEntries; k++)
	for (i=0; i < NUM_DIM; i++)
	{
	 	// Find the separation between the rectangles and normalize it.
	 	double fRectSep =  (list[j].m_high[i] + list[j].m_low[i]) -
	 						(list[k].m_high[i] + list[k].m_low[i]);
		fRectSep /= fSeparation[i];

		if (fRectSep < 0.0)
			fRectSep = -fRectSep;

		if (fRectSep > maxSeparation)
		{
		 	maxSeparation = fRectSep;
			*pIndex1 = j;
			*pIndex2 = k;
		}
	}

	if ((*pIndex1 <0) || (*pIndex2 < 0))
	{
	 	advise("RHTree::LinearPickSeeds(): Invalid indexes.");
		return 1;
	}
	return 0;
}

// Pick the first two seeds from the vector.
int RHTree::QuadraticPickSeeds(RWGOrderedVector(RecordEntry) &list, int *pIndex1, int *pIndex2)
{
	int j, k, nNumEntries;
	double fMaxAreaWasted = 0.0;
	double fAreaWasted;
	RecordEntry cover;

	nNumEntries = list.entries();
	if (nNumEntries <2)
	{
	 	advise("RHTree::QuadraticPickSeeds(): Too few entries.");
		return 1;
	}

	for (j=0; j < nNumEntries; j++)
	for (k=j+1; k < nNumEntries; k++)
	{
		cover.ClearEntry();
		cover.CoverEntry(list[j]);
		cover.CoverEntry(list[k]);
	 	fAreaWasted = cover.Area() - list[j].Area() - list[k].Area();

		if ((fAreaWasted > fMaxAreaWasted) || ((0==j) && (1 == k)))
		{
		 	fMaxAreaWasted = fAreaWasted;
			*pIndex1 = j;
			*pIndex2 = k;
		}
	}
	
	if ((*pIndex1 <0) || (*pIndex2 < 0))
	{
	 	advise("RHTree::QuadraticPickSeeds(): Invalid indexes.");
		return 1;
	}

	return 0;
}

int RHTree::LinearPickNext(RWGOrderedVector(RecordEntry) &list, RecordEntry &cover1, RecordEntry &cover2)
{
 	return list.entries() - 1;
}

int RHTree::QuadraticPickNext(RWGOrderedVector(RecordEntry) &list, RecordEntry &cover1, RecordEntry &cover2)
{
 	size_t i;
 	int nNext=-1;
	double fMaxDif;
 	for (i =0; i < list.entries(); i++)
	{
		double fIncDif = cover1.AreaIncrease(list[i]) - cover2.AreaIncrease(list[i]);
		if (fIncDif < 0.0)
			fIncDif = -fIncDif;
	 	if ((fIncDif > fMaxDif) || (0 == i))
		{
			nNext = i;
			fMaxDif = fIncDif;
		}
	}
 	return nNext;
}

// Add all the entries to the page in question.
int RHTree::AddAllEntries(DataPage* pNewPage, RecordEntry &cover,
	RWGOrderedVector(RecordEntry) &listEntry)
{
	size_t i;
	RecordEntry record;
	for (i=0; i < listEntry.entries(); i++)
	{
		record = listEntry[i];
		pNewPage->AddEntry(listEntry[i]);
		pNewPage->m_numaliveentries++;
		cover.CoverEntry(listEntry[i]);
	}

	listEntry.clear();
	return 0;
}


// Choose the leaf nodes.
RWoffset RHTree::ChooseLeafNode(const RecordEntry &record, long lTime)
{
	int index;
	HeadPage *pHead;
	RWoffset offset;

	m_offset.clear();
	m_index.clear();
	m_depth = 0;
	offset = m_cache->GetFirstPageOffset();

	pHead = (HeadPage *) m_cache->GetPage(offset, HEADPAGETYPE);

	if (NULL == pHead)
	{
	 	advise("RHTree::ChooseLeafNode(): NULL head page.");
		return RWNIL;
	}

	index = pHead->FindRootIndex(lTime);

	m_offset.insert(offset);		// Of the head page...
	m_index.insert(index);		// index in the head page !!
	m_depth++;

	if (index < 0)
	{
	 	advise("RHTree::ChooseLeafNode(): Negative Index.");
		return RWNIL;
	}

	offset = pHead->GetEntry(index).m_info;
	if (RWNIL == offset)
	{
	 	advise("RHTree::ChooseLeafNode(): NULL offset.");
		return RWNIL;
	}

	while (TRUE)
	{
		// Read the next page...
		DataPage* pPage = (DataPage *) m_cache->GetPage(offset, DIRPAGETYPE);
		
		if (DATAPAGETYPE == pPage->GetPageType())
		{
			m_offset.insert(offset);
			m_index.insert(-1);
			m_depth++;
			return offset;
		}
		index = pPage->ClosestAliveEntry(record, lTime);
		if (index < 0)
		{
		 	advise("RHTree::ChooseLeafNode(): Negative Index.");
			return RWNIL;
		}
		m_offset.insert(offset);
		m_index.insert(index);
		m_depth++;

		offset = pPage->GetEntry(index).m_info;
		if (RWNIL == offset)
		{
		 	advise("RHTree::ChooseLeafNode(): NULL offset.");
			return RWNIL;
		}
	}

	return RWNIL;		
}

int RHTree::QueryDirectoryRecordFunction(RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone)
{

	bDone = FALSE;
	bOverlap = PageRecord.OverlapsAndIsAlive(QueryRecord);	

	if (bOverlap) {
	    genStats.queryDirNodes++;

	    if (bQueryDebug) {
	      cerr << "Query Dir Node: ";
	      PageRecord.PrintEntry(cerr);
	    }
	}
	return 0;
}

int RHTree::QueryDataRecordFunction(RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone)
{

  bDone = FALSE;
  bOverlap = PageRecord.OverlapsAndIsAlive(QueryRecord);	

	if (bOverlap) {
	  genStats.queryLeafNodes++;
	  genStats.IncNumQueryResults();

	  if (bQueryDebug) {
		  cerr << "Result Leaf Node: ";
	 	  PageRecord.PrintEntry(cerr);
	  }
	}
	return 0;
}

int RHTree::DeleteQueryDataRecordFunction(RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone)
{

	bDone = PageRecord.IdenticalEntry(QueryRecord) && PageRecord.IsAlive(QueryRecord.m_del);
	bOverlap = PageRecord.OverlapsAndIsAlive(QueryRecord);

	if (bDeleteDebug)
	{
		if (bDone)
		{
		 	cerr << "Deleted Record ";
		 	PageRecord.PrintEntry(cerr);
		}
	}

	return 0;
}

int RHTree::DeleteQueryDirectoryRecordFunction(RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone)
{

	bDone = FALSE;
	bOverlap = PageRecord.DoesIncludeAndIsAlive(QueryRecord);	

	if (bDeleteDebug)
	{
		if (bOverlap)
		{
		  cerr << "Query Page Record ";
		  PageRecord.PrintEntry(cerr);
		}
	}
	return 0;
}
