
#if !defined(__RHTREE_H_)
#define __RHTREE_H_

#include <iostream.h>
#include <rw/cstring.h>
#include <rw/filemgr.h>
#include <rw/gordvec.h>
#include <rw/gsortvec.h>
#include <rw/hashtab.h>
#include <rw/collint.h>
#include <limits.h>

#if defined(_DEBUG)
#define ASSERT(f) \
	if (!(f)) advise("Assertion Failed")
#define VERIFY(f)	ASSERT(f)
#else
#define ASSERT(f)	((void) 0)
#define VERIFY(f)	((void) f)
#endif

// Used for a higher level of debug. More intensive checking.
#if defined(_DEBUG1)
#define ASSERT1(f) \
	if (!(f)) advise("Assertion Failed")
#define VERIFY1(f)	ASSERT1(f)
#else
#define ASSERT1(f)	((void) 0)
#define VERIFY1(f)	((void) f)
#endif

// Macro Definitions
#define DOUBLE_MAX		LONG_MAX

#define NUM_DIM			2		// Number of dimensions
#define MAX_ENTRY		50//10//50		// The page size in terms of the number of entries
#define	MIN_WVENTRY		11//10//3//10		// The minimum number of entries for any page.
#define MIN_SVENTRY		20//4//19		// Min. number of entries for a new page
// mario: DO NOT INCREASE THIS BEYOND 40. THE INDEX WILL NOT WORK!!
#define MAX_SVENTRY		40//8//40		// Max. number of entries for a new page, Psvo = 40/50 = 0.8
#define MID_ENTRY		30				// Used for computing the middle break point for some of the cases.
 
#define	MAX_ENTRYCAPACITY		(MAX_ENTRY +2)		// Should be atleast 2 more then MAX_ENTRY

#define MAX_CACHE		20		// Size of the cache
#define MAX_FIRSTPAGE	10000		// Entries in the first page which keeps track of the root
#define	MAX_PATH		10		// Length of the path taken while searching in the rtree..

#define INDEX_ROOTPAGE	1		// Index of the root page in the path 
					// during a query, insert or delete

//The number for each operation in the input file
#define OP_DELETE		0
#define OP_INSERT		1
#define OP_QUERY		2

#define	TRUE			1
#define	FALSE			0

#define MIN(a,b)		((a) > (b) ? (b) : (a))
#define MAX(a,b)		((a) > (b) ? (a) : (b))

// Used to store a list of the offsets etc...
declare(RWGVector, RWoffset)
declare(RWGOrderedVector, RWoffset)
declare(RWGVector, int)
declare(RWGOrderedVector, int)
declare(RWGVector, double)
declare(RWGOrderedVector, double)


// This is used to sort a set of entries on the basis of a double
// type and get their indexes.
typedef struct _sortstruct
{
	double fValue;
	int nIndex;
} SortStruct;

// Used to compare the structures.
extern int CompareSortStruct(const SortStruct* pA, const SortStruct* pB);

declare(RWGVector, SortStruct);
declare(RWGSortedVector, SortStruct);

class RecordEntry;

// Used for analyzing the results.
// Return Value: TRUE means that there is a match in this record else there is not.
// When bDone is set the searching is over.
// bMatch means that this path should be searched further.
typedef int (*QUERYFNPTR) (RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone);

// ------------------------------------------------------------------------------------

// Abstract Base class for the records..
class Entry
{
public:
	Entry() {}
	virtual ~Entry() {}

	virtual int WriteRecord(RWFileManager *pFM) const=0;
	virtual int ReadRecord(RWFileManager *pFM)=0;
	virtual int GetSize() {return 0;}
	virtual int operator==(const Entry &) const;
	virtual double Area() const;
	virtual int PrintEntry(ostream &os) const;
};

// Data record entry.
class RecordEntry: public Entry
{
public:
	double m_low[NUM_DIM];
	double m_high[NUM_DIM];
	long m_ins;
	long m_del;
	RWoffset m_info;

	RecordEntry();
	virtual ~RecordEntry() { }
	
	virtual int WriteRecord(RWFileManager *pFM) const;
	virtual int ReadRecord(RWFileManager *pFM);
	virtual int GetSize();
	virtual int operator==(const RecordEntry& record) const;
	virtual int IdenticalEntry(const RecordEntry& record) const;
	int IsAlive(long lTime);
	virtual double Area() const;
	virtual double Margin() const;
	virtual double OverlapArea(const RecordEntry& record) const;
	virtual double AreaIncrease(const RecordEntry& record) const;
	virtual int DoesInclude(const RecordEntry &record) const;
	virtual int DoesIncludeAndIsAlive(const RecordEntry &record) const;
	virtual int Overlaps(const RecordEntry &record) const;
	virtual int OverlapsAndIsAlive(const RecordEntry &record) const;

	// Computes the adjusted entry which includes the record
	RecordEntry ReturnAdjustedEntry(const RecordEntry &record) const;

	// Extends the entry to cover the record passed.
	RecordEntry &CoverEntry(const RecordEntry &record);

	virtual int ClearEntry();
	virtual int PrintEntry(ostream &os) const;
};

declare(RWGVector, RecordEntry)
declare(RWGOrderedVector, RecordEntry)


// ------------------------------------------------------------------------------------
// Page information
enum PageType {UNKNOWNPAGETYPE, HEADPAGETYPE, DIRPAGETYPE, DATAPAGETYPE };

//base class for paging
class StoragePage
{
public:
	int m_type;
	RWoffset m_offset;
	
	StoragePage() { m_type = UNKNOWNPAGETYPE; m_offset = RWNIL;}
	virtual ~StoragePage() {}

	virtual int ReadPage(RWFileManager* pFM, RWoffset offset){ return 0;}
	virtual int WritePage(RWFileManager* pFM, RWoffset offset) { return 0;}
	virtual int GetSize();
	void SetOffset(RWoffset offset) { m_offset = offset; }
	RWoffset GetOffset() {return m_offset;}
	int SetPageType(int nType) { return m_type = nType;}
	int GetPageType() { return m_type;}
	virtual RecordEntry &GetEntry(int index) = 0;
	virtual int ClearPage();
	virtual int PrintPage(ostream &os) const;

};


// this is the head page in the file manager
class HeadPage: public StoragePage
{
public:
	int m_numpages;
	int m_depth;
	int m_numentries;
	RecordEntry m_rootpage[MAX_FIRSTPAGE];	// Used to maintain the pointer to the first page.

	HeadPage():StoragePage()
	{
		m_type = HEADPAGETYPE;
		m_numpages = 0;
		m_numentries = 0;
		m_depth = 0;
	}

	virtual ~HeadPage()	{ }

	virtual int WritePage(RWFileManager *pFM, RWoffset offset);
	virtual int ReadPage(RWFileManager *pFM, RWoffset offset);
	virtual int GetSize();
	int FindRootIndex(long lTime);
	virtual RecordEntry &GetEntry(int index);
	virtual int AddEntry(const RecordEntry &record);
	virtual int ModifyEntry(int index, const RecordEntry &record);
	virtual int ClearPage();
	virtual int PrintPage(ostream &os) const;

};


// Used For storing data.
class DataPage: public StoragePage
{
public:
	int m_numentries;
	int m_numaliveentries;
	RecordEntry m_entry[MAX_ENTRYCAPACITY];

	DataPage():StoragePage() { m_type = DATAPAGETYPE; m_numentries = 0; m_numaliveentries =0;}
	virtual ~DataPage() { }

	virtual int WritePage(RWFileManager *pFM, RWoffset offset);
	virtual int ReadPage(RWFileManager *pFM, RWoffset offset);
	virtual int GetSize();
	
	// It copies the current page's active entries to the new data page.
	int VersionSplit(DataPage* pPage, long lTime);
	
	int ClosestAliveEntry(const RecordEntry& record, long lTime);
	virtual RecordEntry &GetEntry(int index);
	RecordEntry ComputeCoverEntry(long lTime);
	virtual int AddEntry(const RecordEntry &);
	virtual int ModifyEntry(int index, const RecordEntry &record);
	virtual int ClearPage();
	
	int FindSibling(const RecordEntry record, int SkipIndex, long lTime);
	int AddAliveEntries(DataPage *pPage, long lTime);
	int AddAliveEntriesChangeInsertionTime(DataPage *pPage, long lTime);
	virtual int PrintPage(ostream &os) const;
	virtual int UpdateNumAliveEntries(long lTime);
	virtual int ComputedNumberOfAliveEntries(long lTime);

private:

};

// ------------------------------------------------------------------------------------
// Cache management
class PageCache
{
private:
	//Attributes

	int numPages;			// Number of pages in the cache
	StoragePage *m_pagecache[MAX_CACHE];
	RWoffset m_offsetpage[MAX_CACHE];
	int m_lockpagecount[MAX_CACHE];
	int m_modifiedflag[MAX_CACHE];
	RWFileManager *m_pfm;
	RWCString m_filename;

public:
	//Functions
	PageCache();
	~PageCache();

	//Basic Functions
	int CreateIndex(const char *szFileName);
	int OpenIndex(const char *szFileName);
	int CloseIndex();
	int RemoveIndex();
	int FlushCache(); // Dumps everything to disk.
	RWFileManager* GetFileManager() {return m_pfm;}
	int NumOfLockedPages();

	// Page manipulation functions.
	RWoffset GetFirstPageOffset();
	int PutFirstPageOffset(RWoffset offset);
	RWoffset NewPage(int nType);
	StoragePage* GetPage(RWoffset nOffset, int nType);
	int PutPage(RWoffset nOffset, int nType);
	int LockPage(RWoffset nOffset, int nType);
	int UnLockPage(RWoffset nOffset, int nType);
	RecordEntry GetPageEntry(RWoffset offset, int nType, int index);

private:

	//Functions
	int IncreasePriorityPage(int nNum);	// Moves a used page up.
	int FindPage(RWoffset offset, int nType);		//Finds the page in the cache
	StoragePage* ReadDiskPage(RWoffset nOffset, int nType);		//Read from the disk
	int DumpPageToDisk();					//Dumps a page to the disk

};


// ------------------------------------------------------------------------------------
// The definition of the rtree with history.
class RHTree
{

public:
	//Various type of splitting algorithms
	enum SplitAlgoType {QuadraticAlgorithm = 1, LinearAlgorithm,
		RStarTreeAlgorithm};

	// Various types of sibling finding algorithms
	enum SiblingAlgoType {AreaMinimizeAlgorithm =1, OverlapMaximizeAlgorithm,
		MarginMinimizeAlgorithm, AreaParameterAlgorithm,
		RandomSiblingAlgorithm};
	
	// Types of values computed for the distributions
	enum DistributionValueType{MarginValueType=1, OverlapValueType,
		AreaValueType};

	//Attributes
	long currentTime;

	// Global choices for the RHTree
	static int bQueryDebug;					// If the query debugging is enabled.
	static int bDeleteDebug;				// If the delete debugging is enabled. 

	static int m_bQueryProfileEnabled;		// If the query profiling is enabled. 
	static int m_nQueryProfileBlockSize;	// How many query results to group together while profiling

	static int bCacheEnabled;				// If the caching is enabled.
	static int m_SplitAlgorithmType;		// Type of the split algorithm type.

	static int m_SiblingAlgorithmType;		// Sibling finding algorithm type.
	static double m_fAreaPerimeterTradeoff;	// Area and parameter tradeoff

	static int m_bVermaUpdatingEnabled;		// If Verma modifications to German algo is enabled

	static int m_bTransactionIntervalQueryEnabled; // Transaction Interval Query enabled?

	RWHashTable	m_ScannedPageTable;

public:
	//Methods
	RHTree();
	~RHTree();

	// For General Information
	static void PrintRHTreeChoices(ostream &os);

	// Implementation
	int CreateTree(const char *szFileName);
	int RemoveTree();
	int CloseTree();
	int OpenTree(const char *szFileName);
	int InsertRecord(RecordEntry &record);
	int DeleteRecord(RecordEntry &record);
	int QueryRecords(RecordEntry &record, QUERYFNPTR pDirFunction, QUERYFNPTR pDataFunction,
		int bCleanup, int &bDone);

	static int QueryDataRecordFunction(RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone);
	static int QueryDirectoryRecordFunction(RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone);
	static int DeleteQueryDataRecordFunction(RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone);
	static int DeleteQueryDirectoryRecordFunction(RecordEntry &QueryRecord, RecordEntry &PageRecord, int &bOverlap,
	int &bDone);

	int FlushCache() { return (m_cache) ? m_cache->FlushCache():1;}

private:
	//Functions
	RWoffset ChooseLeafNode(const RecordEntry &record, long lTime);
	
	//To find the root page in the current path
	int IsRootPage(RWoffset offset)
	{ return (m_depth > INDEX_ROOTPAGE) && (m_offset[INDEX_ROOTPAGE] == offset);}

	// This splits the page in question and gives back another page of the same type
	// with the split entries.
	int SplitPage(DataPage *pPage1, DataPage *pPage2,
		DataPage *pNewPage1, RecordEntry &cover1, DataPage *pNewPage2, RecordEntry &cover2,
		long lTime);
	int RStarTreeSplitPage(DataPage *pPage1, DataPage *pPage2,
		DataPage *pNewPage1, RecordEntry &cover1, DataPage *pNewPage2, RecordEntry &cover2,
		long lTime);
	int RTreeSplitPage(DataPage *pPage1, DataPage *pPage2,
		DataPage *pNewPage1, RecordEntry &cover1, DataPage *pNewPage2, RecordEntry &cover2,
		long lTime);
	int ChooseSplitNumForGivenAxis(RWGOrderedVector(RecordEntry) &listEntry,
		int nAxis, size_t &nNumFirstSplit);
	int ChooseAxisToSplit(RWGOrderedVector(RecordEntry) &listEntry,
		int &nAxis);
	double ComputeGoodnessValue(RWGOrderedVector(RecordEntry) &listEntry,
		size_t nNumFirstSplit, int nValueType, double &fTotalArea);
	int SortEntryByAxis(RWGOrderedVector(RecordEntry) &listEntry, int nAxis);

	
	// This is the primary function for updating a node and handling the logic
	int UpdateNode(RWoffset offset, int nType, int nNumModifedRecords, int indexArray[],
		RecordEntry recordArray[], int nNumNewRecords, RecordEntry newRecordArray[],
		int nNetChange,	long lTime);

	int AddAllEntries(DataPage* pNewPage, RecordEntry &cover,
		RWGOrderedVector(RecordEntry) &list);

	// This is to pick seeds while splitting one or more pages..
	int LinearPickSeeds(RWGOrderedVector(RecordEntry) &list, int *pIndex1, int *pIndex2);
	int QuadraticPickSeeds(RWGOrderedVector(RecordEntry) &list, int *pIndex1, int *pIndex2);
	int LinearPickNext(RWGOrderedVector(RecordEntry) &list, RecordEntry &cover1, RecordEntry &cover2);
	int QuadraticPickNext(RWGOrderedVector(RecordEntry) &list,
			RecordEntry &cover1, RecordEntry &cover2);

	// Used while querying.
	int ScanNode(RWoffset offset, RecordEntry &record,
		QUERYFNPTR pDirFunction, QUERYFNPTR pDataFunction, int &bDone);

	// To copy entries from source to target page.
	int CopyEntries(DataPage *pSourcePage, DataPage *pTargetPage,
		int nNumEntriesToCopy, long lTime);
	int ClearPath();

public:
	//Attributes

private:
	//Attributes
	PageCache *m_cache;
	RWCString m_filename;
	
	int m_depth;
	RWGOrderedVector(RWoffset) m_offset; 	// Used to store the path taken through the tree.
	RWGOrderedVector(int) m_index;	// to store the index of the entries in the path.
};

class CStats
{
	public:
	//Functions
		CStats();
		~CStats();
		int ClearStats();
		// Print the stat with certain message
		int PrintStats(ostream &os, const char* sMessage = NULL, long nNumberUpdates = 0) const;

		long IncNewPages(int nType) { return ++m_newpages;}
		long IncReadPages(int nType) {return ++m_readpages;}
		long IncWritePages(int nType) { return ++m_writepages;}

		long NumberOfNewPages() { return m_newpages;}
		long NumberOfPagesRead() { return m_readpages;}
		long NumberOfPagesWritten() { return m_writepages;}

		CStats operator+(CStats c_stat);
	
	private:
		long m_newpages;
		long m_readpages;
		long m_writepages;
};

extern CStats stats;

class CGeneralStats
{
	public:
	//Functions
		CGeneralStats();
		~CGeneralStats();

		int ClearGeneralStats();
		// PrintStats with certain message
		int PrintGeneralStats(ostream &os, const char* sMessage = NULL) const;

		CGeneralStats operator+(CGeneralStats c_stat);

		long IncNumInserts() {return ++m_numinserts;}
		long IncNumDeletes() {return ++m_numdeletes;}
		long IncNumQueries() {return ++m_numqueries;}
		long IncNumQueryResults() {return ++m_numqueryresults;}

		long NumberOfInserts() { return m_numinserts;}
		long NumberOfDeletes() { return m_numdeletes;}
		long NumberOfQueries() { return m_numqueries;}
		long NumberofQueryResults() { return m_numqueryresults;}

		double queryDirNodes;
		double queryLeafNodes;

	private:
	//Attributes
		long m_numinserts;
		long m_numdeletes;
		long m_numqueries;
		long m_numqueryresults;
};

extern CGeneralStats genStats;

// For storing a query profile structure
class QueryProfileStruct
{
public:
	double m_fPageSumSquare;
	double m_fPageSum;
	long m_lNumberQueries;
	long m_lPageMax;
	long m_lPageMin;
	QueryProfileStruct();
	virtual int operator==(const QueryProfileStruct& record) const;

};

// Vector for the above query struct
declare(RWGVector, QueryProfileStruct)
declare(RWGOrderedVector, QueryProfileStruct)

//
// Name: CQueryProfileStats
//
// Purpose:
//		Used for profiling the queries.
//
class CQueryProfileStats
{
public:
	//Functions
	CQueryProfileStats();
	~CQueryProfileStats();

	int ClearStats();
	int UpdateProfileStats(CGeneralStats& GenStats, CStats& Stats);
	int RememberStats(CGeneralStats& GenStats, CStats& Stats)
	{
		m_bRemembered = TRUE; 
		m_remGenStats = GenStats;
		m_remStats = Stats;
		return 0;
	}

	// PrintStats with certain message
	int PrintStats(ostream &os, const char* sMessage = NULL) const;

private:
	//Attributes
	int m_bRemembered; // whether remembered or not.
	CGeneralStats m_remGenStats;
	CStats m_remStats;
	RWGOrderedVector(QueryProfileStruct) m_QueryProfileVector;
};

#endif // __RHTREE_H_

