
// For the logging functions
#include "rhis.h"
#include "rhtree.h"
#include <math.h>


CStats::CStats()
{
 	ClearStats();
}

CStats::~CStats()
{
}

int CStats::ClearStats()
{
 	m_newpages = 0;
	m_readpages = 0;
	m_writepages = 0;

	return 0;
}

int CStats::PrintStats(ostream &os, const char* sMessage /* = NULL */,
	long nNumberOperations /* =0 */) const
{
	RWCString sMsg = "";

	if (sMessage)
		sMsg = sMessage;

	os << sMsg << "New Pages = " << m_newpages << endl;
	os << sMsg << "Read Pages = " << m_readpages << endl;
	os << sMsg << "Write Pages = " << m_writepages << endl;
	os << sMsg << "Read+Writes Pages = " << m_readpages + m_writepages << endl;
	if (nNumberOperations != 0)
	{
		os << sMsg << "Number of Operations = " << nNumberOperations << endl;
		os << sMsg << "Avg Read+Write Pages/Operation = " <<
			((double) m_readpages + m_writepages)/nNumberOperations << endl;
	}

	return 0;
}

CStats CStats::operator+(CStats c_stat)
{
	CStats t_stat;

	t_stat.m_newpages = m_newpages + c_stat.m_newpages;
	t_stat.m_readpages = m_readpages + c_stat.m_readpages;
	t_stat.m_writepages = m_writepages + c_stat.m_writepages;

	return t_stat;
}

// General Statistics
CGeneralStats::CGeneralStats()
{
 	ClearGeneralStats();
}

CGeneralStats::~CGeneralStats()
{
}

int CGeneralStats::ClearGeneralStats()
{
 	m_numinserts = 0;
	m_numdeletes = 0;
	m_numqueries = 0;
	m_numqueryresults = 0;

	queryDirNodes = 0.0;
	queryLeafNodes = 0.0;

	return 0;
}

int CGeneralStats::PrintGeneralStats(ostream &os, const char* sMessage /* = NULL */) const
{
	RWCString sMsg = "";

	if (sMessage)
		sMsg = sMessage;

	os << sMsg << "Number of Inserts = " << m_numinserts << endl;
	os << sMsg << "Number of Deletes = " << m_numdeletes << endl;
	os << sMsg << "Number of Queries = " << m_numqueries << endl;
	os << sMsg << "Number of Query Results = " << m_numqueryresults << endl;
	os << sMsg << "Average Number of Directory Nodes Accessed = " << queryDirNodes / m_numqueries << endl;
	os << sMsg << "Average Number of Leaf Nodes Accessed = " << queryLeafNodes / m_numqueries << endl;

	return 0;
}

CGeneralStats CGeneralStats::operator+(CGeneralStats c_stat)
{
	CGeneralStats t_stat;

	t_stat.m_numinserts = m_numinserts + c_stat.m_numinserts;
	t_stat.m_numdeletes = m_numdeletes + c_stat.m_numdeletes;
	t_stat.m_numqueries = m_numqueries + c_stat.m_numqueries;
	t_stat.m_numqueryresults = m_numqueryresults + c_stat.m_numqueryresults;

	return t_stat;
}

// Implementation of the Query profiler class

// Firstly a helper class to hold the information

QueryProfileStruct::QueryProfileStruct()
{
	m_fPageSumSquare = 0.0;
	m_fPageSum = 0.0;
	m_lNumberQueries = 0;
	m_lPageMin = LONG_MAX;
	m_lPageMax = -1;
}

int QueryProfileStruct::operator==(const QueryProfileStruct &QPS) const
{
	if ((m_fPageSumSquare == QPS.m_fPageSumSquare) &&
		(m_fPageSum == QPS.m_fPageSum) &&
		(m_lNumberQueries == QPS.m_lNumberQueries) &&
		(m_lPageMin == QPS.m_lPageMin) &&
		(m_lPageMax == QPS.m_lPageMax))
		return TRUE;
	else
		return FALSE;
}

// Now the actual class for handling statistics
CQueryProfileStats::CQueryProfileStats()
{
	ClearStats();
}

CQueryProfileStats::~CQueryProfileStats()
{
}

int CQueryProfileStats::ClearStats()
{
	m_bRemembered = FALSE;
	m_remGenStats.ClearGeneralStats();
	m_remStats.ClearStats();
	m_QueryProfileVector.clear();

	return 0;
}

int CQueryProfileStats::UpdateProfileStats(CGeneralStats& GenStats,
	CStats& Stats)
{
	ASSERT(m_bRemembered);
	m_bRemembered = FALSE;

	long lNumQueryResults = GenStats.NumberofQueryResults() -
			m_remGenStats.NumberofQueryResults();
	ASSERT(lNumQueryResults >=0);
	ASSERT(RHTree::m_nQueryProfileBlockSize > 0);
	int nQueryBlock = (int) (lNumQueryResults / RHTree::m_nQueryProfileBlockSize);

	// See if the array has enough space. If not then grow it by
	// inserting empty entries.
	while (nQueryBlock > (int) (m_QueryProfileVector.entries()-1))
	{
		QueryProfileStruct emptyProfileStruct;
		m_QueryProfileVector.insert(emptyProfileStruct);
	}

	long lPagesRead = Stats.NumberOfPagesRead() - m_remStats.NumberOfPagesRead();
	m_QueryProfileVector[nQueryBlock].m_fPageSumSquare += 
		((double) lPagesRead * (double) lPagesRead);
	m_QueryProfileVector[nQueryBlock].m_fPageSum += lPagesRead;
	m_QueryProfileVector[nQueryBlock].m_lNumberQueries += 1;

	if (m_QueryProfileVector[nQueryBlock].m_lPageMax < lPagesRead)
		m_QueryProfileVector[nQueryBlock].m_lPageMax = lPagesRead;

	if (m_QueryProfileVector[nQueryBlock].m_lPageMin > lPagesRead)
		m_QueryProfileVector[nQueryBlock].m_lPageMin = lPagesRead;

	return 0;
}

int CQueryProfileStats::PrintStats(ostream &os, const char* sMessage /* = NULL */) const
{
	RWCString sMsg = "";

	if (sMessage)
		sMsg = sMessage;

	os << "Query profiling stats" << endl;
	for (size_t i=0 ; i < m_QueryProfileVector.entries(); i++)
	{
		double fAvg, fVar, fNumQuery, fStdDev;

		if (0 != m_QueryProfileVector[i].m_lNumberQueries)
			fNumQuery = m_QueryProfileVector[i].m_lNumberQueries;
		else
			fNumQuery = 1;

		fAvg = m_QueryProfileVector[i].m_fPageSum/fNumQuery;
		fVar = m_QueryProfileVector[i].m_fPageSumSquare/fNumQuery
			- fAvg * fAvg;
		fStdDev = sqrt(fVar);
		os << sMsg << "Query Profile :" << i << ":" << fAvg << ":"
			<< fStdDev << ":"
			<< m_QueryProfileVector[i].m_lPageMax << ":"
			<< m_QueryProfileVector[i].m_lPageMin << ":"
			<< m_QueryProfileVector[i].m_lNumberQueries
			<< endl;
	}

	return 0;
}

