// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

/*
	Creates a boxes file from a data file.
	Split: Uses a split file to create boxes according to splits at specific time instants.
	Nosplit: Uses only one box per data trajectory.
	Piecewise: Splits each data trajectory at function end-points.
*/

#include "global.h"

const int SPLIT = 0;
const int NOSPLIT = 1;
const int PIECEWISE = 2;

int main(int argc, char** argv)
{
	char dataFile[1000] = "";
	char numberOfSplitsFile[1000] = "";
	int policy = SPLIT;

	for (int i = 1; i < argc; i++)
	{
		if (strcmp(argv[i], "-df") == 0)
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing data file." << endl;
				return -1;
			}
			strncpy(dataFile, argv[i], 1000);
		}
		else if (strcmp(argv[i], "-nf") == 0)
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing number of splits file." << endl;
				return -1;
			}
			strncpy(numberOfSplitsFile, argv[i], 1000);
		}
		else if (strcmp(argv[i], "-p") == 0)
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing policy." << endl;
				return -1;
			}
			if (strcasecmp(argv[i], "split") == 0) policy = SPLIT;
			else if (strcasecmp(argv[i], "nosplit") == 0) policy = NOSPLIT;
			else if (strcasecmp(argv[i], "piecewise") == 0) policy = PIECEWISE;
		}
		else
		{
			cerr << "Usage: " << endl
				<< "  -df data file" << endl
				<< "  -nf number of splits file" << endl
				<< "  -p policy (split | nosplit | piecewise)" << endl;
			return -1;
		}
	}
	
	if (policy == SPLIT)
	{
		ifstream df(dataFile);
		if (! df)
		{
			cerr << "Cannot open file " << dataFile << "." << endl;
			return -1;
		}

		ifstream nf(numberOfSplitsFile);
		if (! nf)
		{
			cerr << "Cannot open file " << numberOfSplitsFile << "." << endl;
			return -1;
		}

		map<int, int> splitsMap;
		char buf[1000];
		while (nf.getline(buf, 1000))
		{
			int i, j;
			sscanf(buf, "%d %d", &i, &j);
			splitsMap.insert(pair<int, int>(i, j));
		}

		cout.precision(PRECISION);
		cout << fixed;

		MovingObject o;
		while (parseObject(df, o))
		{
			SplitInformation si;
			if (splitsMap[o.m_id] > 0) greedySplit(o, si);

			vector<Point> c;
			o.toPointArray(c);

			int start = c[0].m_t;
			int end = c[c.size() - 1].m_t;

			int s = 0;
			int sp = splitsMap[o.m_id];

			for (int i = 0; i < sp; i++)
			{
				int e = si.m_splitPoints[splitsMap[o.m_id]][i];

				double maxx = -numeric_limits<double>::max();
				double maxy = -numeric_limits<double>::max();
				double minx = numeric_limits<double>::max();
				double miny = numeric_limits<double>::max();

				for (int j = s; j <= e; j++)
				{
					maxx = std::max(maxx, c[j].m_x);
					maxy = std::max(maxy, c[j].m_y);
					minx = std::min(minx, c[j].m_x);
					miny = std::min(miny, c[j].m_y);
				}

				cout << o.m_id << " " << minx << " " << maxx << " "
					<< miny << " " << maxy << " " << (start + s) << " " << (start + e) << endl;

				s = e;
			}

			double maxx = -numeric_limits<double>::max();
			double maxy = -numeric_limits<double>::max();
			double minx = numeric_limits<double>::max();
			double miny = numeric_limits<double>::max();

			for (unsigned int j = s; j < c.size(); j++)
			{
				maxx = std::max(maxx, c[j].m_x);
				maxy = std::max(maxy, c[j].m_y);
				minx = std::min(minx, c[j].m_x);
				miny = std::min(miny, c[j].m_y);
			}

			cout << o.m_id << " " << minx << " " << maxx << " "
				<< miny << " " << maxy << " " << (start + s) << " " << end << endl;
		}
	}
	else if (policy == NOSPLIT)
	{
		ifstream df(dataFile);
		if (! df)
		{
			cerr << "Cannot open file " << dataFile << "." << endl;
			return -1;
		}

		MovingObject o;
		while (parseObject(df, o))
		{
			vector<Point> c;
			o.toPointArray(c);

			int start = c[0].m_t;
			int end = c[c.size()- 1].m_t;

			double maxx = -numeric_limits<double>::max();
			double maxy = -numeric_limits<double>::max();
			double minx = numeric_limits<double>::max();
			double miny = numeric_limits<double>::max();

			for (unsigned int j = 0; j < c.size(); j++)
			{
				maxx = std::max(maxx, c[j].m_x);
				maxy = std::max(maxy, c[j].m_y);
				minx = std::min(minx, c[j].m_x);
				miny = std::min(miny, c[j].m_y);
			}

			cout.precision(PRECISION);
			cout << fixed;
			cout << o.m_id << " " << minx << " " << maxx << " "
				<< miny << " " << maxy << " " << start << " " << end << endl;
		}
	}
	else if (policy == PIECEWISE)
	{
		ifstream df(dataFile);
		if (! df)
		{
			cerr << "Cannot open file " << dataFile << "." << endl;
			return -1;
		}

		char buf[1000];

		while (df)
		{
			while (df.getline(buf, 1000) && strlen(buf) == 0);
			if (! df) return 0;

			int id = -1;
			vector<Function> func;
			vector<int> pieces;

			while (df && strlen(buf) != 0)
			{
				istringstream ss(buf);

				int start, end;
				ss >> id >> start >> end;
				pieces.push_back(end - start);

				int coefsx;
				ss >> coefsx;
				double cx[coefsx];
				for (int i = 0; i < coefsx; i++) ss >> cx[i];

				int coefsy;
				ss >> coefsy;
				double cy[coefsy];
				for (int i = 0; i < coefsy; i++) ss >> cy[i];

				Polynomial px = Polynomial(coefsx, cx);
				Polynomial py = Polynomial(coefsy, cy);
				func.push_back(Function(start, end, px, py));

				df.getline(buf, 1000);
			}
			
			vector<Point> cuts;

			for (unsigned int i = 0; i < func.size(); i++)
			{
				for (int T = func[i].m_start; T < func[i].m_end; T++)
				{
					cuts.push_back(Point(T, func[i].m_x.getValue(T), func[i].m_y.getValue(T)));
				}

				if (i == func.size() - 1)
				{
					cuts.push_back(Point(func[i].m_end, func[i].m_x.getValue(func[i].m_end), func[i].m_y.getValue(func[i].m_end)));
				}
			}

			int start = 0;
			for (unsigned int cIndex = 0; cIndex < pieces.size(); cIndex++)
			{
				int end = pieces[cIndex];
				double maxx = -numeric_limits<double>::max();
				double maxy = -numeric_limits<double>::max();
				double minx = numeric_limits<double>::max();
				double miny = numeric_limits<double>::max();

				for (int j = start; j <= start + end; j++)
				{
					maxx = std::max(maxx, cuts[j].m_x);
					maxy = std::max(maxy, cuts[j].m_y);
					minx = std::min(minx, cuts[j].m_x);
					miny = std::min(miny, cuts[j].m_y);
				}

				cout.precision(PRECISION);
				cout << fixed;
				cout << id << " " << minx << " " << maxx << " "
					<< miny << " " << maxy << " " << cuts[0].m_t + start << " " << (cuts[0].m_t + start + end) << endl;

				start = start + end;
			}
		}
	}
}
