// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

using namespace std;

int main(int argc, char** argv)
{
	if (argc != 2)
	{
		cerr << "Usage: " << argv[0] << " data_file." << endl;
		return -1;
	}

	ifstream df(argv[1]);
	if (! df)
	{
		cerr << "Cannot open " << argv[1] << "." << endl;
		return -1;
	}

	cout.precision(PRECISION);
	cout << fixed;

	MovingObject o;
	unsigned long totalObjects = 0;
	double averageLifetime = 0.0;
	double averageVolume = 0.0;
	double averageNumberOfFunctions = 0.0;
	map<unsigned long, unsigned long> objectsPerInstant;
	double minX = std::numeric_limits<double>::max();
	double maxX = std::numeric_limits<double>::min();
	double minY = std::numeric_limits<double>::max();
	double maxY = std::numeric_limits<double>::min();

	while(parseObject(df, o))
	{
		totalObjects++;
		averageLifetime += (double) (o.m_functions[o.m_functions.size() - 1].m_end - o.m_functions[0].m_start);
		averageNumberOfFunctions += o.m_functions.size();
		averageVolume += o.getVolume();
		Rectangle r = o.getMBR();
		minX = std::min(minX, r.m_xlow);
		maxX = std::max(maxX, r.m_xhigh);
		minY = std::min(minY, r.m_ylow);
		maxY = std::max(maxY, r.m_yhigh);

		for (long T = o.m_functions[0].m_start; T < o.m_functions[o.m_functions.size() - 1].m_end; T++)
		{
			map<unsigned long, unsigned long>::iterator it = objectsPerInstant.find(T);
			if (it != objectsPerInstant.end())
			{
				unsigned long N = (*it).second + 1L;
				objectsPerInstant.erase(it);
				objectsPerInstant.insert(pair<unsigned long, unsigned long>(T, N));
			}
			else
			{
				objectsPerInstant.insert(pair<unsigned long, unsigned long>(T, 1L));
			}
		}
	}

	double average = 0.0;
	unsigned long count = 0;

	cout << "Obejcts per time instant: " << endl;
	for (map<unsigned long, unsigned long>::iterator it = objectsPerInstant.begin(); it != objectsPerInstant.end(); it++)
	{
		cout << "  T: " << (*it).first << ", N: " << (*it).second << endl;
		average += static_cast<double>((*it).second);
		count++;
	}
	cout << "Average alive objects per time instant: " << average / static_cast<double>(count) << endl;
	cout << "Total objects: " << totalObjects << endl;
	cout << "Average lifetime: " << averageLifetime / (double) totalObjects << endl;
	cout << "Average number of Functions per object: " << averageNumberOfFunctions / (double) totalObjects << endl;
	cout << "Average volume: " << averageVolume / (double) totalObjects << endl;
	cout << "Universe: " << minX << ", " << maxX << ", " << minY << ", " << maxY << ", Volume: " << (maxX - minX) * (maxY - minY) << endl;

	return 0;
}
