// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

/*
	Assigns a number of splits per data trajectory, using dynamic programming to
	minimize the total volume of the resulting boxes.
*/

#include "global.h"

class Entry
{
public:
	int m_totalSplits;
	float* m_volumes;
	int m_id;

	Entry(SplitInformation& si);
	Entry(const Entry& e);
	~Entry() { delete[] m_volumes; }

	Entry& operator=(const Entry& e);

	float getVolume(int s)
	{
		return (s <= m_totalSplits) ? m_volumes[s] : m_volumes[m_totalSplits];
	}
};

Entry:: Entry(SplitInformation& si)
	:	m_totalSplits(0), m_id(si.m_id)
{
	m_totalSplits = si.getTotalSplits();
	m_volumes = new float[m_totalSplits + 1];
	for (int cIndex = 0; cIndex <= m_totalSplits; cIndex++) m_volumes[cIndex] = (float) si.m_volumes[cIndex];
}

Entry::Entry(const Entry& e)
{
	m_totalSplits = e.m_totalSplits;
	m_id = e.m_id;

	m_volumes = new float[m_totalSplits + 1];
	memcpy(m_volumes, e.m_volumes, (m_totalSplits + 1) * sizeof(float));
}

Entry& Entry::operator=(const Entry& e)
{
	if (this != &e)
	{
		m_totalSplits = e.m_totalSplits;
		m_id = e.m_id;

		delete[] m_volumes;
		m_volumes = new float[m_totalSplits + 1];
		memcpy(m_volumes, e.m_volumes, (m_totalSplits + 1) * sizeof(float));
	}

	return *this;
}

class DEntry
{
public:
	double m_volume;
	int m_splits;
	DEntry() : m_volume(-std::numeric_limits<double>::max()), m_splits(-1) {}
};

class DHistory
{
public:
	int m_totalSplitsLow, m_totalSplitsHigh, m_splitsForThisObject, m_id;
	DHistory() : m_totalSplitsLow(-1), m_totalSplitsHigh(-1), m_splitsForThisObject(-1), m_id(-1) {}
};

struct maxId : public binary_function<Entry*, Entry*, bool>
{
	bool operator()(Entry* __x, Entry* __y) const { return __x->m_id < __y->m_id; }
};

int main(int argc, char** argv)
{
	if (argc != 3)
	{
		cerr << "Usage:" << argv[0] << " data_file total_splits" << endl;
		return -1;
	}

	int splits = atoi(argv[2]);

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open file " << argv[1] << "." << endl;
		return -1;
	}

	priority_queue<Entry*, vector<Entry*>, maxId> pq;
	DEntry *a, *b;
	stack<DHistory*> history;

	try
	{
	 	a = new DEntry[splits + 1];
	 	b = new DEntry[splits + 1];
	}
	catch (...)
	{
		cerr << "Out of memory." << endl;
		return -1;
	}

	MovingObject o;
	int count = 0;
	double totalVolume = 0.0;
	double skippedVolume = 0.0;

	cpu_usage_aid cpu;
	begin_cpu_usage(&cpu);

	while (parseObject(fin, o))
	{
		count++;
		if (count % 10000 == 0) cerr << count << endl;

		SplitInformation si;
		dynamicSplit(o, si);
		totalVolume += si.m_volumes[0];

		if (si.getTotalSplits() == 0)
		{
			skippedVolume += si.m_volumes[0];
			cout << o.m_id << "\t0" << endl;
			continue;
		}

		Entry* e = new Entry(si);
		pq.push(e);
	}

	cpu_usage splitcpu;
	end_cpu_usage(&cpu, &splitcpu);
	cerr << "Total split time: " << splitcpu.total_time << endl;

	if (pq.empty()) return 0;

	Entry* e = pq.top(); pq.pop();

	for (int cSplit = 0; cSplit <= splits; cSplit++)
	{
		a[cSplit].m_volume = e->getVolume(cSplit);
		a[cSplit].m_splits = min(cSplit, e->m_totalSplits);

		if (cSplit < e->m_totalSplits)
		{
			DHistory* h = new DHistory();
			h->m_id = e->m_id;
			h->m_totalSplitsLow = cSplit;
			h->m_totalSplitsHigh = cSplit;
			h->m_splitsForThisObject = cSplit;
			history.push(h);
		}
	}

	DHistory* h = new DHistory();
	h->m_id = e->m_id;
	h->m_totalSplitsLow = e->m_totalSplits;
	h->m_totalSplitsHigh = splits;
	h->m_splitsForThisObject = e->m_totalSplits;
	history.push(h);
	delete e;

	count = 0;

	while (! pq.empty())
	{
		count++;
		//if (count % 1000 == 0) cerr << count << endl;

		e = pq.top(); pq.pop();

		for (int cSplit = 0; cSplit <= splits; cSplit++)
		{
			b[cSplit].m_volume = std::numeric_limits<double>::max();

			for (int cCurrentSplit = 0; cCurrentSplit <= std::min(cSplit, e->m_totalSplits); cCurrentSplit++)
			{
				int remainingSplits = cSplit - cCurrentSplit;
				double d = a[remainingSplits].m_volume + e->getVolume(cCurrentSplit);

				if (d < b[cSplit].m_volume)
				{
					b[cSplit].m_volume = d;
					b[cSplit].m_splits = cCurrentSplit;
				}
			}
		}

		// compress the information so that we do not have to keep one DHistory entry per possible split.
		// This is not a very efficient compression scheme, but it is simple.
		int cSplit = 1;
		h = new DHistory();
		h->m_totalSplitsLow = 0;
		h->m_totalSplitsHigh = 0;
		h->m_splitsForThisObject = b[0].m_splits;
		h->m_id = e->m_id;
		while (cSplit <= splits)
		{
			if (b[cSplit].m_splits != h->m_splitsForThisObject)
			{
				h->m_totalSplitsHigh = cSplit - 1;
				history.push(h);
				h = new DHistory();
				h->m_totalSplitsLow = cSplit;
				h->m_totalSplitsHigh = cSplit;
				h->m_splitsForThisObject = b[cSplit].m_splits;
				h->m_id = e->m_id;
			}
			cSplit++;
		}
		h->m_totalSplitsHigh = cSplit - 1;
		history.push(h);

		memcpy(a, b, (splits + 1) * sizeof(DEntry));

		delete e;
	}

	cerr << "Total: " << totalVolume << endl;
	cerr << "Optimal: " << a[splits].m_volume + skippedVolume << endl;
	cerr << "Gain: " << totalVolume - (a[splits].m_volume  + skippedVolume) << endl;

	h = history.top(); history.pop();
	cout << h->m_id << " " << a[splits].m_splits << endl;
	long rem = splits - a[splits].m_splits;
	int id = h->m_id;
	while (h->m_id == id)
	{
		delete h;
		if (history.empty()) break;
		h = history.top(); history.pop();
	}
	if (history.empty()) return 0;
	id = h->m_id;

	while (1)
	{
		if (id != h->m_id)
		{
			// a history entry for the remaining number of splits was not found for the current object. That
			// means there is a bug in the history compression code (see above). Normally there should be an entry
			// with an interval containing any possible number of remaining splits.
			cerr << "Ids do not much. This is a bug!" << endl;
			return -1;
		}

		// find the entry that corresponds to the number of remaining splits that we have to allocate.
		if (h->m_totalSplitsLow <= rem && rem <= h->m_totalSplitsHigh)
		{
			cout << h->m_id << " " << h->m_splitsForThisObject << endl;
			rem -= h->m_splitsForThisObject;

			// get rid of all entries belonging to this object.
			while (h->m_id == id)
			{
				delete h;
				if (history.empty()) break;
				h = history.top(); history.pop();
			}
			if (history.empty()) break;
			id = h->m_id;
		}
		else
		{
			// this was not the history entry we need. Try the next one.
			delete h;
			if (history.empty()) break;
			h = history.top(); history.pop();
		}
	}

	delete[] a;
	delete[] b;

	cpu_usage cpuu;
	end_cpu_usage(&cpu, &cpuu);

	cerr << "Total time: " << cpuu.total_time << endl;
	cerr << "System time: " << cpuu.system_time << endl;
	cerr << "User time: " << cpuu.user_time << endl;

	return 0;
}
