// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "AttractorGenerator.h"

void AttractorGenerator::adjustSpeed(MyMovingObject* o)
{
	double theta;
	Segment s = Segment(0, o->m_sx, o->m_sy, m_attractors[o->m_attractorIndex].m_x, m_attractors[o->m_attractorIndex].m_y);

	double v = generateSpeed();

	// remember that s.m_slope is in [-PI/2, PI/2].

	if (o->m_towardsAttractor == false)
	{
		bool left = false;
		if (flipCoin()) left = true;

		if (o->m_sx <= m_attractors[o->m_attractorIndex].m_x && ! left)
		{
			double d1 = nextUniformDouble(-M_PI_2, s.m_angle - (10.0 * M_PI / 180.0));
			double d2 = nextUniformDouble(s.m_angle + (10.0 * M_PI / 180.0), M_PI_2);
			theta = (nextUniformDouble() > 0.5) ? d1 : d2;
		}
		else if (o->m_sx >= m_attractors[o->m_attractorIndex].m_x && left)
		{
			double d1 = nextUniformDouble(M_PI_2, (M_PI + s.m_angle) - (10.0 * M_PI / 180.0));
			double d2 = nextUniformDouble((M_PI + s.m_angle) + (10.0 * M_PI / 180.0), 3.0 * M_PI_2);
			theta = (flipCoin()) ? d1 : d2;
		}
		else
		{
			if (left) v *= -1.0;
			theta = nextUniformDouble(0.2, M_PI_2 - 0.2);
			if (flipCoin()) theta *= -1;
		}
	}
	else
	{
		if (o->m_sx >= m_attractors[o->m_attractorIndex].m_x)
		{
			theta = nextUniformDouble(M_PI + s.m_angle - (5.0 * M_PI / 180.0), M_PI + s.m_angle + (5.0 * M_PI / 180.0));
		}
		else
		{
			theta = nextUniformDouble(s.m_angle - (5.0 * M_PI / 180.0), s.m_angle + (5.0 * M_PI / 180.0));
		}
	}

	o->m_vx = cos(theta) * v;
	o->m_vy = sin(theta) * v;
}

void AttractorGenerator::createRandomAttractors()
{
	if (m_attractors != 0) delete[] m_attractors;
	m_attractors = new Point[m_cAttractors];

	cerr << "Attractors:" << endl;
	for (int cIndex = 0; cIndex < m_cAttractors; cIndex++)
	{
		m_attractors[cIndex].m_x = nextUniformDouble(0.0, m_maxX);
		m_attractors[cIndex].m_y = nextUniformDouble(0.0, m_maxY);
		cerr << "  " << cIndex << ": " << m_attractors[cIndex].m_x << " " << m_attractors[cIndex].m_y << endl;
	}
}

void AttractorGenerator::generate()
{
	m_dataset.clear();

	for (int cIndex = 0; cIndex < m_simulationLength; cIndex++)
	{
		m_updateArray[cIndex].clear();
	}

	int updatesPerInstant = (int) floor(((double) m_datasetSize) * m_agility);

	// initialize all objects at random positions.
	for (int cObject = 0; cObject < m_datasetSize; cObject++)
	{
		MyMovingObject* o = new MyMovingObject();

		o->m_id = cObject;
		if (m_initialDistribution == ZIPF)
		{
			o->m_sx = nextZipfDouble(0.0, m_maxX, 1.0);
			o->m_sy = nextZipfDouble(0.0, m_maxY, 1.0);
		}
		else if (m_initialDistribution == GAUSSIAN)
		{
			o->m_sx = nextGaussianDouble(0.0, m_maxX);
			o->m_sy = nextGaussianDouble(0.0, m_maxY);
		}
		else
		{
			o->m_sx = nextUniformDouble(0.0, m_maxX);
			o->m_sy = nextUniformDouble(0.0, m_maxY);
		}
		o->m_st = 0;
    o->m_kt = -1;

		o->m_attractorIndex = nextUniformInt(0, m_cAttractors);
		o->m_towardsAttractor = (flipCoin()) ? true : false;

		adjustSpeed(o);

		double ax[2] = {o->m_sx - (o->m_vx * o->m_st), o->m_vx};
		double ay[2] = {o->m_sy - (o->m_vy * o->m_st), o->m_vy};
		Polynomial px = Polynomial(2, ax);
		Polynomial py = Polynomial(2, ay);
		Function f = Function(o->m_st, -1, px, py);
		o->m_functions.push_back(f);

		m_dataset.insert(pair<int, MovingObject*>(o->m_id, static_cast<MovingObject*>(o)));

		int t1 = o->m_st + m_maximumUpdateInterval;
		int t2 = o->m_st + (int) floor(abs(m_attractors[o->m_attractorIndex].m_x - o->m_sx) / abs(o->m_vx));
		int t = (o->m_towardsAttractor) ? min(t1, t2) : t1;
		if (t == 0) t = 1;
		assert(t >= 0);

		int t3;
		for (t3 = 1; t3 <= m_simulationLength; t3++)
		{
			if (o->getX(t3) > m_maxX || o->getY(t3) > m_maxY || o->getX(t3) < 0.0 || o->getY(t3) < 0.0) break;
		}

		o->m_outOfBounds = (t3 <= t) ? true : false;
		o->m_willReachAttractor = (o->m_towardsAttractor && t2 < t1 && t < t3) ? true : false;
		t = min(t, t3);

		assert(t >= 0);
		if (t < m_simulationLength)
		{
			m_updateArray[t].insert(o->m_id);
			o->m_kt = t;
		}

		if (cObject % 10000 == 0) cerr << cObject << endl;
	}

	for (int Tnow = 1; Tnow < m_simulationLength; Tnow++)
	{
		cerr << "Time: " << Tnow;

		int cTotalUpdates = 0;
		int cNeedToUpdate = updatesPerInstant;
		set<int> updated;

		if (! m_updateArray[Tnow].empty())
		{
			for (set<int>::iterator itUpdateArray = m_updateArray[Tnow].begin(); itUpdateArray != m_updateArray[Tnow].end(); itUpdateArray++)
			{
				map<int, MovingObject*>::iterator itDataset = m_dataset.find(*itUpdateArray);
				assert(itDataset != m_dataset.end());
				MyMovingObject* o = static_cast<MyMovingObject*>((*itDataset).second);
				cNeedToUpdate--;
				cTotalUpdates++;
				updated.insert(o->m_id);

				Function& f = o->m_functions.back();
				f.m_end = Tnow;

				o->m_sx = o->getX(Tnow);
				o->m_sy = o->getY(Tnow);
				o->m_st = Tnow;

				if (o->m_willReachAttractor)
				{
					// object reached attractor.
					o->m_towardsAttractor = false;
					o->m_willReachAttractor = false;

					adjustSpeed(o);

					double ax[2] = {o->m_sx - (o->m_vx * o->m_st), o->m_vx};
					double ay[2] = {o->m_sy - (o->m_vy * o->m_st), o->m_vy};
					Polynomial px = Polynomial(2, ax);
					Polynomial py = Polynomial(2, ay);
					Function f2 = Function(o->m_st, -1, px, py);
					o->m_functions.push_back(f2);

					int t = o->m_st + m_maximumUpdateInterval;
					if (t == Tnow) t++;

					int t3;
					for (t3 = Tnow + 1; t3 <= m_simulationLength; t3++)
					{
						if (o->getX(t3) > m_maxX || o->getY(t3) > m_maxY || o->getX(t3) < 0.0 || o->getY(t3) < 0.0) break;
					}

					o->m_outOfBounds = (t3 < t) ? true : false;
					t = min(t, t3);

					assert(t >= 0);
					if (t < m_simulationLength)
					{
						m_updateArray[t].insert(o->m_id);
						o->m_kt = t;
					}
				}
				else
				{
					if (o->m_towardsAttractor == false || o->m_outOfBounds == true)
					{
						// go towards a new attractor.
						int attractorIndex = nextUniformInt(0, m_cAttractors);
						while (o->m_attractorIndex == attractorIndex) attractorIndex = nextUniformInt(0, m_cAttractors);
						o->m_attractorIndex = attractorIndex;
						o->m_towardsAttractor = (o->m_outOfBounds == true || nextUniformDouble() <= m_attractionProbability) ? true : false;
					}

					adjustSpeed(o);

					double ax[2] = {o->m_sx - (o->m_vx * o->m_st), o->m_vx};
					double ay[2] = {o->m_sy - (o->m_vy * o->m_st), o->m_vy};
					Polynomial px = Polynomial(2, ax);
					Polynomial py = Polynomial(2, ay);
					Function f2 = Function(o->m_st, -1, px, py);
					o->m_functions.push_back(f2);

					int t1 = o->m_st + m_maximumUpdateInterval;
					int t2 = o->m_st + (int) floor(abs(m_attractors[o->m_attractorIndex].m_x - o->m_sx) / abs(o->m_vx));
					if (t2 == Tnow) t2++;
					int t = (o->m_towardsAttractor) ? min(t1, t2) : t1;

					int t3;
					for (t3 = Tnow + 1; t3 <= m_simulationLength; t3++)
					{
						if (o->getX(t3) > m_maxX || o->getY(t3) > m_maxY || o->getX(t3) < 0.0 || o->getY(t3) < 0.0) break;
					}

					o->m_outOfBounds = (t3 <= t) ? true : false;
					o->m_willReachAttractor = (o->m_towardsAttractor && t2 < t1 && t < t3) ? true : false;
					t = min(t, t3);

					assert(t >= 0);
					if (t < m_simulationLength)
					{
						m_updateArray[t].insert(o->m_id);
						o->m_kt = t;
					}
				}
			}
		}

		while (cNeedToUpdate > 0)
		{
			int id = nextUniformInt(0, m_datasetSize);
			set<int>::iterator itUpdated = updated.find(id);

			while (itUpdated != updated.end())
			{
				id = nextUniformInt(0, m_datasetSize);
				itUpdated = updated.find(id);
			}
			cNeedToUpdate--;
			cTotalUpdates++;
			updated.insert(id);

			map<int, MovingObject*>::iterator itDataset = m_dataset.find(id);
			MyMovingObject* o = static_cast<MyMovingObject*>((*itDataset).second);

			Function& f = o->m_functions.back();
			f.m_end = Tnow;

			o->m_sx = o->getX(Tnow);
			o->m_sy = o->getY(Tnow);
			o->m_st = Tnow;

			adjustSpeed(o);

			double ax[2] = {o->m_sx - (o->m_vx * o->m_st), o->m_vx};
			double ay[2] = {o->m_sy - (o->m_vy * o->m_st), o->m_vy};
			Polynomial px = Polynomial(2, ax);
			Polynomial py = Polynomial(2, ay);
			Function f2 = Function(o->m_st, -1, px, py);
			o->m_functions.push_back(f2);

			int t1 = o->m_st + m_maximumUpdateInterval;
			int t2 = o->m_st + (int) floor(abs(m_attractors[o->m_attractorIndex].m_x - o->m_sx) / abs(o->m_vx));
			if (t2 == Tnow) t2++;
			int t = (o->m_towardsAttractor) ? min(t1, t2) : t1;

			int t3;
			for (t3 = Tnow + 1; t3 <= m_simulationLength; t3++)
			{
				if (o->getX(t3) > m_maxX || o->getY(t3) > m_maxY || o->getX(t3) < 0.0 || o->getY(t3) < 0.0) break;
			}

			o->m_outOfBounds = (t3 <= t) ? true : false;
			o->m_willReachAttractor = (o->m_towardsAttractor && t2 < t1 && t < t3) ? true : false;
			t = min(t, t3);

			assert(t >= 0);
			if (o->m_kt >= 0) m_updateArray[o->m_kt].erase(o->m_id);

			if (t < m_simulationLength)
			{
				m_updateArray[t].insert(o->m_id);
				o->m_kt = t;
			}
		}

		cerr << ", Updates: " << cTotalUpdates << endl;
	}

	for (map<int, MovingObject*>::iterator itDataset = m_dataset.begin(); itDataset != m_dataset.end(); itDataset++)
	{
		MyMovingObject* o = static_cast<MyMovingObject*>((*itDataset).second);

		for (unsigned int cIndex = 0; cIndex < o->m_functions.size(); cIndex++)
		{
			Function& f = o->m_functions[cIndex];
			if (f.m_end == -1) f.m_end = m_simulationLength;

			cout << o->m_id << " " << f.m_start << " "  << f.m_end << " "	<< f.m_x.m_cLength << " ";

			for (int cCoef = 0; cCoef < f.m_x.m_cLength; cCoef++)
			{
				cout << f.m_x.m_pCoefficients[cCoef] << " ";
			}

			cout << f.m_y.m_cLength << " ";
			for (int cCoef = 0; cCoef < f.m_y.m_cLength; cCoef++)
			{
				cout << f.m_y.m_pCoefficients[cCoef] << " ";
			}

			cout << "0 0" << endl;
		}
		cout << endl;
	}

	delete[] m_attractors;
}

