// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "PolynomialGenerator.h"

void PolynomialGenerator::printObject(MyMovingObject* o, bool bKilled = false)
{
  int id;
  if (bKilled)
  {
    id = m_lastId;
    m_lastId++;
  }
  else
  {
    id = o->m_id;
  }

	for (unsigned int cIndex = 0; cIndex < o->m_functions.size(); cIndex++)
	{
		Function& f = o->m_functions[cIndex];
		if (f.m_end == -1) f.m_end = m_simulationLength;

  	cout << id << " " << f.m_start << " "  << f.m_end << " "	<< f.m_x.m_cLength << " ";

		for (int cCoef = 0; cCoef < f.m_x.m_cLength; cCoef++)
		{
			cout << f.m_x.m_pCoefficients[cCoef] << " ";
		}

		cout << f.m_y.m_cLength << " ";
		for (int cCoef = 0; cCoef < f.m_y.m_cLength; cCoef++)
		{
			cout << f.m_y.m_pCoefficients[cCoef] << " ";
		}

		cout << "0 0" << endl;

		for (int T = o->m_functions[cIndex].m_start; T < o->m_functions[cIndex].m_end; T++)
		{
			if (! (o->m_functions[cIndex].m_x.getValue(T) <= m_maxX + 1.0 && o->m_functions[cIndex].m_y.getValue(T) <= m_maxY + 1.0 &&
						 o->m_functions[cIndex].m_x.getValue(T) >= -1.0 && o->m_functions[cIndex].m_y.getValue(T) >= -1.0))
						 {
						 	cerr << endl << "1: " << id << " " << T << " " << o->m_functions[cIndex].m_x.getValue(T) << " " << o->m_functions[cIndex].m_y.getValue(T) << endl;
						 	exit(-1);
						}
		}

		if (cIndex == o->m_functions.size() - 1)
		{
			int T = o->m_functions[cIndex].m_end;
			if (! (o->m_functions[cIndex].m_x.getValue(T) <= m_maxX + 1.0 && o->m_functions[cIndex].m_y.getValue(T) <= m_maxY + 1.0 &&
						 o->m_functions[cIndex].m_x.getValue(T) >= -1.0 && o->m_functions[cIndex].m_y.getValue(T) >= -1.0))
						{
						 	cerr << endl << "2: " << id << " " << T << " " << o->m_functions[cIndex].m_x.getValue(T) << " " << o->m_functions[cIndex].m_y.getValue(T) << endl;
						 	exit(-1);
						}
		}

	}
  cout << endl;
}

MovingObject* PolynomialGenerator::createObject(int id, int st, double xmin, double xmax, double ymin, double ymax, int dist, bool bKilled)
{
	double x, y;
  if (dist == ZIPF)
  {
    x = nextZipfDouble(xmin, xmax, 1.0);
    y = nextZipfDouble(ymin, ymax, 1.0);
  }
  else if (dist == GAUSSIAN)
  {
    x = nextGaussianDouble(xmin, xmax);
    y = nextGaussianDouble(ymin, ymax);
  }
  else
  {
    x = nextUniformDouble(xmin, xmax);
    y = nextUniformDouble(ymin, ymax);
  }

  return createObject(id, st, x, y, bKilled);
}

MovingObject* PolynomialGenerator::createObject(int id, int st, double x, double y, bool bKilled = false)
{
	MyMovingObject* o = 0;

	map<int, MovingObject*>::iterator itDataset = m_dataset.find(id);

	if (itDataset != m_dataset.end())
	{
		o = static_cast<MyMovingObject*>((*itDataset).second);
	}
	else
	{
	  o = new MyMovingObject();
  	m_dataset.insert(pair<int, MovingObject*>(id, o));
	}

  if (bKilled)
  {
    printObject(o, true);
    o->m_functions.clear();
  }

	o->m_px = createPolynomial(x, st);
	o->m_py = createPolynomial(y, st);
  o->m_id = id;
  o->m_st = st;
  o->m_kt = -1;

	Function f = Function(o->m_st, -1, o->m_px, o->m_py);
	o->m_functions.push_back(f);

  int t1 = o->m_st + m_maximumUpdateInterval;
  int t2;

  for (t2 = st; t2 <= m_simulationLength; t2++)
  {
    if (o->getX(t2) > m_maxX || o->getY(t2) > m_maxY || o->getX(t2) < 0.0 || o->getY(t2) < 0.0) break;
  }

  o->m_outOfBounds = (t2 < t1) ? true : false;

  int t = min(t1, t2);

  if (t <= st) t = st + 1;
  if (t < m_simulationLength)
  {
    m_updateArray[t].insert(o->m_id);
    o->m_kt = t;
  }

  return o;
}

Polynomial PolynomialGenerator::createPolynomial(double start, unsigned int st)
{
  int degree = (flipCoin()) ? 1 : 2;
  double coefficients[degree + 1];
  double decreasing = (flipCoin()) ? 1.0 : -1.0;

  if (degree == 1)
  {
  	double v = generateSpeed() * decreasing;
    coefficients[0] = start - (v * st);
    coefficients[1] = v;
  }
  else if (degree == 2)
  {
  	double v2 = generateSpeed() * decreasing;
  	double v1 = generateSpeed() * decreasing;
  	double v0 = start;
    coefficients[0] = v0 - (v1 * st) + (v2 * st * st);
    coefficients[1] = v1 - (2 * v2 * st);
    coefficients[2] = v2;
  }

  return Polynomial(degree + 1, coefficients);
}

void PolynomialGenerator::generate()
{
  for (map<int, MovingObject*>::iterator itDataset = m_dataset.begin(); itDataset != m_dataset.end(); itDataset++)
  {
  	delete (*itDataset).second;
  }
  m_dataset.clear();

  for (int cIndex = 0; cIndex < m_simulationLength; cIndex++)
  {
    m_updateArray[cIndex].clear();
  }

  int updatesPerInstant = (int) ceil(((double) m_datasetSize) * m_agility);

  for (int cObject = 0; cObject < m_datasetSize; cObject++)
  {
		createObject(cObject, 0, 0.0, m_maxX, 0.0, m_maxY, m_initialDistribution, false);
    if (cObject % 10000 == 0) cerr << cObject << endl;
  }
  m_lastId = m_datasetSize;

  for (int Tnow = 1; Tnow < m_simulationLength; Tnow++)
  {
    cerr << "Time: " << Tnow;

    int cTotalUpdates = 0;
    int cNeedToUpdate = updatesPerInstant;
    set<int> updated;
    set<int>::iterator itUpdateArray = m_updateArray[Tnow].begin();

    while (cNeedToUpdate > 0 || itUpdateArray != m_updateArray[Tnow].end())
    {
      int id;
      bool bKilled = false;
      if (itUpdateArray != m_updateArray[Tnow].end())
      {
      	bKilled = true;
      	id = *itUpdateArray;
      	itUpdateArray++;
      }
      else
      {
				id = nextUniformInt(0, m_datasetSize);
				set<int>::iterator itUpdated = updated.find(id);

				while (itUpdated != updated.end())
				{
					id = nextUniformInt(0, m_datasetSize);
					itUpdated = updated.find(id);
				}
      }
      updated.insert(id);
      cNeedToUpdate--;
      cTotalUpdates++;

      map<int,MovingObject*>::iterator itDataset = m_dataset.find(id);
      assert(itDataset != m_dataset.end());
      MyMovingObject* o = static_cast<MyMovingObject*>((*itDataset).second);
			if (o->m_kt >= 0) m_updateArray[o->m_kt].erase(o->m_id);

			Function& f = o->m_functions.back();
      f.m_end = Tnow;

			assert(o->getX(Tnow) <= m_maxX + 1.0 && o->getY(Tnow) <= m_maxY + 1.0 && o->getX(Tnow) >= -1.0 && o->getY(Tnow) >= -1.0);

      if (bKilled && o->m_outOfBounds)
      {
        createObject(o->m_id, Tnow, 0.0, m_maxX, 0.0, m_maxY, UNIFORM, true);
      }
      else
      {
      	createObject(o->m_id, Tnow, o->getX(Tnow), o->getY(Tnow), false);
      }
    }

    cerr << ", Updates: " << cTotalUpdates << endl;
  }

	for (map<int, MovingObject*>::iterator itDataset = m_dataset.begin(); itDataset != m_dataset.end(); itDataset++)
	{
		MyMovingObject* o = static_cast<MyMovingObject*>((*itDataset).second);
    printObject(o);
	}
}
