// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "RoadGenerator.h"

RoadGenerator::RoadGenerator(int ds, int sl, int mui, double a, string filename)
	: Generator(ds, sl, mui, a)
{
	FILE* pRoads = fopen(filename.c_str(), "r");
	if (pRoads == NULL)
	{
		cerr << "Cannot open " << filename << endl;
		exit(-1);
	}

	char buf[1000];
	int fid = 0;
	int sid = 0;

	fgets(buf, 1000, pRoads);

	while (! feof(pRoads))
	{
		Path road;

		while(! feof(pRoads) && strcmp(buf, "\n") != 0)
		{
			float x1, x2, y1, y2;
			sscanf(buf, "%f %f %f %f", &x1, &y1, &x2, &y2);
			Segment s(sid, x1, y1, x2, y2);
			road.m_segments.push_back(s);
			fgets(buf, 1000, pRoads);
			sid++;
		}

		road.m_id = fid;
		m_roads.push_back(road);

		fgets(buf, 1000, pRoads);
		fid++;
	}

	m_updateArray = new set<int>[m_simulationLength];
}

void RoadGenerator::generate()
{
	m_dataset.clear();

	for (int cIndex = 0; cIndex < m_simulationLength; cIndex++)
	{
		m_updateArray[cIndex].clear();
	}

	int updatesPerInstant = (int) ceil(((double) m_datasetSize) * m_agility);

	// initialize all objects at random positions.
	for (int cObject = 0; cObject < m_datasetSize; cObject++)
	{
		MyMovingObject* o = new MyMovingObject();

		o->m_id = cObject;

		int cRoad = nextUniformInt(0, m_roads.size());
		Path& road = m_roads[cRoad];
		int cSegment = nextUniformInt(0, road.m_segments.size());
		Segment& s = road.m_segments[cSegment];

		o->m_roadIndex = cRoad;
		o->m_segmentIndex = cSegment;
		o->m_segmentId = s.m_id;

		if (m_initialDistribution == ZIPF)
		{
			o->m_sx = nextZipfDouble(s.m_x1, s.m_x2, 1.0);
		}
		else if (m_initialDistribution == GAUSSIAN)
		{
			o->m_sx = nextGaussianDouble(s.m_x1, s.m_x2);
		}
		else
		{
			o->m_sx = nextUniformDouble(s.m_x1, s.m_x2);
		}
		o->m_sy = s.m_slope * (o->m_sx - s.m_x1) + s.m_y1;
		o->m_st = 0;

		o->m_v = generateSpeed();
		if (flipCoin()) o->m_v *= -1.0;

		o->m_vx = cos(s.m_angle) * o->m_v;
		o->m_vy = sin(s.m_angle) * o->m_v;

		o->m_x = sqrt(pow(o->m_sx - s.m_x1, 2.0) + pow(o->m_sy - s.m_y1, 2.0));
		o->m_a = o->m_x - o->m_v * o->m_st;

		double ax[2] = {o->m_sx - (o->m_vx * o->m_st), o->m_vx};
		double ay[2] = {o->m_sy - (o->m_vy * o->m_st), o->m_vy};
		Polynomial px = Polynomial(2, ax);
		Polynomial py = Polynomial(2, ay);
		Function f = Function(o->m_st, -1, px, py);
		o->m_functions.push_back(f);

		m_dataset.insert(pair<int, MovingObject*>(o->m_id, o));

		o->m_goingTowardsX2 = false;
		double destination = (o->m_v > 0.0) ? max(s.m_x1, s.m_x2) : min(s.m_x1, s.m_x2);
		if (destination >= s.m_x2 - DOUBLE_EPSILON && destination <= s.m_x2 + DOUBLE_EPSILON) o->m_goingTowardsX2 = true;

		int t1 = o->m_st + m_maximumUpdateInterval;

		int t2;
		if (o->m_goingTowardsX2) t2 = (int) floor(o->m_st + (s.m_x2 - o->m_sx) / o->m_vx);
		else t2 = (int) floor(o->m_st + (s.m_x1 - o->m_sx) / o->m_vx);

		int t;
		if (t1 < t2)
		{
			t = t1;
			o->m_willReachEndOfSegment = false;
		}
		else
		{
			t = t2;
			o->m_willReachEndOfSegment = true;
		}
		if (t == 0) t = 1;
		assert(t >= 0);
		if (t < m_simulationLength)
		{
			m_updateArray[t].insert(o->m_id);
			o->m_kt = t;
		}

		if (cObject % 10000 == 0) cerr << cObject << endl;
	}

	int nextID = 0;

	for (int Tnow = 1; Tnow < m_simulationLength; Tnow++)
	{
		cerr << "Time: " << Tnow;

		int cTotalUpdates = 0;
		int cNeedToUpdate = updatesPerInstant;
		set<int> updated;

		if (! m_updateArray[Tnow].empty())
		{
			for (set<int>::iterator itUpdateArray = m_updateArray[Tnow].begin(); itUpdateArray != m_updateArray[Tnow].end(); itUpdateArray++)
			{
				map<int, MovingObject*>::iterator itDataset = m_dataset.find(*itUpdateArray);
				assert(itDataset != m_dataset.end());
				MyMovingObject* o = static_cast<MyMovingObject*>((*itDataset).second);
				cNeedToUpdate--;
				cTotalUpdates++;
				updated.insert(*itUpdateArray);

				Function& f = o->m_functions.back();
				f.m_end = Tnow;

				// change freeway only at line segment end-points.

				if (! o->m_willReachEndOfSegment)
				{
					o->m_sx = o->getX(Tnow);
					o->m_sy = o->getY(Tnow);
					o->m_v = (o->m_v > 0.0) ? generateSpeed() : generateSpeed() * -1.0;
				}
				else if (! o->m_goingTowardsX2 && o->m_segmentIndex != 0 /*&& nextUniformDouble() <= 0.9*/)
				{
					o->m_segmentIndex -= 1;
					Segment& s = m_roads[o->m_roadIndex].m_segments[o->m_segmentIndex];
					o->m_sx = s.m_x2;
					o->m_sy = s.m_y2;
					o->m_v = generateSpeed();
					if (s.m_x2 > s.m_x1) o->m_v *= -1.0;
				}
				else if (o->m_goingTowardsX2 && o->m_segmentIndex != (int) m_roads[o->m_roadIndex].m_segments.size() - 1 /*&& nextUniformDouble() <= 0.9*/)
				{
					o->m_segmentIndex += 1;
					Segment& s = m_roads[o->m_roadIndex].m_segments[o->m_segmentIndex];
					o->m_sx = s.m_x1;
					o->m_sy = s.m_y1;
					o->m_v = generateSpeed();
					if (s.m_x1 > s.m_x2) o->m_v *= -1.0;
				}
				else
				{
					for (unsigned int cIndex = 0; cIndex < o->m_functions.size(); cIndex++)
					{
						Function& f = o->m_functions[cIndex];

						cout << nextID << " " << f.m_start << " "  << f.m_end << " "	<< f.m_x.m_cLength << " ";

						for (int cCoef = 0; cCoef < f.m_x.m_cLength; cCoef++)
						{
							cout << f.m_x.m_pCoefficients[cCoef] << " ";
						}

						cout << f.m_y.m_cLength << " ";
						for (int cCoef = 0; cCoef < f.m_y.m_cLength; cCoef++)
						{
							cout << f.m_y.m_pCoefficients[cCoef] << " ";
						}

						cout << "0 0" << endl;
					}
					nextID++;
					cout << endl;
					o->m_functions.clear();

					int cRoad = nextUniformInt(0, m_roads.size());
					int cSegment = nextUniformInt(0, m_roads[cRoad].m_segments.size());
					Segment& s = m_roads[cRoad].m_segments[cSegment];

					o->m_roadIndex = cRoad;
					o->m_segmentIndex = cSegment;

					o->m_v = generateSpeed();
					if (flipCoin()) o->m_v *= -1.0;

					o->m_sx = nextUniformDouble(s.m_x1, s.m_x2);
					o->m_sy = s.m_slope * (o->m_sx - s.m_x1) + s.m_y1;
				}

				Segment& news = m_roads[o->m_roadIndex].m_segments[o->m_segmentIndex];
				o->m_st = Tnow;
				o->m_segmentId = news.m_id;
				o->m_x = sqrt(pow(o->m_sx - news.m_x1, 2.0) + pow(o->m_sy - news.m_y1, 2.0));
				o->m_a = o->m_x - o->m_v * o->m_st;
				o->m_vx = cos(news.m_angle) * o->m_v;
				o->m_vy = sin(news.m_angle) * o->m_v;

				double ax[2] = {o->m_sx - (o->m_vx * o->m_st), o->m_vx};
				double ay[2] = {o->m_sy - (o->m_vy * o->m_st), o->m_vy};
				Polynomial px = Polynomial(2, ax);
				Polynomial py = Polynomial(2, ay);
				Function f2 = Function(o->m_st, -1, px, py);
				o->m_functions.push_back(f2);

				o->m_goingTowardsX2 = false;
				double destination = (o->m_v > 0.0) ? max(news.m_x1, news.m_x2) : min(news.m_x1, news.m_x2);
				if (destination >= news.m_x2 - DOUBLE_EPSILON && destination <= news.m_x2 + DOUBLE_EPSILON) o->m_goingTowardsX2 = true;

				int t1 = o->m_st + m_maximumUpdateInterval;

				int t2;
				if (o->m_goingTowardsX2) t2 = (int) floor(o->m_st + (news.m_x2 - o->m_sx) / o->m_vx);
				else t2 = (int) floor(o->m_st + (news.m_x1 - o->m_sx) / o->m_vx);
				if (t2 == Tnow) t2++;

				int t;
				if (t1 < t2)
				{
					t = t1;
					o->m_willReachEndOfSegment = false;
				}
				else
				{
					t = t2;
					o->m_willReachEndOfSegment = true;
				}

				assert(t >= 0);
				if (t < m_simulationLength)
				{
					m_updateArray[t].insert(o->m_id);
					o->m_kt = t;
				}
			}
		}

		while (cNeedToUpdate > 0)
		{
			int id = nextUniformInt(0, m_datasetSize);
			set<int>::iterator itUpdated = updated.find(id);

			while (itUpdated != updated.end())
			{
				id = nextUniformInt(0, m_datasetSize);
				itUpdated = updated.find(id);
			}
			cNeedToUpdate--;
			cTotalUpdates++;
			updated.insert(id);

			map<int, MovingObject*>::iterator itDataset = m_dataset.find(id);
			MyMovingObject* o = static_cast<MyMovingObject*>((*itDataset).second);

			Function& f = o->m_functions.back();
			f.m_end = Tnow;

			// update speed only.

			o->m_sx = o->getX(Tnow);
			o->m_sy = o->getY(Tnow);

			o->m_v = (o->m_v > 0.0) ? generateSpeed() : generateSpeed() * -1.0;

			Segment& news = m_roads[o->m_roadIndex].m_segments[o->m_segmentIndex];
			o->m_st = Tnow;
			o->m_x = sqrt(pow(o->m_sx - news.m_x1, 2.0) + pow(o->m_sy - news.m_y1, 2.0));
			o->m_a = o->m_x - o->m_v * o->m_st;
			o->m_vx = cos(news.m_angle) * o->m_v;
			o->m_vy = sin(news.m_angle) * o->m_v;

			double ax[2] = {o->m_sx - (o->m_vx * o->m_st), o->m_vx};
			double ay[2] = {o->m_sy - (o->m_vy * o->m_st), o->m_vy};
			Polynomial px = Polynomial(2, ax);
			Polynomial py = Polynomial(2, ay);
			Function f2 = Function(o->m_st, -1, px, py);
			o->m_functions.push_back(f2);

			o->m_goingTowardsX2 = false;
			double destination = (o->m_v > 0.0) ? max(news.m_x1, news.m_x2) : min(news.m_x1, news.m_x2);
			if (destination >= news.m_x2 - DOUBLE_EPSILON && destination <= news.m_x2 + DOUBLE_EPSILON) o->m_goingTowardsX2 = true;

			int t1 = o->m_st + m_maximumUpdateInterval;

			int t2;
			if (o->m_goingTowardsX2) t2 = (int) floor(o->m_st + (news.m_x2 - o->m_sx) / o->m_vx);
			else t2 = (int) floor(o->m_st + (news.m_x1 - o->m_sx) / o->m_vx);
			if (t2 == Tnow) t2++;
			int t;
			if (t1 < t2)
			{
				t = t1;
				o->m_willReachEndOfSegment = false;
			}
			else
			{
				t = t2;
				o->m_willReachEndOfSegment = true;
			}

			assert(t >= 0);
			if (t < m_simulationLength)
			{
				m_updateArray[t].insert(o->m_id);
				o->m_kt = t;
			}
		}

		cerr << ", Updates: " << cTotalUpdates << endl;
	}

	for (map<int, MovingObject*>::iterator itDataset = m_dataset.begin(); itDataset != m_dataset.end(); itDataset++)
	{
		MyMovingObject* o = static_cast<MyMovingObject*>((*itDataset).second);

		for (unsigned int cIndex = 0; cIndex < o->m_functions.size(); cIndex++)
		{
			Function& f = o->m_functions[cIndex];
			if (f.m_end == -1) f.m_end = m_simulationLength;

			cout << nextID << " " << f.m_start << " "  << f.m_end << " "	<< f.m_x.m_cLength << " ";

			for (int cCoef = 0; cCoef < f.m_x.m_cLength; cCoef++)
			{
				cout << f.m_x.m_pCoefficients[cCoef] << " ";
			}

			cout << f.m_y.m_cLength << " ";
			for (int cCoef = 0; cCoef < f.m_y.m_cLength; cCoef++)
			{
				cout << f.m_y.m_pCoefficients[cCoef] << " ";
			}

			cout << "0 0" << endl;
		}
		nextID++;
		cout << endl;
	}
}

