// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#ifndef _global_h
#define _global_h

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <math.h>
#include <assert.h>

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <sstream>
#include <vector>
#include <map>
#include <set>
#include <stack>

#define DOUBLE_POSITIVE_INFINITY std::numeric_limits<double>::max()
#define DOUBLE_NEGATIVE_INFINITY -std::numeric_limits<double>::max()
#define DOUBLE_EPSILON std::numeric_limits<double>::epsilon()
#define FLOAT_POSITIVE_INFINITY std::numeric_limits<float>::max()
#define FLOAT_NEGATIVE_INFINITY -std::numeric_limits<float>::max()
#define FLOAT_EPSILON std::numeric_limits<float>::epsilon()

using namespace std;

#define DELETE 1
#define INSERT 2
#define QUERY 3

#define GAUSSIAN 0
#define UNIFORM 1
#define ZIPF 2

#define TMAX 100
#define XMAX 1.0
#define YMAX 1.0

class Point
{
public:
  double m_x;
  double m_y;
};

class Segment
{
public:
  double m_x1, m_x2, m_y1, m_y2;
  double m_slope, m_angle;
  int m_id;
  bool m_isInitialized;

  Segment()
  {
    m_x1 = m_x2 = m_y1 = m_y2 = 0.0;
    m_isInitialized = false;
  }

  Segment(int id, double x1, double y1, double x2, double y2)
  {
    // avoid horizontal or vertical segments.
    //assert(x1 != x2 && y1 != y2);

    this->m_x1 = x1;
    this->m_x2 = x2;
    this->m_y1 = y1;
    this->m_y2 = y2;

    this->m_id = id;

    if (m_x2 !=  m_x1) m_slope = (m_y2 - m_y1) / (m_x2 - m_x1);
    else m_slope = (m_y2 - m_y1) / 0.0000001;

    m_angle = atan(m_slope);

    m_isInitialized = true;
  }

  bool isInitialized()
  {
    return m_isInitialized;
  }
};

class Path
{
public:
  vector<Segment> m_segments;
  int m_id;
};

class Polynomial
{
public:
  int m_cLength;
  double* m_pCoefficients;

  Polynomial() : m_cLength(0), m_pCoefficients(0) {}
  Polynomial(int cLength, double* pCoefficients) : m_cLength(cLength)
  {
    m_pCoefficients = new double[m_cLength];
    memcpy(m_pCoefficients, pCoefficients, m_cLength * sizeof(double));
  }

  virtual ~Polynomial() { if (m_pCoefficients != 0) delete[] m_pCoefficients; }

  Polynomial(const Polynomial& p)
  {
    m_cLength = p.m_cLength;
    m_pCoefficients = new double[m_cLength];
    memcpy(m_pCoefficients, p.m_pCoefficients, m_cLength * sizeof(double));
  }

  Polynomial& operator=(const Polynomial& p)
  {
    if (this != &p)
    {
      if (m_cLength != p.m_cLength)
      {
        if (m_pCoefficients != 0) delete[] m_pCoefficients;
        m_cLength = p.m_cLength;
        m_pCoefficients = new double[m_cLength];
      }
      memcpy(m_pCoefficients, p.m_pCoefficients, m_cLength * sizeof(double));
    }
    return *this;
  }

  int getDegree() const { return m_cLength - 1; }

  double getValue(double t) const
  {
    double ret = 0.0;
    
    for (int cCoefficient = 0; cCoefficient < m_cLength; cCoefficient++)
    {
      ret += std::pow(t, (double) cCoefficient) * m_pCoefficients[cCoefficient];
    }

    return ret;
  }
};

class Function
{
public:
	int m_start;
	int m_end;
	Polynomial m_x;
	Polynomial m_y;

	Function(int start, int end, Polynomial& x, Polynomial& y)
	:  m_start(start), m_end(end), m_x(x), m_y(y) {}
};

class MovingObject
{
public:
  int m_id;
  int m_st;
  int m_kt;
};

extern void initRandom();
extern long nextUniformLong(long, long);
extern int nextUniformInt(int, int);
extern double nextUniformDouble();
extern double nextUniformDouble(double, double);
extern double nextZipfDouble(double, double, double);
extern double nextGaussianDouble();
extern double nextGaussianDouble(double, double);
extern bool flipCoin();

extern ostream& operator<<(ostream& os, const Segment& s);
extern ostream& operator<<(ostream& os, const Polynomial& p);
extern ostream& operator<<(ostream& os, const MovingObject& o);

#endif

