// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

int main(int argc, char** argv)
{
  std::vector<Segment> roads;

  int ds = 500, sl = 100;
  double a = 0.01;
  double extent = 1.0;
  char s[1000] = "road1.txt";
  int dt = 2;

  for (int i = 1; i < argc; i++)
  {
    if (! strcmp(argv[i], "-ds"))
    {
      i++;
      if (i >= argc)
      {
        cerr << "Missing dataset size." << endl;
        return -1;
      }
      ds = atoi(argv[i]);
    }
    else if (! strcmp(argv[i], "-e"))
    {
      i++;
      if (i >= argc)
      {
        cerr << "Missing universe extent." << endl;
        return -1;
      }
      extent = atof(argv[i]);
    }
    else if (! strcmp(argv[i], "-rf"))
    {
      i++;
      if (i >= argc)
      {
        cerr << "Missing road file." << endl;
        return -1;
      }
      strcpy(s, argv[i]);
    }
    else if (! strcmp(argv[i], "-sl"))
    {
      i++;
      if (i >= argc)
      {
        cerr << "Missing simulation length." << endl;
        return -1;
      }
      sl = atoi(argv[i]);
    }
    else if (! strcmp(argv[i], "-dt"))
    {
      i++;
      if (i >= argc)
      {
        cerr << "Missing dt." << endl;
        return -1;
      }
      dt = atoi(argv[i]);
    }
    else if (! strcmp(argv[i], "-a"))
    {
      i++;
      if (i >= argc)
      {
        cerr << "Missing alpha." << endl;
        return -1;
      }
      a = atof(argv[i]);
    }
    else
    {
      cerr << "Usage: " << endl
           << "  -ds dataset size" << endl
           << "  -sl simulation length" << endl
           << "  -dt query time interval" << endl
           << "  -a query extent" << endl
           << "  -e universe extent" << endl
           << "  -rf road file" << endl;
      return -1;
    }
  }

  initRandom();

	ifstream fin(s);
	if (! fin)
	{
		cerr << "Cannot open file " << s << "." << endl;
		return -1;
	}

	char buf[1000];
	int fid = 0;
	int sid = 0;

	while (fin)
	{
		while (fin.getline(buf, 1000) && strlen(buf) == 0); // eat up newlines

		while (fin && strlen(buf) != 0)
		{
			float x1, x2, y1, y2;
			sscanf(buf, "%f %f %f %f", &x1, &y1, &x2, &y2);
			Segment s(sid, x1, y1, x2, y2);
			roads.push_back(s);
			fin.getline(buf, 1000);
			sid++;
		}

		fid++;
	}

  double dx = a * extent / 2.0;
  double dy = a * extent / 2.0;

	cout.precision(8);
	cout << fixed;

  for (int cCount = 0; cCount < ds; cCount++)
 	{
    int cSeg = nextUniformInt(0, roads.size() - 1);
 	  Segment& seg = roads[cSeg];

    double x = nextUniformDouble(seg.m_x1, seg.m_x2);
 	  double y = seg.m_slope * (x - seg.m_x1) + seg.m_y1;
    int t1 = nextUniformInt(0, sl);

		cout << cCount << " " << x - dx << " " << x + dx << " "
			<< y - dy << " " << y + dy << " " << t1 << " " << std::min(t1 + dt - 1, sl) << endl;
  }
}
