// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

// for internal use
float combine_time( struct timeval* ); // return the combined time in second
void  subtract_timeval( const struct timeval* left, const struct timeval* right, struct timeval* result ); 

void initRandom()
{
  srand48(time(NULL));
}

long nextUniformLong(long low, long high)
{
  return (long) floor(((double) low + ((double) (high - low)) * drand48()));
}

int nextUniformInt(int low, int high)
{
  return (int) floor(((double) low + ((double) (high - low)) * drand48()));
}

double nextUniformDouble(double low, double high)
{
  return (low + ((high - low) * drand48()));
}

double nextUniformDouble()
{
  return drand48();
}

// mean 0.0, standard deviation 1.0
double nextGaussianDouble()
{
  static bool haveNextNextGaussian = false;
  static double nextNextGaussian;

  if (haveNextNextGaussian) {
    haveNextNextGaussian = false;
    return nextNextGaussian;
  } else {
    double v1, v2, s;

    do {
      v1 = 2 * drand48() - 1;   // between -1.0 and 1.0
      v2 = 2 * drand48() - 1;   // between -1.0 and 1.0
      s = v1 * v1 + v2 * v2;
    } while (s >= 1 || s == 0);

    double multiplier = std::sqrt(-2 * std::log(s)/s);
    nextNextGaussian = v2 * multiplier;
    haveNextNextGaussian = true;
    return v1 * multiplier;
  }
}

double nextZipfDouble(double low, double high, double p)
{
  double x;
  int i;
  double r, HsubV, sum;
  int V = 100;

  /* calculate the V-th harmonic number HsubV. WARNING: V>1 */
  HsubV = 0.0;
  for(i = 1; i <= V; i++) HsubV += 1.0 / pow((double) i, p);

  r = nextUniformDouble() * HsubV;
  sum = 1.0; i=1;
  while(sum < r)
  {
    i++;
    sum += 1.0/ pow((double)i, p);
  }

  /* i follows Zipf distribution and lies between 1 and V */
  /* x lies between 0. and 1. and then between low and high */
  x = ((double) i - 1.0) / ((double) V - 1.0);
  x = (high - low) * x + low;

  return x;
}

double nextGaussianDouble(double m, double std)
{
  return m + (std * nextGaussianDouble());
}

bool flipCoin()
{
  if (nextUniformDouble() >= 0.5) return true;
  return false;
}

int begin_cpu_usage( cpu_usage_aid* a) {
  int status;
  struct timezone dummy;

  status = getrusage( RUSAGE_SELF, &(a->my_rusage) );
  if ( status != 0 ) return 0;
  status = gettimeofday( &(a->my_timeval), &dummy );
  if ( status != 0 ) return 0;
  return 1;
}

int end_cpu_usage( const cpu_usage_aid* a, cpu_usage* c) {
  int status;
  struct timezone dummy;
  cpu_usage_aid a2;
  struct timeval dif_timeval;
  
  status = getrusage( RUSAGE_SELF, &(a2.my_rusage) );
  if ( status != 0 ) return 0;
  status = gettimeofday( &(a2.my_timeval), &dummy );
  if ( status != 0 ) return 0;

  // total_time

  subtract_timeval( &(a2.my_timeval), &(a->my_timeval), &dif_timeval );  
  c->total_time = dif_timeval.tv_sec + dif_timeval.tv_usec*1.0/1000000;
  
  // system_time
  
  subtract_timeval( &(a2.my_rusage.ru_stime), &(a->my_rusage.ru_stime), &dif_timeval );
  c->system_time = combine_time( &dif_timeval );

  // user_time

  subtract_timeval( &(a2.my_rusage.ru_utime), &(a->my_rusage.ru_utime), &dif_timeval );
  c->user_time = combine_time( &dif_timeval );

  // num_read_IOs, num_write_IOs

  c->num_read_IOs = a2.my_rusage.ru_inblock - a->my_rusage.ru_inblock;
  c->num_write_IOs = a2.my_rusage.ru_oublock - a->my_rusage.ru_oublock;
  return 1;
}

float combine_time( struct timeval* t ) {
  return t->tv_sec + t->tv_usec*1.0/1000000;
}

void  subtract_timeval( const struct timeval* a, const struct timeval* b, struct timeval* result ) {
  (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;                             
  (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;                          
  if ((result)->tv_usec < 0) {                                              
    --(result)->tv_sec;                                                     
    (result)->tv_usec += 1000000;                                           
  }                                                 
}

void MovingObject::toPointArray(vector<Point>& c)
{
	c.clear();

	for (unsigned int cIndex = 0; cIndex < m_functions.size(); cIndex++)
	{
		for (int T = m_functions[cIndex].m_start; T < m_functions[cIndex].m_end; T++)
		{
			c.push_back(Point(T, m_functions[cIndex].m_x.getValue(T), m_functions[cIndex].m_y.getValue(T)));
		}

		if (cIndex == m_functions.size() - 1)
		{
			int T = m_functions[cIndex].m_end;
			c.push_back(Point(T, m_functions[cIndex].m_x.getValue(T), m_functions[cIndex].m_y.getValue(T)));
		}
	}

/*
	for (unsigned int i = 0; i < c.size(); i++)
	{
		cerr << c[i].m_t << " " << c[i].m_x << " " << c[i].m_y << endl;
	}
*/
}

double MovingObject::getVolume()
{
	vector<Point> c;
	toPointArray(c);
	return volume(c, 0, c.size() - 1);
}

double MovingObject::getStartTime()
{
	return m_functions[0].m_start;
}

double MovingObject::getEndTime()
{
	return m_functions[m_functions.size() - 1].m_end;
}

Rectangle MovingObject::getMBR()
{
	vector<Point> c;
	toPointArray(c);

	double t1 = std::numeric_limits<double>::max();
	double x1 = std::numeric_limits<double>::max();
	double y1 = std::numeric_limits<double>::max();
	double t2 = -std::numeric_limits<double>::max();
	double x2 = -std::numeric_limits<double>::max();
	double y2 = -std::numeric_limits<double>::max();

	for (unsigned long i = 0; i < c.size() - 1; i++)
	{
		t1 = std::min(t1, (double) c[i].m_t);
		t2 = std::max(t2, (double) c[i].m_t);
		x1 = std::min(x1, c[i].m_x);
		x2 = std::max(x2, c[i].m_x);
		y1 = std::min(y1, c[i].m_y);
		y2 = std::max(y2, c[i].m_y);
	}

	return Rectangle(x1, x2, y1, y2, t1, t2);
}

bool parseObject(ifstream& f, MovingObject& o)
{
	o.m_functions.clear();
	o.m_id = -1;

	if (! f) return false;

	char buf[1000];
	while (f.getline(buf, 1000) && strlen(buf) == 0); // eat up newlines
	if (! f) return false;

	while (f && strlen(buf) != 0)
	{
		istringstream ss(buf);

		unsigned long start, end, coefsx, coefsy, id;

		ss >> id;
		if (! ss.good()) break;
		o.m_id = id;

		ss >> start >> end >> coefsx;

		double cx[coefsx];
		for (unsigned long i = 0; i < coefsx; i++)
		{
			ss >> cx[i];
		}

		ss >> coefsy;
		double cy[coefsy];
		for (unsigned long i = 0; i < coefsy; i++)
		{
			ss >> cy[i];
		}

		Polynomial px = Polynomial(coefsx, cx);
		Polynomial py = Polynomial(coefsy, cy);
		o.m_functions.push_back(Function(start, end, px, py));

		f.getline(buf, 1000);
	}

	return true;
}

double volume(vector<Point>& c, int s, int e)
{
	double t1 = std::numeric_limits<double>::max();
	double x1 = std::numeric_limits<double>::max();
	double y1 = std::numeric_limits<double>::max();
	double t2 = -std::numeric_limits<double>::max();
	double x2 = -std::numeric_limits<double>::max();
	double y2 = -std::numeric_limits<double>::max();

	for (int i = s; i <= e; i++)
	{
		t1 = std::min(t1, (double) c[i].m_t);
		t2 = std::max(t2, (double) c[i].m_t);
		x1 = std::min(x1, c[i].m_x);
		x2 = std::max(x2, c[i].m_x);
		y1 = std::min(y1, c[i].m_y);
		y2 = std::max(y2, c[i].m_y);
	}

	return (x2 - x1) * (y2 - y1) * (t2 - t1);
}

double rvolume(vector<Point>& c, int s)
{
	return volume(c, s, c.size() - 1);
}

double lvolume(vector<Point>& c, int e)
{
	return volume(c, 0, e);
}

class _globalMEntry
{
public:
	double g;
	double v;
	int l;
	int r;
	int m;
	
	_globalMEntry() : g(std::numeric_limits<double>::max()), v(std::numeric_limits<double>::max()), l(-1), r(-1), m(-1) {}
};

struct _globalMinGain : public binary_function<_globalMEntry, _globalMEntry, bool>
{
  bool operator()(_globalMEntry __x, _globalMEntry __y) const
  {
		if (__x.g < __y.g) return false;
		if (__x.g > __y.g) return true;
		if (__x.l <= __y.l) return false;
		return true;
  }
};

void greedySplit(MovingObject& o, SplitInformation& si)
{
	si.m_volumes.clear();
	si.m_splitPoints.clear();

	vector<Point> cuts;
	o.toPointArray(cuts);

	if (cuts.size() < 3)
	{
		si.m_volumes.push_back(rvolume(cuts, 0));
		si.m_splitPoints.push_back(vector<int>());
		si.m_id = o.m_id;
		return;
	}

	unsigned int sp = cuts.size() - 2;

	priority_queue<_globalMEntry, vector<_globalMEntry>, _globalMinGain> pq;

	priority_queue<unsigned int, vector<unsigned int>, greater<unsigned int> > a;
	priority_queue<unsigned int, vector<unsigned int>, greater<unsigned int> > s[sp + 1];

	for (unsigned int i = 1; i < cuts.size() - 1; i++)
	{
		double lv = volume(cuts, i - 1, i);
		double rv = volume(cuts, i, i + 1);
		double mv = volume(cuts, i - 1, i + 1);
		double g = mv - (lv + rv);

		_globalMEntry e;
		e.l = i - 1;
		e.r = i + 1;
		e.m = i;
		e.g = g;
		e.v = mv;

		pq.push(e);
	}

	s[sp] = a;

	for (unsigned int i = sp - 1; i > 0; i--)
	{
		_globalMEntry e = pq.top(); pq.pop();

		a.push(e.m);
		s[i] = a;

		vector<_globalMEntry> temp;

		bool f1 = false, f2 = false;
		while (! pq.empty())
		{
			_globalMEntry t = pq.top(); pq.pop();

			if (t.r == e.m)
			{
				t.r = e.r;
				t.m = e.l;
				t.g = volume(cuts, t.l, t.r) - (volume(cuts, t.l, e.l) + e.v);
				f1 = true;
			}
			else if (t.l == e.m)
			{
				t.l = e.l;
				t.m = e.r;
				t.g = volume(cuts, t.l, t.r) - (volume(cuts, e.r, t.r) + e.v);
				f2 = true;
			}
			temp.push_back(t);
			if (f1 && f2) break;
		}

		for (unsigned int cIndex = 0; cIndex < temp.size(); cIndex++) pq.push(temp[cIndex]);
	}

	si.m_id = o.m_id;
	si.m_volumes.push_back(rvolume(cuts, 0));
	si.m_splitPoints.push_back(vector<int>());

	for (unsigned int i = 1; i <= sp; i++)
	{
		vector<int> splits;
		for (unsigned int cSplit = 1; cSplit < cuts.size() - 1; cSplit++)
		{
			if (s[i].empty() || cSplit != s[i].top()) splits.push_back(cSplit);
			else s[i].pop();
		}

		double v = 0.0;
		int k = 0;
		for (unsigned int j = 0; j < splits.size(); j++)
		{
			int l = splits[j];
			v += volume(cuts, k, l);
			k = l;
		}
		v += rvolume(cuts, k);

		si.m_volumes.push_back(v);
		si.m_splitPoints.push_back(splits);
	}
}

class _globalDEntry
{
public:
	double g;
	int p;
	int k;
	
	_globalDEntry() : g(std::numeric_limits<double>::max()), p(-1), k(-1) {}
};

void dynamicSplit(MovingObject& o, SplitInformation& si)
{
	si.m_volumes.clear();
	si.m_splitPoints.clear();

	vector<Point> cuts;
	o.toPointArray(cuts);

	if (cuts.size() < 3)
	{
		si.m_volumes.push_back(rvolume(cuts, 0));
		si.m_splitPoints.push_back(vector<int>());
		si.m_id = o.m_id;
		return;
	}

	unsigned int sp = cuts.size() - 2;
	_globalDEntry** a;

	try
	{
		a = new _globalDEntry*[cuts.size()];
		for (unsigned int cIndex = 0; cIndex < cuts.size(); cIndex++)
		{
			a[cIndex] = new _globalDEntry[sp + 1];
		}
	}
	catch (...)
	{
		cerr << "Out of memory." << endl;
		exit(-1);
	}

	a[1][0].g = lvolume(cuts, 1);
	a[1][0].p = 0;

	//_globalDEntry le;

	for (unsigned int i = 2; i < cuts.size(); i++)
	{
		a[i][0].g = lvolume(cuts, i);

		//bool stop = false;

		for (unsigned int k = 1; k <= i - 1; k++)
		{
			_globalDEntry e;

/*
			if (stop)
			{
				a[i][k].g = le.g;
				a[i][k].p = le.p;
				a[i][k].k = le.k;
				continue;
			}
*/
			for (unsigned int p = 1; p < i; p++)
			{
				double f = a[p][std::min(k - 1, p - 1)].g + volume(cuts, p, i);

				if (f <= e.g)
				{
					e.g = f;
					e.p = p;
					e.k = k;
				}
			}

			a[i][k] = e;

/*
			// avoid some unnecessary computations, if volume gain is insignificant.
			if (a[i][k - 1].g - a[i][k].g < 0.000001)
			{
				stop = true;
				le.g = e.g;
				le.p = e.p;
				le.k = k;
			}
*/
		}
	}

	si.m_id = o.m_id;

	for (unsigned int k = 0; k <= sp; k++)
	{
		si.m_volumes.push_back(a[cuts.size() - 1][k].g);

		vector<int> splits;
		int i = cuts.size() - 1;
		int j = k;

		while (j > 0)
		{
			splits.push_back(a[i][j].p);
			int t = a[i][j].k - 1;
			i = a[i][j].p;
			j = t;
		}

		std::sort(splits.begin(), splits.end());
		si.m_splitPoints.push_back(splits);
	}

	for (unsigned int i = 0; i < cuts.size(); i++) delete[] a[i];
	delete[] a;
}

std::ostream& operator<<(std::ostream& os, const Point& p)
{
	os << p.m_t << ", " << p.m_x << ", " << p.m_y;

	return os;
}

std::ostream& operator<<(std::ostream& os, const Polynomial& p)
{
	os << p.m_cLength << " ";

	for (int i = 0; i < p.m_cLength - 1; i++)
	{
		os << p.m_pCoefficients[i] << " ";
	}
	os << p.m_pCoefficients[p.m_cLength - 1];

	return os;
}

std::ostream& operator<<(std::ostream& os, const Function& f)
{
	os << f.m_start << " " << f.m_end << " " << f.m_x << " " << f.m_y;
	return os;
}

std::ostream& operator<<(std::ostream& os, const MovingObject& o)
{
	for (unsigned int cIndex = 0; cIndex < o.m_functions.size(); cIndex++)
	{
		const Function& f = o.m_functions[cIndex];

		os << o.m_id << " " << f.m_start << " "  << f.m_end << " "	<< f.m_x.m_cLength << " ";

		for (int cCoef = 0; cCoef < f.m_x.m_cLength; cCoef++)
		{
			os << f.m_x.m_pCoefficients[cCoef] << " ";
		}

		os << f.m_y.m_cLength << " ";
		for (int cCoef = 0; cCoef < f.m_y.m_cLength; cCoef++)
		{
			os << f.m_y.m_pCoefficients[cCoef] << " ";
		}

		os << "0 0" << endl;
	}

	return os;
}

