// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#ifndef _GLOBAL_H
#define _GLOBAL_H

#include <assert.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <queue>
#include <stack>
#include <vector>
#include <map>

#include <fcntl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

using namespace std;

#define POSITIVE_INFINITY std::numeric_limits<float>::max()
#define NEGATIVE_INFINITY -POSITIVE_INFINITY

#define DELETE 0
#define INSERT 1
#define QUERY 2

#define PRECISION 8

#define DYNAMIC_SPLIT 0
#define GREEDY_SPLIT 1

#define TMAX 100.0
#define XMAX 1.0
#define YMAX 1.0

#define MAXIMUM_SPLITS_KEPT 101

typedef struct {
  struct rusage my_rusage;
  struct timeval my_timeval;
} cpu_usage_aid;

typedef struct {
  float total_time; // including IO time.
  float system_time;
  float user_time;
  long num_read_IOs;
  long num_write_IOs;
} cpu_usage;

// The following two functions should be called in pair to report the cpu usage of a procedure.
// The first one is called before the procedure. It records some information in c, which is
// passed to the second one. The second one is called after the procedure.
// returns 0 - failure;  1 - success
int begin_cpu_usage( cpu_usage_aid* c);
int end_cpu_usage( const cpu_usage_aid*, cpu_usage* );

void initRandom();
long nextUniformLong(long, long);
int nextUniformInt(int, int);
double nextUniformDouble();
double nextUniformDouble(double, double);
double nextZipfDouble(double, double, double);
double nextGaussianDouble();
double nextGaussianDouble(double, double);
bool flipCoin();

class Point
{
public:
	int m_t;
	double m_x, m_y;

	Point(int t, double x, double y) : m_t(t), m_x(x), m_y(y) {}
};

class Rectangle
{
public:
	Rectangle(double x1, double x2, double y1, double y2, double t1, double t2)
		: m_xlow(x1), m_xhigh(x2), m_ylow(y1), m_yhigh(y2), m_tlow(t1), m_thigh(t2)
	{
		assert(x1 <= x2);
		assert(y1 <= y2);
		assert(t1 <= t2);
	}

	bool intersects(Rectangle& b)
	{
		if (m_xlow > b.m_xhigh || m_xhigh < b.m_xlow ||
				m_ylow > b.m_yhigh || m_yhigh < b.m_ylow) return false;

		return true;
	}

	bool intersectsInTime(Rectangle& r)
	{
		//if (m_startTime != r.m_startTime && (m_endTime <= r.m_startTime || m_startTime >= r.m_endTime)) return false;
		if (m_thigh <= r.m_tlow || m_tlow >= r.m_thigh) return false;
		return intersects(r);
	}

	double getMinDist(const Rectangle& r)
	{
		double ret = 0.0;

		if (r.m_xlow < m_xlow)
			ret += pow(m_xlow - r.m_xlow, 2.0);
		else if (r.m_xlow > m_xhigh)
			ret += pow(r.m_xlow - m_xhigh, 2.0);

		if (r.m_ylow < m_ylow)
			ret += pow(m_ylow - r.m_ylow, 2.0);
		else if (r.m_ylow > m_yhigh)
			ret += pow(r.m_ylow - m_yhigh, 2.0);

		return ret;
	}

	double getIntersectingArea(Rectangle& b)
	{
		if (m_xlow > b.m_xhigh || m_xhigh < b.m_xlow ||
				m_ylow > b.m_yhigh || m_yhigh < b.m_ylow) return 0.0;

		double f1 = std::max(m_xlow, b.m_xlow);
		double f2 = std::min(m_xhigh, b.m_xhigh);
		double f3 = std::max(m_ylow, b.m_ylow);
		double f4 = std::min(m_yhigh, b.m_yhigh);

		return (f2 - f1) * (f4 - f3);
	}

	double getArea()
	{
		return (m_xhigh - m_xlow) * (m_yhigh - m_ylow);
	}

	double m_xlow, m_xhigh, m_ylow, m_yhigh, m_tlow, m_thigh;
};


class Polynomial
{
public:
	int m_cLength;
	double* m_pCoefficients;

	Polynomial() : m_cLength(0), m_pCoefficients(0) {}
	Polynomial(int cLength, double* pCoefficients) : m_cLength(cLength)
	{
		m_pCoefficients = new double[m_cLength];
		memcpy(m_pCoefficients, pCoefficients, m_cLength * sizeof(double));
	}

	virtual ~Polynomial() { if (m_pCoefficients != 0) delete[] m_pCoefficients; }

	Polynomial(const Polynomial& p)
	{
		m_cLength = p.m_cLength;
		m_pCoefficients = new double[m_cLength];
		memcpy(m_pCoefficients, p.m_pCoefficients, m_cLength * sizeof(double));
	}

	Polynomial& operator=(const Polynomial& p)
	{
		if (this != &p)
		{
			if (m_cLength != p.m_cLength)
			{
				if (m_pCoefficients != 0) delete[] m_pCoefficients;
				m_cLength = p.m_cLength;
				m_pCoefficients = new double[m_cLength];
			}
			memcpy(m_pCoefficients, p.m_pCoefficients, m_cLength * sizeof(double));
		}
		return *this;
	}

	int getDegree() const { return m_cLength - 1; }

	double getValue(double t) const
	{
		double ret = 0.0;

		for (int cCoefficient = 0; cCoefficient < m_cLength; cCoefficient++)
		{
			ret += std::pow(t, (double) cCoefficient) * m_pCoefficients[cCoefficient];
		}

		return ret;
	}
};

class Function
{
public:
	int m_start;
	int m_end;
	Polynomial m_x;
	Polynomial m_y;

	Function(int start, int end, Polynomial& x, Polynomial& y)
	:  m_start(start), m_end(end), m_x(x), m_y(y) {}
};

class MovingObject
{
public:
	void toPointArray(vector<Point>& c);
	double getVolume();
	double getStartTime();
	double getEndTime();
	Rectangle getMBR();
	int m_id;
	vector<Function> m_functions;
};

class SplitInformation
{
public:
	int getTotalSplits() { return m_volumes.size() - 1; }

	int m_id;
	vector<double> m_volumes;
	vector<vector<int> > m_splitPoints;
};

std::ostream& operator<<(std::ostream& os, const Point& p);
std::ostream& operator<<(std::ostream& os, const Polynomial& p);
std::ostream& operator<<(std::ostream& os, const Function& f);
std::ostream& operator<<(std::ostream& os, const MovingObject& o);

double volume(vector<Point>& c, int s, int e);
bool parseObject(ifstream& f, MovingObject& o);
void greedySplit(MovingObject& o, SplitInformation& si);
void dynamicSplit(MovingObject& o, SplitInformation& si);

#endif
