// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

/*
	Assigns a number of splits per data trajectory, using greedy programming to
	minimize the total volume of the resulting boxes.
	It uses a pre-calculated split file containing all possible splits per data trajectory.
*/

#include "global.h"

class Entry
{
public:
	int m_assignedSplits;
	int m_totalSplits;
	int m_id;
	float* m_volumes;

	Entry(SplitInformation& si);
	Entry(const Entry& e);
	~Entry() { delete[] m_volumes; }

	Entry& operator=(const Entry& e);

	float gainFromLastSplit();
	float gainWithOneMoreSplit();
};

Entry:: Entry(SplitInformation& si)
	:	m_assignedSplits(0),
		m_totalSplits(0),
		m_id(si.m_id)
{
	m_totalSplits = si.getTotalSplits() + 1; // plus the zero splits case.
	int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
	m_volumes = new float[s];
	for (int cIndex = 0; cIndex < s; cIndex++) m_volumes[cIndex] = (float) si.m_volumes[cIndex];
}

Entry::Entry(const Entry& e)
{
	m_assignedSplits = e.m_assignedSplits;
	m_totalSplits = e.m_totalSplits;
	m_id = e.m_id;

	int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
	m_volumes = new float[s];
	memcpy(m_volumes, e.m_volumes, s * sizeof(float));
}

Entry& Entry::operator=(const Entry& e)
{
	if (this != &e)
	{
		m_assignedSplits = e.m_assignedSplits;
		m_totalSplits = e.m_totalSplits;
		m_id = e.m_id;

		delete[] m_volumes;
		int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
		m_volumes = new float[s];
		memcpy(m_volumes, e.m_volumes, s * sizeof(float));
	}

	return *this;
}

float Entry::gainFromLastSplit()
{
	if (m_assignedSplits == 0) return numeric_limits<float>::max();
	return (m_volumes[m_assignedSplits - 1] - m_volumes[m_assignedSplits]);
}

float Entry::gainWithOneMoreSplit()
{
	if (m_assignedSplits >= MAXIMUM_SPLITS_KEPT)
	{
		cerr << "Increase the number of splits kept per object." << endl;
		exit(-1);
	}
	else if (m_assignedSplits >= m_totalSplits)
	{
		return 0.0f;
	}
	else
	{
		return m_volumes[m_assignedSplits] - m_volumes[m_assignedSplits + 1];
	}
}

struct maxGainWithOneMoreSplit : public binary_function<Entry*, Entry*, bool>
{
	bool operator()(Entry* __x, Entry* __y) const { return __x->gainWithOneMoreSplit() < __y->gainWithOneMoreSplit(); }
};

int main(int argc, char** argv)
{
	if (argc != 3)
	{
		cerr << "Usage: " << argv[0] << " split_file total_splits." << endl;
		return -1;
	}

	int splits = atoi(argv[2]);

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open file " << argv[1] << "." << endl;
		return -1;
	}

	priority_queue<Entry*, vector<Entry*>, maxGainWithOneMoreSplit> pq;
	char buf[1000];
	double totalVolume = 0.0;
	double skippedVolume = 0.0;

	while (fin.getline(buf, 1000))
	{
		while(fin && strlen(buf) == 0) fin.getline(buf, 1000);

		SplitInformation si;
		int totalSplits, currentSplit;
		float volume;

		while (fin && strlen(buf) != 0)
		{
			istringstream ss(buf);
			ss >> si.m_id >> totalSplits >> currentSplit >> volume;
			si.m_volumes.push_back(volume);
			fin.getline(buf, 1000);
		}

		totalVolume += si.m_volumes[0];

		if (si.getTotalSplits() == 0)
		{
			skippedVolume += si.m_volumes[0];
			cout << si.m_id << "\t0" << endl;
			continue;
		}

		Entry* e = new Entry(si);
		pq.push(e);
	}

	cpu_usage_aid cpu;
	begin_cpu_usage(&cpu);

	// distribute considering one split at a time.
	while (splits > 0)
	{
		Entry* e = pq.top(); pq.pop();
		//cerr << splits << " " << e.id << " " << e.gainWithOneMoreSplit() << endl;

		if (e->gainWithOneMoreSplit() <= 0.0f) break;
		e->m_assignedSplits++;
		pq.push(e);
		splits--;
	}

	cpu_usage cpuu;
	end_cpu_usage(&cpu, &cpuu);

	double greedy = 0.0;

	while (! pq.empty())
	{
		Entry* t = pq.top(); pq.pop();

		cout << t->m_id << "\t" << t->m_assignedSplits << endl;
		greedy += t->m_volumes[t->m_assignedSplits];
		delete t;
	}

	cerr << "Total: " << totalVolume << endl;
	cerr << "Greedy: " << greedy + skippedVolume << endl;
	cerr << "Gain: " << totalVolume - (greedy + skippedVolume) << endl;
	cerr << "Total time: " << cpuu.total_time << endl;
	cerr << "System time: " << cpuu.system_time << endl;
	cerr << "User time: " << cpuu.user_time << endl;

	return 0;
}
