// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

/*
	Assigns a number of splits per data trajectory, using look ahead greedy programming to
	minimize the total volume of the resulting boxes.
*/
#include "global.h"

class Entry
{
public:
	int m_assignedSplits;
	int m_totalSplits;
	int m_id;
	float* m_volumes;

	Entry(SplitInformation& si);
	Entry(const Entry& e);
	~Entry() { delete[] m_volumes; }

	Entry& operator=(const Entry& e);

	float gainFromLastSplit();
	float gainWithOneMoreSplit();
	float gainWithTwoMoreSplits();
	float test();
};

Entry:: Entry(SplitInformation& si)
	:	m_assignedSplits(0),
		m_totalSplits(si.getTotalSplits() + 1), // plus the zero splits case.
		m_id(si.m_id)
{
	int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
	m_volumes = new float[s];

	for (int cIndex = 0; cIndex < s; cIndex++) m_volumes[cIndex] = (float) si.m_volumes[cIndex];
}

Entry::Entry(const Entry& e)
{
	m_assignedSplits = e.m_assignedSplits;
	m_totalSplits = e.m_totalSplits;
	m_id = e.m_id;

	int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
	m_volumes = new float[s];
	memcpy(m_volumes, e.m_volumes, s * sizeof(float));
}

Entry& Entry::operator=(const Entry& e)
{
	if (this != &e)
	{
		m_assignedSplits = e.m_assignedSplits;
		m_totalSplits = e.m_totalSplits;
		m_id = e.m_id;

		delete[] m_volumes;
		int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
		m_volumes = new float[s];
		memcpy(m_volumes, e.m_volumes, s * sizeof(float));
	}

	return *this;
}

float Entry::gainFromLastSplit()
{
	if (m_assignedSplits == 0) return numeric_limits<float>::max();
	return (m_volumes[m_assignedSplits - 1] - m_volumes[m_assignedSplits]);
}

float Entry::gainWithOneMoreSplit()
{
	if (m_assignedSplits >= MAXIMUM_SPLITS_KEPT)
	{
		cerr << "Increase the number of splits kept per object." << endl;
		exit(-1);
	}
	else if (m_assignedSplits >= m_totalSplits)
	{
		return 0.0f;
	}
	else
	{
		return m_volumes[m_assignedSplits] - m_volumes[m_assignedSplits + 1];
	}
}

float Entry::gainWithTwoMoreSplits()
{
	if (m_assignedSplits >= MAXIMUM_SPLITS_KEPT - 1)
	{
		cerr << "Increase the number of splits kept per object." << endl;
		exit(-1);
	}
	else if (m_assignedSplits + 2 >= m_totalSplits)
	{
		return 0.0f;
	}
	else
	{
		return m_volumes[m_assignedSplits + 1] - m_volumes[m_assignedSplits + 2];
	}
}

float Entry::test()
{
		float f1 = gainWithOneMoreSplit();
		float f2 = gainWithTwoMoreSplits();
		return (f2 > f1) ? f2 : 0.0f;
}

struct maxGainWithOneMoreSplit : public binary_function<Entry*, Entry*, bool>
{
	bool operator()(Entry* __x, Entry* __y) const { return __x->gainWithOneMoreSplit() < __y->gainWithOneMoreSplit(); }
};

struct minGainFromLastSplit : public binary_function<Entry*, Entry*, bool>
{
	bool operator()(Entry* __x, Entry* __y) const { return __x->gainFromLastSplit() > __y->gainFromLastSplit(); }
};

struct maxGainWithTwoMoreSplits : public binary_function<Entry*, Entry*, bool>
{
	bool operator()(Entry* __x, Entry* __y) const
	{
		float xf1 = __x->gainWithOneMoreSplit();
		float xf2 = __x->gainWithTwoMoreSplits();
		float yf1 = __y->gainWithOneMoreSplit();
		float yf2 = __y->gainWithTwoMoreSplits();

		float x = (xf2 > xf1) ? xf2 : 0.0;
		float y = (yf2 > yf1) ? yf2 : 0.0;

		return x < y;
	}
};

int main(int argc, char** argv)
{
	if (argc != 3)
	{
		cerr << "Usage: " << argv[0] << " data_file total_splits." << endl;
		return -1;
	}

	int splits = atoi(argv[2]);

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open file " << argv[1] << "." << endl;
		return -1;
	}

	priority_queue<Entry*, vector<Entry*>, maxGainWithOneMoreSplit> pq;
	MovingObject o;

	cpu_usage_aid cpu;
	begin_cpu_usage(&cpu);

	double totalVolume = 0.0;
	double skippedVolume = 0.0;
	int count = 0;
	while (parseObject(fin, o))
	{
		count++;
		if (count % 10000 == 0) cerr << count << endl;

		SplitInformation si;
		greedySplit(o, si);

		totalVolume += si.m_volumes[0];

		if (si.getTotalSplits() == 0)
		{
			skippedVolume += si.m_volumes[0];
			cout << o.m_id << "\t0" << endl;
			continue;
		}

		Entry* e = new Entry(si);
		pq.push(e);
	}

	// distribute considering one split at a time.
	while (splits > 0)
	{
		Entry* e = pq.top(); pq.pop();
		//cerr << splits << " " << e.id << " " << e.gainWithOneMoreSplit() << endl;
		if (e->gainWithOneMoreSplit() <= 0.0f) break;
		e->m_assignedSplits++;
		pq.push(e);
		splits--;
	}

	// re-arrange according to minimum gain for the last split and maximum gain for two new splits.
	vector<Entry*> gainMin;
	vector<Entry*> gainMax;
	vector<Entry*>::iterator itMin;
	vector<Entry*>::iterator itMax;

	while (! pq.empty())
	{
		Entry* e = pq.top(); pq.pop();

		// smallest element last.
		itMin = upper_bound(gainMin.begin(), gainMin.end(), e, minGainFromLastSplit());
		gainMin.insert(itMin, e);

		// largest element last.
		itMax = upper_bound(gainMax.begin(), gainMax.end(), e, maxGainWithTwoMoreSplits());
		gainMax.insert(itMax, e);
	}

/*
	for (itMax = gainMax.begin(); itMax != gainMax.end(); itMax++)
	{
		cerr << (*itMax)->m_id << " " << (*itMax)->test() << endl;
	}

	cerr << endl;

	for (itMin = gainMin.begin(); itMin != gainMin.end(); itMin++)
	{
		cerr << (*itMin)->gainFromLastSplit() << endl;
	}
*/

	bool bContinue = true;
	int changes = 0;

	while (bContinue)
	{
		bContinue = false;

		// remove top element from gainMax list.
		Entry* max = gainMax.back(); gainMax.pop_back();
		if (max->gainWithTwoMoreSplits() <= 0.0f) break;

		// remove two top elements from gainMin list.
		Entry* min1 = gainMin.back(); gainMin.pop_back();
		Entry* min2 = gainMin.back(); gainMin.pop_back();

		// check if max is the same as mins and remove something else instead.
		if (max == min1)
		{
			Entry* t = min1;
			min1 = min2;
			min2 = gainMin.back(); gainMin.pop_back();
			gainMin.push_back(t);
		}
		else if (max == min2)
		{
			Entry* t = min2;
			min2 = gainMin.back(); gainMin.pop_back();
			gainMin.push_back(t);
		}

		assert(max != min1 && max != min2 && min1 != min2);

		// if it is worth rearranging...
		if (max->gainWithTwoMoreSplits() > min1->gainFromLastSplit() + min2->gainFromLastSplit())
		{
			changes++;
			bContinue = true;

			// locate max in gainMin list and remove it.
			itMin = lower_bound(gainMin.begin(), gainMin.end(), max, minGainFromLastSplit());
			while (*itMin != max)
			{
				assert(itMin != gainMin.end());
				itMin++;
			}
			gainMin.erase(itMin);

			// locate mins in gainMax list and remove them.
			itMax = lower_bound(gainMax.begin(), gainMax.end(), min1, maxGainWithTwoMoreSplits());
			while (*itMax != min1)
			{
				assert(itMax != gainMax.end());
				itMax++;
			}
			gainMax.erase(itMax);

			itMax = lower_bound(gainMax.begin(), gainMax.end(), min2, maxGainWithTwoMoreSplits());
			while (*itMax != min2)
			{
				assert(itMax != gainMax.end());
				itMax++;
			}
			gainMax.erase(itMax);

			// rearrange the splits.
			max->m_assignedSplits += 2;
			min1->m_assignedSplits--;
			min2->m_assignedSplits--;

			{
			// this block does not work correctly when using -O1 flag with g++ v3.2
			// gainMax list is sorted before the insertions, and ends up messed up after the insertions.
			// It works great in debug build or without -O1.

			// add max and mins in gainMax list.
			vector<Entry*>::iterator it1 = upper_bound(gainMax.begin(), gainMax.end(), min1, maxGainWithTwoMoreSplits());
			gainMax.insert(it1, min1);
			vector<Entry*>::iterator it2 = upper_bound(gainMax.begin(), gainMax.end(), min2, maxGainWithTwoMoreSplits());
			gainMax.insert(it2, min2);
			vector<Entry*>::iterator it3 = upper_bound(gainMax.begin(), gainMax.end(), max, maxGainWithTwoMoreSplits());
			gainMax.insert(it3, max);

			// add max and mins in gainMin list.
			vector<Entry*>::iterator it4 = upper_bound(gainMin.begin(), gainMin.end(), min1, minGainFromLastSplit());
			gainMin.insert(it4, min1);
			vector<Entry*>::iterator it5 = upper_bound(gainMin.begin(), gainMin.end(), min2, minGainFromLastSplit());
			gainMin.insert(it5, min2);
			vector<Entry*>::iterator it6 = upper_bound(gainMin.begin(), gainMin.end(), max, minGainFromLastSplit());
			gainMin.insert(it6, max);
			}
		}
	}

	cpu_usage cpuu;
	end_cpu_usage(&cpu, &cpuu);

	double lagreedy = 0.0;

	for (itMin = gainMin.begin(); itMin != gainMin.end(); itMin++)
	{
		cout << (*itMin)->m_id << "\t" << (*itMin)->m_assignedSplits << endl;
		lagreedy += (*itMin)->m_volumes[(*itMin)->m_assignedSplits];
		delete (*itMin);
	}

	cerr << "Number of changes: " << changes << endl;
	cerr << "Total: " << totalVolume << endl;
	cerr << "LAGreedy: " << lagreedy + skippedVolume << endl;
	cerr << "Gain: " << totalVolume - (lagreedy + skippedVolume) << endl;
	cerr << "Total time: " << cpuu.total_time << endl;
	cerr << "System time: " << cpuu.system_time << endl;
	cerr << "User time: " << cpuu.user_time << endl;

	return 0;
}
