// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

/*
	Assigns a number of splits per data trajectory, using look ahead greedy programming to
	minimize the total volume of the resulting boxes.
	It uses a pre-calculated split file containing all possible splits per data trajectory.
*/

#include "global.h"

class Entry
{
public:
	int m_assignedSplits;
	int m_totalSplits;
	int m_id;
	float* m_volumes;

	Entry(SplitInformation& si);
	Entry(const Entry& e);
	~Entry() { delete[] m_volumes; }

	Entry& operator=(const Entry& e);

	float gainFromLastSplit();
	float gainWithOneMoreSplit();
	float gainWithTwoMoreSplits();
};

Entry:: Entry(SplitInformation& si)
	:	m_assignedSplits(0),
		m_totalSplits(si.getTotalSplits() + 1), // plus the zero splits case.
		m_id(si.m_id)
{
	int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
	m_volumes = new float[s];

	for (int cIndex = 0; cIndex < s; cIndex++) m_volumes[cIndex] = (float) si.m_volumes[cIndex];
}

Entry::Entry(const Entry& e)
{
	m_assignedSplits = e.m_assignedSplits;
	m_totalSplits = e.m_totalSplits;
	m_id = e.m_id;

	int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
	m_volumes = new float[s];
	memcpy(m_volumes, e.m_volumes, s * sizeof(float));
}

Entry& Entry::operator=(const Entry& e)
{
	if (this != &e)
	{
		m_assignedSplits = e.m_assignedSplits;
		m_totalSplits = e.m_totalSplits;
		m_id = e.m_id;

		delete[] m_volumes;
		int s = min(m_totalSplits, MAXIMUM_SPLITS_KEPT);
		m_volumes = new float[s];
		memcpy(m_volumes, e.m_volumes, s * sizeof(float));
	}

	return *this;
}

float Entry::gainFromLastSplit()
{
	if (m_assignedSplits == 0) return numeric_limits<float>::max();
	return (m_volumes[m_assignedSplits - 1] - m_volumes[m_assignedSplits]);
}

float Entry::gainWithOneMoreSplit()
{
	if (m_assignedSplits >= MAXIMUM_SPLITS_KEPT)
	{
		cerr << "Increase the number of splits kept per object." << endl;
		exit(-1);
	}
	else if (m_assignedSplits >= m_totalSplits)
	{
		return 0.0f;
	}
	else
	{
		return m_volumes[m_assignedSplits] - m_volumes[m_assignedSplits + 1];
	}
}

float Entry::gainWithTwoMoreSplits()
{
	if (m_assignedSplits >= MAXIMUM_SPLITS_KEPT - 1)
	{
		cerr << "Increase the number of splits kept per object." << endl;
		exit(-1);
	}
	else if (m_assignedSplits + 2 >= m_totalSplits)
	{
		return 0.0f;
	}
	else
	{
		return m_volumes[m_assignedSplits + 1] - m_volumes[m_assignedSplits + 2];
	}
}

struct minGainFromLastSplit : public binary_function<Entry*, Entry*, bool>
{
	bool operator()(Entry* __x, Entry* __y) const { return __x->gainFromLastSplit() > __y->gainFromLastSplit(); }
};

struct maxGainWithOneMoreSplit : public binary_function<Entry*, Entry*, bool>
{
	bool operator()(Entry* __x, Entry* __y) const { return __x->gainWithOneMoreSplit() < __y->gainWithOneMoreSplit(); }
};

struct maxGainWithTwoMoreSplits : public binary_function<Entry*, Entry*, bool>
{
	bool operator()(Entry* __x, Entry* __y) const
	{
		float xf1 = __x->gainWithOneMoreSplit();
		float xf2 = __x->gainWithTwoMoreSplits();
		float yf1 = __y->gainWithOneMoreSplit();
		float yf2 = __y->gainWithTwoMoreSplits();

		float x = (xf2 > xf1) ? xf2 : -numeric_limits<float>::max();
		float y = (yf2 > yf1) ? yf2 : -numeric_limits<float>::max();

		return x < y;
	}
};

int main(int argc, char** argv)
{
	if (argc != 3)
	{
		cerr << "Usage: " << argv[0] << " split_file total_splits." << endl;
		return -1;
	}

	int splits = atoi(argv[2]);

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open file " << argv[1] << "." << endl;
		return -1;
	}

	priority_queue<Entry*, vector<Entry*>, maxGainWithOneMoreSplit> pq;
	char buf[1000];
	double totalVolume = 0.0;
	double skippedVolume = 0.0;

	while (fin.getline(buf, 1000))
	{
		while(fin && strlen(buf) == 0) fin.getline(buf, 1000);

		SplitInformation si;
		int totalSplits, currentSplit;
		float volume;

		while (fin && strlen(buf) != 0)
		{
			istringstream ss(buf);
			ss >> si.m_id >> totalSplits >> currentSplit >> volume;
			si.m_volumes.push_back(volume);
			fin.getline(buf, 1000);
		}

		totalVolume += si.m_volumes[0];

		if (si.getTotalSplits() == 0)
		{
			skippedVolume += si.m_volumes[0];
			cout << si.m_id << "\t0" << endl;
			continue;
		}

		Entry* e = new Entry(si);
		pq.push(e);
	}

	cpu_usage_aid cpu;
	begin_cpu_usage(&cpu);

	// distribute considering one split at a time.
	while (splits > 0)
	{
		Entry* e = pq.top(); pq.pop();
		//cerr << splits << " " << e.id << " " << e.gainWithOneMoreSplit() << endl;
		if (e->gainWithOneMoreSplit() <= 0.0f) break;
		e->m_assignedSplits++;
		pq.push(e);
		splits--;
	}

	// re-arrange according to minimum gain for the last split and maximum gain for two new splits.
	vector<Entry*> minGain;
	vector<Entry*> maxGain;

	while (! pq.empty())
	{
		Entry* e = pq.top(); pq.pop();

		// smallest element last.
		vector<Entry*>::iterator it = upper_bound(minGain.begin(), minGain.end(), e, minGainFromLastSplit());
		minGain.insert(it, e);

		// largest element last.
		it = upper_bound(maxGain.begin(), maxGain.end(), e, maxGainWithTwoMoreSplits());
		maxGain.insert(it, e);
	}

/*
	for (unsigned int cIndex = 0; cIndex < maxGain.size(); cIndex++)
	{
		cerr << maxGain[cIndex]->gainWithTwoMoreSplits() << endl;
	}

	for (unsigned int cIndex = 0; cIndex < minGain.size(); cIndex++)
	{
		cerr << minGain[cIndex]->gainFromLastSplit() << endl;
	}
*/

	bool bContinue = true;
	int changes = 0;

	while (bContinue)
	{
		bContinue = false;

		// remove top element from maxGain list.
		Entry* max = maxGain.back(); maxGain.pop_back();
		if (max->gainWithTwoMoreSplits() <= 0.0f) break;
		  // maxGain is missing one element here. Don't used it outside of the while loop!

		// remove two top elements from minGain list.
		Entry* min1 = minGain.back(); minGain.pop_back();
		Entry* min2 = minGain.back(); minGain.pop_back();

		// check if max is the same as mins and remove something else instead.
		if (max == min1)
		{
			Entry* t = min1;
			min1 = min2;
			min2 = minGain.back(); minGain.pop_back();
			minGain.push_back(t);
		}
		else if (max == min2)
		{
			Entry* t = min2;
			min2 = minGain.back(); minGain.pop_back();
			minGain.push_back(t);
		}

		// if it is worth rearranging...
		if (max->gainWithTwoMoreSplits() > min1->gainFromLastSplit() + min2->gainFromLastSplit())
		{
			changes++;
			bContinue = true;

			// locate max in minGain list and remove it.
			vector<Entry*>::iterator it = lower_bound(minGain.begin(), minGain.end(), max, minGainFromLastSplit());
			while (*it != max)
			{
				assert(it != minGain.end());
				it++;
			}
			minGain.erase(it);

			// locate mins in maxGain list and remove them.
			it = lower_bound(maxGain.begin(), maxGain.end(), min1, maxGainWithTwoMoreSplits());
			while (*it != min1)
			{
				assert(it != minGain.end());
				it++;
			}
			maxGain.erase(it);
			it = lower_bound(maxGain.begin(), maxGain.end(), min2, maxGainWithTwoMoreSplits());
			while (*it != min2)
			{
				assert(it != minGain.end());
				it++;
			}
			maxGain.erase(it);

			// rearrange the splits.
			max->m_assignedSplits += 2;
			min1->m_assignedSplits--;
			min2->m_assignedSplits--;

			// add max and mins in minGain list.
			it = upper_bound(minGain.begin(), minGain.end(), max, minGainFromLastSplit());
			minGain.insert(it, max);
			it = upper_bound(minGain.begin(), minGain.end(), min1, minGainFromLastSplit());
			minGain.insert(it, min1);
			it = upper_bound(minGain.begin(), minGain.end(), min2, minGainFromLastSplit());
			minGain.insert(it, min2);

			// add max and mins in maxGain list.
			it = upper_bound(maxGain.begin(), maxGain.end(), max, maxGainWithTwoMoreSplits());
			maxGain.insert(it, max);
			it = upper_bound(maxGain.begin(), maxGain.end(), min1, maxGainWithTwoMoreSplits());
			maxGain.insert(it, min1);
			it = upper_bound(maxGain.begin(), maxGain.end(), min2, maxGainWithTwoMoreSplits());
			maxGain.insert(it, min2);
		}
	}

	cpu_usage cpuu;
	end_cpu_usage(&cpu, &cpuu);

	double lagreedy = 0.0;

	for (unsigned int i = 0; i < minGain.size(); i++)
	{
		cout << minGain[i]->m_id << "\t" << minGain[i]->m_assignedSplits << endl;
		lagreedy += minGain[i]->m_volumes[minGain[i]->m_assignedSplits];
		delete minGain[i];
	}

	cerr << "Number of changes: " << changes << endl;
	cerr << "Total: " << totalVolume << endl;
	cerr << "LAGreedy: " << lagreedy + skippedVolume << endl;
	cerr << "Gain: " << totalVolume - (lagreedy + skippedVolume) << endl;
	cerr << "Total time: " << cpuu.total_time << endl;
	cerr << "System time: " << cpuu.system_time << endl;
	cerr << "User time: " << cpuu.user_time << endl;

	return 0;
}
