// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

// example of a Visitor pattern.
// findes the index and leaf IO for answering the query and prints
// the resulting data IDs to stdout.
class MyVisitor : public IVisitor
{
public:
	int m_indexIO;
	int m_leafIO;

public:
	MyVisitor() : m_indexIO(0), m_leafIO(0) {}

	void visitNode(const INode& n)
	{
		if (n.isLeaf()) m_leafIO++;
		else m_indexIO++;
	}

	void visitData(const IData& d)
	{
		//Region* pr = static_cast<Region*>(d.getShape());
			// do something.
		//delete pr;

		// data should be an array of characters representing a Region as a string.
		//byte* pData = 0;
		//unsigned long cLen = 0;
		//d.getData(cLen, &pData);
			// do something.
		//delete[] pData;

		cout << d.getIdentifier() << endl;
			// the ID of this data entry is an answer to the query. I will just print it to stdout.
	}
};

int main(int argc, char** argv)
{
	try
	{
		if (argc != 4)
		{
			cerr << "Usage: " << argv[0] << " input_file tree_file buffer_size." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open query file " << argv[1] << "." << endl;
			return -1;
		}

		string baseName = argv[2];
		IStorageManager* diskfile = StorageManager::loadDiskStorageManager(baseName);
		StorageManager::IBuffer* file = StorageManager::createNewRandomEvictionsBuffer(*diskfile, atoi(argv[3]), false);
		ISpatialIndex* tree = MVRTree::loadMVRTree(*file, 1);

		int count = 0;
		unsigned long indexIO = 0;
		unsigned long leafIO = 0;
		unsigned long hits = 0;

		int id;
		float x1, x2, y1, y2, t1, t2;
		double plow[2], phigh[2];

		cpu_usage_aid cpu;
		begin_cpu_usage(&cpu);

		while (fin)
		{
			fin >> id >> x1 >> x2 >> y1 >> y2 >> t1 >> t2;
			if (! fin.good()) continue;

			plow[0] = x1; plow[1] = y1;
			phigh[0] = x2; phigh[1] = y2;

			MyVisitor vis;

			TimeRegion r = TimeRegion(plow, phigh, t1, t2, 2);
			file->clear();
			tree->intersectsWithQuery(r, vis);
				// this will find all data that intersect with the query range.

			indexIO += vis.m_indexIO;
			leafIO += vis.m_leafIO;
				// example of the Visitor pattern usage, for calculating how many nodes
				// were visited.

			hits += file->getHits();

			if ((count % 1000) == 0) cerr << count << endl;
			count++;
		}

		cpu_usage cpuu;
		end_cpu_usage(&cpu, &cpuu);

		cerr << "Operations: " << count << endl;
		cerr << *tree;
		cerr << "Index I/O: " << indexIO << endl;
		cerr << "Leaf I/O: " << leafIO << endl;
		cerr << "Buffer hits: " << hits << endl;
		cerr << "Actual I/O: " << indexIO + leafIO - hits << endl;
		cerr << "Average I/Os per query: " << static_cast<double>(indexIO + leafIO - hits) / static_cast<double>(count) << endl;
		cerr << "Total time: " << cpuu.total_time << endl;
		cerr << "System time: " << cpuu.system_time << endl;
		cerr << "User time: " << cpuu.user_time << endl;

		delete tree;
		delete file;
		delete diskfile;
	}
	catch (Tools::Exception& e)
	{
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
