// Efficient Indexing of Spatiotemporal Objects
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

int main(int argc, char** argv)
{
	try
	{
		if (argc != 4)
		{
			cerr << "Usage: " << argv[0] << " input_file tree_file capacity." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open data file " << argv[1] << "." << endl;
			return -1;
		}

		string baseName = argv[2];
		IStorageManager* diskfile = StorageManager::createNewDiskStorageManager(baseName, 2048);
		StorageManager::IBuffer* file = StorageManager::createNewRandomEvictionsBuffer(*diskfile, 100, false);

		long indexIdentifier;
		ISpatialIndex* tree = RTree::createNewRTree(*file, 0.7, atoi(argv[3]), atoi(argv[3]), 3, SpatialIndex::RTree::RV_RSTAR, indexIdentifier);

		int count = 0;
		int id;
		float x1, x2, y1, y2, t1, t2;
		double plow[3], phigh[3];

		cpu_usage_aid cpu;
		begin_cpu_usage(&cpu);

		while (fin)
		{
			fin >> id >> x1 >> x2 >> y1 >> y2 >> t1 >> t2;
			if (! fin.good()) continue; // skip newlines etc.

			plow[0] = static_cast<double>(x1); plow[1] = static_cast<double>(y1); plow[2] = static_cast<double>(t1);
			phigh[0] = static_cast<double>(x2); phigh[1] = static_cast<double>(y2); phigh[2] = static_cast<double>(t2);
			Region r = Region(plow, phigh, 3);

			try
			{
				tree->insertData(0, 0, r, id);
			}
			catch (Tools::Exception& e)
			{
				std::string s = e.what();
				cerr << s << endl;
				return -1;
			}
			catch (...)
			{
				cerr << "other exception" << endl;
			}

			if (count % 10000 == 0) cerr << count << endl;
			count++;
		}

		cpu_usage cpuu;
		end_cpu_usage(&cpu, &cpuu);

		cerr << "Operations: " << count << endl;
		cerr << *tree;
		cerr << "Total time: " << cpuu.total_time << endl;
		cerr << "System time: " << cpuu.system_time << endl;
		cerr << "User time: " << cpuu.user_time << endl;
		cerr << "Index ID: " << indexIdentifier << endl;

		bool ret = tree->isIndexValid();
		if (ret == false) cerr << "Structure is INVALID!" << endl;

		delete tree;
		delete file;
		delete diskfile;
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
